/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

bufferlist_t	buffer_list = TAILQ_HEAD_INITIALIZER(buffer_list);

buffer_t *buffer_create(u_char *name, int internal) {
	buffer_t *buffer;

	buffer = ckmalloc(sizeof(buffer_t));
	TAILQ_INIT(&buffer->lines);
	TAILQ_INIT(&buffer->undolist);
	TAILQ_INIT(&buffer->marklist);
	line_insert(&buffer->lines, NULL);
	buffer->linecnt = 1;
	buffer->undoitem = NULL;
	buffer->undodisk = NULL;
	buffer->dispname = filepart(name);
	buffer->name = ckstrdup(name);
	buffer->mark = NULL;
	buffer->markcnt = 0;
	buffer->mark_implicit = 1;
	buffer->modified = 0;
	buffer->internal = internal;
	buffer->dirty = 1;
	buffer->refcnt = 0;
	buffer->nlstyle = NLSTYLE_LF;
	TAILQ_INSERT_HEAD(&buffer_list, buffer, b_list);

	return buffer;
}

void buffer_destroy(buffer_t *buffer) {
	mark_t *mark;

	undo_emptylist(buffer);
	while (!TAILQ_EMPTY(&buffer->marklist)) {
		mark = TAILQ_FIRST(&buffer->marklist);
		TAILQ_REMOVE(&buffer->marklist, mark, m_list);
		free(mark);
	}
	while (!TAILQ_EMPTY(&buffer->lines))
		line_remove(&buffer->lines, TAILQ_FIRST(&buffer->lines));
	TAILQ_REMOVE(&buffer_list, buffer, b_list);
	free(buffer->dispname);
	free(buffer->name);
	free(buffer);
}

int buffer_save(buffer_t *buffer) {
	if (file_write(buffer, buffer->name))
		return -1;

	buffer->modified = 0;
	buffer->undodisk = buffer->undoitem ? buffer->undoitem :
		TAILQ_LAST(&buffer->undolist, undolist);
	buffer->dirty = 1;

	return 0;
}

void buffer_pushmark(buffer_t *buffer, line_t *line, int linenum, int col) {
	mark_t *mark;

	if (buffer->markcnt > options.markringmax) {
		mark = TAILQ_LAST(&buffer->marklist, marklist);
		TAILQ_REMOVE(&buffer->marklist, mark, m_list);
	} else {
		mark = ckmalloc(sizeof(mark_t));
		buffer->markcnt++;
	}
	mark->line = line;
	mark->linenum = linenum;
	mark->col = col;
	TAILQ_INSERT_HEAD(&buffer->marklist, mark, m_list);
	buffer->mark = mark;
	buffer->dirty = 1;
}

void buffer_popmark(buffer_t *buffer) {
	buffer->mark = TAILQ_NEXT(buffer->mark, m_list);
	if (!buffer->mark)
		buffer->mark = TAILQ_FIRST(&buffer->marklist);
}

void buffer_empty(buffer_t *buffer, int undoable) {
	if (!TAILQ_FIRST(&buffer->lines))
		return;

	if (undoable)
		undo_beginblock(buffer);
	while (TAILQ_NEXT(TAILQ_FIRST(&buffer->lines), l_list))
		buffer_rmline(buffer, undoable,
			TAILQ_NEXT(TAILQ_FIRST(&buffer->lines), l_list), 1);
	buffer_rmstr(buffer, undoable, TAILQ_FIRST(&buffer->lines), 0, 0,
		TAILQ_FIRST(&buffer->lines)->length);
	if (undoable)
		undo_endblock(buffer);
}

/*
 * the following buffer operation wrappers do what their line_
 * counterparts do, but also add undo deltas to the buffer's
 * undolist and set the dirty and modified flags.  the routines
 * also call into the notify operations to all views onto that
 * buffer to prevent views from using references to lines/cols
 * after they no longer exist, or using linenums that aren't
 * consisent.
 */

line_t *buffer_addline(buffer_t *buffer, int undoable, line_t *line,
		int linenum) {
	line_t *ret;
	undo_t *undo;
	mark_t *mark;

	/*
	 * because line can be null, the semantics of how linenum
	 * should be set are slightly complicated; they should
	 * satisfy the following debug assertion.
	 */
	assert(line ? linenum >= 0 : linenum == -1);

	if (undoable) {
		undo = ckmalloc(sizeof(undo_t));
		undo->type = UNDO_LINEINSERT;
		undo->linenum = linenum;
		undo_additem(buffer, undo);
	}
	buffer->dirty = 1;
	buffer->modified = 1;
	ret = line_insert(&buffer->lines, line);
	buffer->linecnt++;
	VIEWCALLB(buffer, addline, ret, linenum);

	/* keep mark positions valid */
	TAILQ_FOREACH(mark, &buffer->marklist, m_list)
		if (linenum < mark->linenum)
			mark->linenum++;

	return ret;
}

void buffer_rmline(buffer_t *buffer, int undoable, line_t *line,
		int linenum) {
	undo_t *undo;
	mark_t *mark;
	int needblk;

	if (undoable) {
		if (line->length) {
			undo_beginblock(buffer);
			buffer_rmstr(buffer, undoable, line,
				linenum, 0, line->length);
			needblk = 1;
		} else
			needblk = 0;
		undo = ckmalloc(sizeof(undo_t));
		undo->type = UNDO_LINEDELETE;
		undo->linenum = linenum;
		undo_additem(buffer, undo);
		if (needblk)
			undo_endblock(buffer);
	}
	buffer->dirty = 1;
	buffer->modified = 1;
	VIEWCALLB(buffer, rmline, line, linenum);

	/* keep mark positions valid */
	TAILQ_FOREACH(mark, &buffer->marklist, m_list)
		if (mark->line == line) {
			mark->line = TAILQ_PREV(line, linelist, l_list);
			if (!mark->line) {
				mark->line = TAILQ_NEXT(line, l_list);
				mark->col = 0;
			} else {
				mark->linenum--;
				mark->col = mark->line->length;
			}
		} else if (linenum < mark->linenum)
			mark->linenum--;

	line_remove(&buffer->lines, line);
	buffer->linecnt--;
}

static void adjmark_rmstr(buffer_t *buffer, line_t *line,
		int linenum, int loc, int slen) {
	mark_t *mark;

	TAILQ_FOREACH(mark, &buffer->marklist, m_list) {
		if (mark->line != line || mark->col < loc)
			continue;
		if (mark->col < loc + slen)
			mark->col = loc <= line->length -
				slen ? loc : (loc ? loc - 1 : 0);
		else
			mark->col -= slen;
	}
}

void buffer_rmstr(buffer_t *buffer, int undoable, line_t *line,
		int linenum, int loc, int slen) {
	undo_t *undo;

	if (!slen)
		return;

	if (undoable) {
		undo = ckmalloc(sizeof(undo_t));
		undo->type = UNDO_TEXTDELETE;
		undo->linenum = linenum;
		undo->op.text.col = loc;
		undo->op.text.slen = slen;
		undo->op.text.s = ckmalloc(slen);
		memcpy(undo->op.text.s, &line->text[loc], slen);
		undo_additem(buffer, undo);
	}
	buffer->dirty = 1;
	buffer->modified = 1;
	line_rmstr(line, loc, slen);
	VIEWCALLB(buffer, rmstr, line, linenum, loc, slen);
	adjmark_rmstr(buffer, line, linenum, loc, slen);
}

void buffer_rmch(buffer_t *buffer, int undoable, line_t *line,
		int linenum, int loc) {
	undo_t *undo;

	if (undoable) {
		undo = ckmalloc(sizeof(undo_t));
		undo->type = UNDO_RMCHAR;
		undo->linenum = linenum;
		undo->op.ch.col = loc;
		undo->op.ch.c = line->text[loc];
		undo_additem(buffer, undo);
	}
	buffer->dirty = 1;
	buffer->modified = 1;
	line_rmch(line, loc);
	VIEWCALLB(buffer, rmstr, line, linenum, loc, 1);
	adjmark_rmstr(buffer, line, linenum, loc, 1);
}

static void adjmark_stradd(buffer_t *buffer, line_t *line,
		int linenum, int loc, int slen) {
	mark_t *mark;

	TAILQ_FOREACH(mark, &buffer->marklist, m_list)
		if (mark->line == line && loc < mark->col)
			mark->col += slen;
}

void buffer_putc(buffer_t *buffer, int undoable, line_t *line,
		int linenum, int loc, u_char c) {
	undo_t *undo;

	if (undoable) {
		undo = ckmalloc(sizeof(undo_t));
		undo->type = UNDO_PUTCHAR;
		undo->linenum = linenum;
		undo->op.ch.col = loc;
		undo->op.ch.c = c;
		undo_additem(buffer, undo);
	}
	buffer->dirty = 1;
	buffer->modified = 1;
	line_putc(line, loc, c);
	VIEWCALLB(buffer, stradd, line, linenum, loc, 1);
	adjmark_stradd(buffer, line, linenum, loc, 1);
}

void buffer_nputs(buffer_t *buffer, int undoable, line_t *line, int linenum,
		int loc, u_char *s, int slen) {
	undo_t *undo;

	if (!slen)
		return;

	if (undoable) {
		undo = ckmalloc(sizeof(undo_t));
		undo->type = UNDO_TEXTINSERT;
		undo->linenum = linenum;
		undo->op.text.col = loc;
		undo->op.text.slen = slen;
		undo->op.text.s = ckmalloc(slen);
		memcpy(undo->op.text.s, s, slen);
		undo_additem(buffer, undo);
	}
	buffer->dirty = 1;
	buffer->modified = 1;
	line_nputs(line, loc, s, slen);
	VIEWCALLB(buffer, stradd, line, linenum, loc, slen);
	adjmark_stradd(buffer, line, linenum, loc, slen);
}
