/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef BUFFER_H
#define BUFFER_H

/*
 * buffer structure
 */
struct buffer {
	TAILQ_ENTRY(buffer) b_list;

	u_char		*dispname;	/* the displayed name */
	u_char		*name;		/* buffer name/filename */
	linelist_t	lines;		/* list of lines */
	int		linecnt;	/* number of lines in the buffer */
	undolist_t	undolist;	/* list of undoable operations */
	undo_t		*undoitem;	/* current undo item in the list */
	undo_t		*undodisk;	/* the undo loc that is unmodified */
	int		refcnt;		/* buffer reference count */

	/* style of line-terminators used */
	enum {
		NLSTYLE_UNKNOWN		= -1,
		NLSTYLE_LF,
		NLSTYLE_CRLF,
	} nlstyle;

	/*
	 * buffer position marks; buffer marks are positions in a buffer
	 * marked as being important enough to reference again in the
	 * near future.  some commands may choose to set marks for the
	 * user, or the user may explicitly set the mark to a position
	 * in the buffer.  mark_implicit is set if the mark was set
	 * implicitly without user interaction, and some commands may
	 * see fit to treat an implicit mark differently than a mark
	 * which the user set explicitly.
	 */
	struct mark {
		line_t	*line;
		int	linenum;
		int	col;
		TAILQ_ENTRY(mark) m_list;
	};
	TAILQ_HEAD(marklist, mark) marklist;
	mark_t		*mark;
	u_int		markcnt;

	/* buffer flags */
	u_int	mark_implicit:1;
	u_int	modified:1;
	u_int	internal:1;
	u_int	dirty:1;	/* changed since last draw */
};

TAILQ_HEAD(bufferlist, buffer);

extern bufferlist_t	buffer_list;

buffer_t	*buffer_create(u_char *name, int internal);
void		buffer_destroy(buffer_t *);
int		buffer_save(buffer_t *);
void		buffer_pushmark(buffer_t *, line_t *, int linenum, int col);
void		buffer_popmark(buffer_t *);
void		buffer_empty(buffer_t *, int undoable);
line_t		*buffer_addline(buffer_t *, int undoable, line_t *,
			int linenum);
void		buffer_rmline(buffer_t *, int undoable, line_t *,
			int linenum);
void		buffer_rmstr(buffer_t *, int undoable, line_t *,
			int linenum, int loc, int slen);
void		buffer_rmch(buffer_t *, int undoable, line_t *,
			int linenum, int loc);
void		buffer_putc(buffer_t *, int undoable, line_t *,
			int linenum, int loc, u_char);
void		buffer_nputs(buffer_t *, int undoable, line_t *,
			int linenum, int loc, u_char *s, int slen);

/* add a reference to a buffer */
static __inline buffer_t *buffer_ref(buffer_t *buffer) {
	assert(buffer->refcnt >= 0);
	buffer->refcnt++;
	return buffer;
}

/* remove a buffer reference, destroy if the count is 0 */
static __inline void buffer_rele(buffer_t *buffer) {
	assert(buffer->refcnt > 0);
	if (--buffer->refcnt == 0)
		buffer_destroy(buffer);
}

/*
 * find a buffer line by linenumber, this is slow, so don't use
 * it unless it makes sense to.
 */
static __inline line_t *buffer_lineat(buffer_t *buffer, int linenum,
		int *rlinenum) {
	line_t *line;
	int num;

	line = TAILQ_FIRST(&buffer->lines);
	for (num = 0; num < linenum; num++) {
		if (!TAILQ_NEXT(line, l_list))
			break;
		line = TAILQ_NEXT(line, l_list);
	}
	if (rlinenum)
		*rlinenum = num;

	return line;
}

/* remove dirty flags on all buffers */
static __inline void buffer_cleanse() {
	buffer_t *buffer;

	TAILQ_FOREACH(buffer, &buffer_list, b_list)
		buffer->dirty = 0;
}

/* lookup a buffer by dispname */
static __inline buffer_t *buffer_lookup(u_char *name) {
	buffer_t *buffer;

	TAILQ_FOREACH(buffer, &buffer_list, b_list)
		if (!strcmp(buffer->dispname, name))
			return buffer;
	return NULL;
}

#endif
