/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/*
 * basic cursor movement commands
 */

void command_up(vdefault_t *v) {
	if (!TAILQ_PREV(v->cursor.line, linelist, l_list)) {
		screen_bell();
		return;
	}

	v->cursor.line = TAILQ_PREV(v->cursor.line, linelist, l_list);
	v->cursor.linenum--;

	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_down(vdefault_t *v) {
	if (!TAILQ_NEXT(v->cursor.line, l_list)) {
		screen_bell();
		return;
	}

	v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
	v->cursor.linenum++;

	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_left(vdefault_t *v) {
	if (v->cursor.col == 0) {
		if (!TAILQ_PREV(v->cursor.line, linelist, l_list)) {
			screen_bell();
			return;
		}

		v->cursor.line = TAILQ_PREV(v->cursor.line, linelist, l_list);
		v->cursor.col = v->cursor.line->length;
		v->cursor.linenum--;
	} else
		v->cursor.col--;

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_right(vdefault_t *v) {
	if (v->cursor.col == v->cursor.line->length) {
		if (!TAILQ_NEXT(v->cursor.line, l_list)) {
			screen_bell();
			return;
		}

		v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
		v->cursor.linenum++;
		v->cursor.lastcol = v->cursor.col = v->cursor.screencol = 0;
	} else {
		v->cursor.screencol += VIEWCALLI(v->hdr.top, viswid,
			v->cursor.screencol,
			v->cursor.line->text[v->cursor.col]);
		v->cursor.lastcol = v->cursor.screencol;
		v->cursor.col++;
	}

	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

/*
 * move the cursor to the beginning of the line and if
 * stickyhomekey is enabled allow toggling between the front
 * of the line and the first nonwhitespace character
 */
void command_home(vdefault_t *v) {
	int i;

	if (options.stickyhomekey) {
		i = 0;
		while (i < v->cursor.line->length &&
				(v->cursor.line->text[i] == ' '
				|| v->cursor.line->text[i] == '\t'))
			i++;

		if (v->cursor.col <= i && v->cursor.col != 0)
			v->cursor.col = 0;
		else
			v->cursor.col = i;
	} else
		v->cursor.col = 0;

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_end(vdefault_t *v) {
	v->cursor.col = v->cursor.line->length;
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_viewtop(vdefault_t *v) {
	v->cursor.line = v->view.line;
	v->cursor.linenum = v->view.linenum;
	vdefault_trylastcol(v);
	v->redraw_header = 1;
}

void command_viewbottom(vdefault_t *v) {
	int i;

	v->cursor.line = v->view.line;
	v->cursor.linenum = v->view.linenum;
	for (i = v->hdr.geom.height - 1; i; i--) {
		if (!TAILQ_NEXT(v->cursor.line, l_list))
			break;
		v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
		v->cursor.linenum++;
	}

	vdefault_trylastcol(v);
	v->redraw_header = 1;
}

void command_nextword(vdefault_t *v) {

	/* if at the end of a line, behave like right arrow */
	if (v->cursor.col == v->cursor.line->length) {
		command_right(v);
		return;
	}

	v->cursor.col = word_next(v->cursor.line->text,
		v->cursor.line->length, v->cursor.col);
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_prevword(vdefault_t *v) {

	/* at the start of a line, just act like left arrow */
	if (v->cursor.col <= 1) {
		command_left(v);
		return;
	}

	v->cursor.col = word_prev(v->cursor.line->text, v->cursor.col);
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_nextpara(vdefault_t *v) {
	para_next(&v->cursor.line, &v->cursor.linenum);
	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_prevpara(vdefault_t *v) {
	para_prev(&v->cursor.line, &v->cursor.linenum);
	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

