/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

static void writergn(vdefault_t *v, FILE *fp) {
	region_t r;
	line_t *line;
	u_char *s;
	int linenum, col;
	int slen, last;

	vdefault_getmarkrgn(v, &r);
	line = NULL;
	for (;;) {
		last = region_nextstr(v->hdr.buffer, &r, &line, &linenum,
			&col, &s, &slen);
		if (slen)
			fwrite(s, 1, slen, fp);
		if (last)
			break;

		if (v->hdr.buffer->nlstyle == NLSTYLE_CRLF)
			putc('\r', fp);
		putc('\n', fp);
	}

	v->cursor.line = r.first.line;
	v->cursor.linenum = r.first.linenum;
	v->cursor.col = r.first.col;
	region_delete(v->hdr.buffer, 1, &r);
}

/*
 * run the entire buffer, or the marked portion of the buffer through
 * a filter command specified by the user.
 */
void command_filter(vdefault_t *v) {
	int infds[2], outfds[2];
	line_t *line;
	FILE *fp;
	u_char *cmd;
	int linenum, col;
	pid_t pid;

	/* prompt the user for the command to use */
	cmd = minibuff_prompt("Filter command", NULL);
	if (!cmd)
		return;

	/* create pipes for each direction */
	if (pipe(infds) == -1 || pipe(outfds) == -1) {
		minibuff_printf("Pipe failed: %s", strerror(errno));
		goto bail;
	}

	input_begincancel();

	/*
	 * fork, dup the pipe descriptors into standard input and
	 * standard output, and then run the buffer (or the marked
	 * portion) through the filter.
	 */
	switch (pid = fork()) {
	case 0:
		/* restore ignored signals */
		signal(SIGPIPE, SIG_DFL);
		signal(SIGINT, SIG_DFL);

		/* set up file descriptors */
		dup2(infds[0], STDIN_FILENO);
		dup2(outfds[1], STDOUT_FILENO);
		dup2(outfds[1], STDERR_FILENO);

		/* close the other descriptors in the child */
		close(infds[0]);
		close(infds[1]);
		close(outfds[0]);
		close(outfds[1]);

		/* pass it off to bourne */
		execl(_PATH_BSHELL, "sh", "-c", cmd, NULL);
		_exit(0);
	case -1:
		minibuff_printf("Couldn't fork: %s", strerror(errno));
		goto bail;
	}

	/* close the child's ends of the descriptors in the parent */
	close(infds[0]);
	close(outfds[1]);

	/* this whole thing is undoable */
	undo_beginblock(v->hdr.buffer);

	/*
	 * write the portion of the buffer being filtered out to
	 * the child process, and then remove that portion of
	 * the buffer.
	 */
	fp = fdopen(infds[1], "wb");
	if (v->hdr.buffer->mark_implicit || !v->hdr.buffer->mark) {
		file_fwrite(v->hdr.buffer, fp);
		buffer_empty(v->hdr.buffer, 1);
	} else
		writergn(v, fp);
	fclose(fp);

	/* read data back from the child */
	line = v->cursor.line;
	linenum = v->cursor.linenum;
	col = v->cursor.col;
	fp = fdopen(outfds[0], "rb");
	if (file_fread(v->hdr.buffer, 1, 1, &line, &linenum, &col, fp))
		minibuff_printf("Read error: %s",
			strerror(errno));
	fclose(fp);
	if (input_checkcancel()) {
		kill(pid, SIGTERM);
		goto bail;
	}

	/* put the mark at the end */
	if (v->hdr.buffer->mark) {
		v->hdr.buffer->mark->line = line;
		v->hdr.buffer->mark->linenum = linenum;
		v->hdr.buffer->mark->col = col;
	} else
		buffer_pushmark(v->hdr.buffer, line, linenum, col);
	v->hdr.buffer->mark_implicit = 1;

	undo_endblock(v->hdr.buffer);
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
bail:
	input_endcancel();
	free(cmd);
}
