/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/*
 * flags if the last yank command was successful.  A yank command can
 * fail if conditions aren't appropriate for a yank: i.e., yank will
 * fail if the killring has no buffers in it, and yank-pop will fail
 * if the last command wasn't a successful yank command.
 */
static int yank_succeeded;

/*
 * used by kill commands; returns a pointer to the killbuffer they should
 * use; if the last command wasn't a kill command, allocate a new buffer
 * on the top of the killring to use, otherwise just use the top one.
 */
static __inline killbuffer_t *ringmanage() {
	if (input_lastkeysym == KEYSYM_KILLLINE ||
			input_lastkeysym == KEYSYM_KILLEOL ||
			input_lastkeysym == KEYSYM_KILLWORD ||
			input_lastkeysym == KEYSYM_FWDKILLWORD ||
			input_lastkeysym == KEYSYM_APPENDKILL)
		return TAILQ_FIRST(&killring.buffers);
	return killring_addbuf(&killring);
}

/*
 * kill the region defined by the cursor and the mark; the
 * normal kill routines use this when the mark is explicit.
 */
void command_killrgn(vdefault_t *v) {
	region_t r;
	killbuffer_t *killbuff;
	line_t *line, *kbline;
	u_char *s;
	int linenum, col;
	int slen, last;

	if (!v->hdr.buffer->mark) {
		minibuff_printf("Mark isn't set.");
		return;
	}

	v->rm_protect = PROTECT_CURSOR;
	killbuff = ringmanage();
	kbline = TAILQ_LAST(&killbuff->lines, linelist);
	if (!kbline)
		kbline = line_insert(&killbuff->lines, NULL);

	/* put each line onto the killbuffer */
	vdefault_getmarkrgn(v, &r);
	line = NULL;
	for (;;) {
		last = region_nextstr(v->hdr.buffer, &r, &line, &linenum,
			&col, &s, &slen);
		line_nputs(kbline, kbline->length, s, slen);
		if (last)
			break;

		kbline = line_insert(&killbuff->lines, kbline);
	}
	killbuff->trailnl = 0;

	v->cursor.line = r.first.line;
	v->cursor.linenum = r.first.linenum;
	v->cursor.col = r.first.col;
	region_delete(v->hdr.buffer, 1, &r);
	v->hdr.buffer->mark_implicit = 1;

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
}

/*
 * remove the current line from the buffer and append it to
 * the killring.
 */
void command_killline(vdefault_t *v) {
	killbuffer_t *killbuff;
	line_t *line, *prev;

	killbuff = ringmanage();
	if (!v->hdr.buffer->mark_implicit) {
		command_killrgn(v);
		return;
	}
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/* add new line to the killbuffer */
	line = line_insert(&killbuff->lines, TAILQ_LAST(&killbuff->lines,
		linelist));
	line_nputs(line, 0, v->cursor.line->text, v->cursor.line->length);
	killbuff->trailnl = 1;

	/*
	 * remove the line from the buffer, keep the cursor on the same
	 * line number, if the line is the last line in the buffer, don't
	 * remove the line.
	 */
	if (v->cursor.line != TAILQ_LAST(&v->hdr.buffer->lines, linelist)) {
		prev = TAILQ_PREV(v->cursor.line, linelist, l_list);
		if (v->view.line == v->cursor.line)
			v->view.line = NULL;
		buffer_rmline(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum);
		if (prev) {
			v->cursor.line = TAILQ_NEXT(prev, l_list);
			if (!v->cursor.line) {
				v->cursor.line = prev;
				v->cursor.linenum--;
			}
		} else
			v->cursor.line = TAILQ_FIRST(&v->hdr.buffer->lines);
		if (!v->view.line) {
			v->view.line = v->cursor.line;
			v->view.linenum = v->cursor.linenum;
		}
	} else if (v->cursor.line->length)
		buffer_rmstr(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, 0, v->cursor.line->length);
	v->cursor.col = 0;

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/*
 * kill to the end of the current line; if on the end of the line,
 * kill the line termination.
 */
void command_killeol(vdefault_t *v) {
	killbuffer_t *killbuff;
	line_t *line;

	killbuff = ringmanage();
	if (!v->hdr.buffer->mark_implicit) {
		command_killrgn(v);
		return;
	}
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	line = TAILQ_LAST(&killbuff->lines, linelist);
	if (!line) {
		line = line_insert(&killbuff->lines, NULL);
		killbuff->trailnl = 0;
	}
	if (v->cursor.col < v->cursor.line->length) {
		if (killbuff->trailnl) {
			line = line_insert(&killbuff->lines, line);
			killbuff->trailnl = 0;
		}
		line_nputs(line, line->length,
			&v->cursor.line->text[v->cursor.col],
			v->cursor.line->length - v->cursor.col);
		buffer_rmstr(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, v->cursor.col,
			v->cursor.line->length - v->cursor.col);
	} else {
		line_insert(&killbuff->lines, line);
		killbuff->trailnl = 0;
		line = TAILQ_NEXT(v->cursor.line, l_list);
		if (!line)
			return;
		undo_beginblock(v->hdr.buffer);
		buffer_nputs(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, v->cursor.col,
			line->text, line->length);
		buffer_rmline(v->hdr.buffer, 1, line, v->cursor.linenum + 1);
		undo_endblock(v->hdr.buffer);
	}

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/*
 * kill backwards to the start of the previous word and prepend the
 * word to the first line of the killring; killing across a line
 * adds a new line to the front of the killring.
 */
void command_killword(vdefault_t *v) {
	killbuffer_t *killbuff;
	line_t *line;
	int i;

	killbuff = ringmanage();
	if (!v->hdr.buffer->mark_implicit) {
		command_killrgn(v);
		return;
	}
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/*
	 * in column 0, just add a new line to the killbuffer, because
	 * of the way killword works backwards, the user will expect
	 * the new line to be in front of the first line
	 */
	if (v->cursor.col == 0) {
		if (!TAILQ_PREV(v->cursor.line, linelist, l_list))
			return;
		if (!TAILQ_FIRST(&killbuff->lines))
			killbuff->trailnl = 1;
		line_insert(&killbuff->lines, NULL);
		command_backspace(v);
		return;
	}

	/* get the start of the previous word */
	if (v->cursor.col > 1)
		i = word_prev(v->cursor.line->text, v->cursor.col);
	else
		i = 0;

	/* prepend it to the first line of the killbuffer */
	line = TAILQ_FIRST(&killbuff->lines);
	if (!line) {
		line = line_insert(&killbuff->lines, NULL);
		killbuff->trailnl = 0;
	}
	line_nputs(line, 0, &v->cursor.line->text[i], v->cursor.col - i);
	buffer_rmstr(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		i, v->cursor.col - i);

	v->cursor.col = i;
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

/* kill forward until the end of a word */
void command_fwdkillword(vdefault_t *v) {
	killbuffer_t *killbuff;
	line_t *line;
	int i;

	killbuff = ringmanage();
	if (!v->hdr.buffer->mark_implicit) {
		command_killrgn(v);
		return;
	}
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/*
	 * at the end of line, just add the newline to the killbuffer
	 * and behave like delete.
	 */
	if (v->cursor.col == v->cursor.line->length) {
		if (!TAILQ_NEXT(v->cursor.line, l_list))
			return;
		if (!TAILQ_FIRST(&killbuff->lines))
			line_insert(&killbuff->lines, NULL);
		line_insert(&killbuff->lines, TAILQ_LAST(&killbuff->lines,
			linelist));
		killbuff->trailnl = 0;
		command_delete(v);
		return;
	}

	/* get the next end of word boundary */
	i = word_next(v->cursor.line->text, v->cursor.line->length,
		v->cursor.col);

	/* append the word to the last line of the killbuffer */
	line = TAILQ_LAST(&killbuff->lines, linelist);
	if (!line) {
		line = line_insert(&killbuff->lines, NULL);
		killbuff->trailnl = 0;
	}
	line_nputs(line, line->length, &v->cursor.line->text[v->cursor.col],
		i - v->cursor.col);
	buffer_rmstr(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		v->cursor.col, i - v->cursor.col);

	v->redraw_text = 1;
}

void command_appendkill(vdefault_t *v) {
	minibuff_printf("If the next command is a kill, it will append.");
}

static void doyank(vdefault_t *v, int pushmark) {
	killbuffer_t *killbuff;
	line_t *line, *next, *start_line;
	int start_linenum, start_col;

	killbuff = killring.current;
	if (!killbuff) {
		yank_succeeded = 0;
		return;
	}
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/* save the cursor position */
	start_line = v->cursor.line;
	start_linenum = v->cursor.linenum;
	start_col = v->cursor.col;

	/*
	 * copy the data for each kill buffer line and place it on
	 * the buffer.
	 */
	undo_beginblock(v->hdr.buffer);
	TAILQ_FOREACH(line, &killbuff->lines, l_list) {
		if (line != TAILQ_LAST(&killbuff->lines, linelist) ||
				killbuff->trailnl) {
			next = buffer_addline(v->hdr.buffer, 1, v->cursor.line,
				v->cursor.linenum);
			if (v->cursor.col < v->cursor.line->length) {
				buffer_nputs(v->hdr.buffer, 1, next,
					v->cursor.linenum + 1, 0,
					&v->cursor.line->text[v->cursor.col],
					v->cursor.line->length -
					v->cursor.col);
				buffer_rmstr(v->hdr.buffer, 1, v->cursor.line,
					v->cursor.linenum, v->cursor.col,
					v->cursor.line->length -
					v->cursor.col);
			}
		} else
			next = NULL;

		buffer_nputs(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, v->cursor.col, line->text,
			line->length);

		if (next) {
			v->cursor.linenum++;
			v->cursor.line = next;
			v->cursor.col = 0;
		} else
			v->cursor.col = v->cursor.col + line->length;
	}
	undo_endblock(v->hdr.buffer);

	/* set the mark to the position before the yank */
	if (pushmark) {
		buffer_pushmark(v->hdr.buffer, start_line,
			start_linenum, start_col);
		v->hdr.buffer->mark_implicit = 1;
	} else {
		assert(v->hdr.buffer->mark);
		v->hdr.buffer->mark->line = start_line;
		v->hdr.buffer->mark->linenum = start_linenum;
		v->hdr.buffer->mark->col = start_col;
	}
	minibuff_printf("Mark set.");

	yank_succeeded = 1;
}

void command_yank(vdefault_t *v) {
	doyank(v, 1);
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

static __inline int popok() {
	if ((input_lastkeysym != KEYSYM_YANK &&
			input_lastkeysym != KEYSYM_YANKPOP &&
			input_lastkeysym != KEYSYM_FWDYANKPOP) ||
			!yank_succeeded) {
		minibuff_printf("Previous command was not a yank.");
		yank_succeeded = 0;
		return 0;
	}

	return 1;
}

static void dopop(vdefault_t *v) {
	region_t r;

	v->rm_protect = PROTECT_CURSOR;

	/*
	 * delete the marked region of the buffer (which contains the
	 * text just yanked), adjust the current killbuffer pointer
	 * in the killring, and yank from it.
	 */
	undo_beginblock(v->hdr.buffer);
	vdefault_getmarkrgn(v, &r);
	v->cursor.line = r.first.line;
	v->cursor.linenum = r.first.linenum;
	v->cursor.col = r.first.col;
	region_delete(v->hdr.buffer, 1, &r);
	doyank(v, 0);
	undo_endblock(v->hdr.buffer);

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_yankpop(vdefault_t *v) {
	if (!popok())
		return;

	killring.current = TAILQ_NEXT(killring.current, kb_list);
	if (!killring.current)
		killring.current = TAILQ_FIRST(&killring.buffers);
	dopop(v);
}

void command_fwdyankpop(vdefault_t *v) {
	if (!popok())
		return;

	killring.current = TAILQ_PREV(killring.current, killbufferlist, kb_list);
	if (!killring.current)
		killring.current = TAILQ_LAST(&killring.buffers, killbufferlist);
	dopop(v);
}
