/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* storage of the last accepted search */
static u_char	*last_text;
static int	last_len;

/* used to talk to isearchloop from the searchback */
static enum {
	ACT_NONE,
	ACT_ACCEPT,
	ACT_CANCEL,
	ACT_BACKUP,
	ACT_RECISEARCH,
	ACT_RECRISEARCH
} saction;

/*
 * searching vars needed by the input handler callbacks
 */
static int		sidx;
static int		sprevidx;
static int		searchfromsave;
static int		sfailed;
static int		sfirst;
static int		sfwd;		/* 1 = fwd search, 0 = backward */
static int		switched;	/* 1 if fwd was inverted */
static vdefault_t	*sview;

/* save and current find positions */
static line_t	*find_line, *save_line;
static int	find_linenum, find_col, find_idx;
static int	save_linenum, save_col;

/*
 * move the passed in line pointers forward by count; no error
 * checking is performed, the usage of this below must be certain
 * that it may move forward by that amount
 */
static __inline void mvfwd(int count, line_t **line, int *linenum,
		int *col) {
	while (count--)
		if (++(*col) > (*line)->length) {
			*col = 0;
			*line = TAILQ_NEXT(*line, l_list);
			(*linenum)++;
		}
}

/* input callback for search prompt */
static int searchcback(vminibuff_t *v, int keysym) {
	killbuffer_t *killbuff;

	if (input_lastkeysym == KEYSYM_QUOTENEXT) {
		sidx++;
		return 0;
	}

	switch (keysym) {
	case '\n': case '\r':
	case KEYSYM_ENTER:
		saction = ACT_ACCEPT;
		vminibuff_exit(v);
		break;
	case KEYSYM_CANCEL:
		saction = ACT_CANCEL;
		vminibuff_exit(v);
		break;
	case KEYSYM_BACKSPACE:
		if (sidx > sprevidx) {
			searchfromsave = 1;
			sidx--;
			return 0;
		} else {
			saction = ACT_BACKUP;
			vminibuff_tmpexit(v);
		}
		break;

	/*
	 * pressing the key for nextword causes the text up till the
	 * next word boundary after the cursor to be added to the
	 * search term
	 */
	case KEYSYM_NEXTWORD: {
		vdefault_t *vdef;
		int next;

		if (!sfwd)
			mvfwd(sidx, &sview->cursor.line,
				&sview->cursor.linenum, &sview->cursor.col);

		/* at the end of a line, insert a newline */
		if (sview->cursor.col == sview->cursor.line->length) {
			if (!TAILQ_NEXT(sview->cursor.line, l_list))
				break;
			line_putc(minibuff_line, sidx++, '\n');
			if (!sfwd)
				mvfwd(1, &sview->cursor.line,
					&sview->cursor.linenum,
					&sview->cursor.col);
			goto done;
		}

		/* add text till the next word boundary */
		next = word_next(sview->cursor.line->text,
			sview->cursor.line->length, sview->cursor.col);
		line_nputs(minibuff_line, sidx,
			&sview->cursor.line->text[sview->cursor.col],
			next - sview->cursor.col);
		sidx += next - sview->cursor.col;
		if (!sfwd)
			sview->cursor.col = next;

done:
		if (!sfwd) {
			/*
			 * cause the search to find this term this is done
			 * so that any number of words can be pulled out of
			 * the text before actually looking for other
			 * occurances.
			 */
			find_line = sview->cursor.line;
			find_linenum = sview->cursor.linenum;
			find_col = sview->cursor.col;
		}
		vdef = view_findvdef((viewhdr_t *) v);
		vdef->redraw_text = 1;
		command_end(vdef);
		break;
	}


	/*
	 * prevent multiline yanks.
	 *
	 * XXX: it'd be nicer to just convert the lines into ^J characters.
	 */
	case KEYSYM_YANK:
		killbuff = killring.current;
		if (!killbuff)
			break;
		if (!TAILQ_FIRST(&killbuff->lines))
			break;
		if (TAILQ_NEXT(TAILQ_FIRST(&killbuff->lines), l_list))
			break;
		sidx += TAILQ_FIRST(&killbuff->lines)->length;
		return 0;
	case KEYSYM_RISEARCH:
	case KEYSYM_ISEARCH: {
		vdefault_t *vdef;

		if (sidx > 0 || (keysym == KEYSYM_ISEARCH) != sfwd) {
			saction = keysym == KEYSYM_ISEARCH ?
				ACT_RECISEARCH : ACT_RECRISEARCH;
			vminibuff_tmpexit(v);
			break;
		}

		/* insert the last search text */
		if (!last_text)
			break;
		line_nputs(minibuff_line, 0, last_text, last_len);
		sidx += last_len;
		vdef = view_findvdef((viewhdr_t *) v);
		vdef->redraw_text = 1;
		command_end(vdef);
		break;
	}

	default:
		if (keysym <= UCHAR_MAX && !input_isctrl(keysym)) {
			sidx++;
			return 0;
		} else
			screen_bell();
	}

	return 1;
}

/*
 * post input callback for searching; perform the search for the text
 * that we have so far
 */
static void searchpostcb(vminibuff_t *vmini, int keysym) {
	line_t *search_line;
	int search_col, search_linenum;
	int r;

	/* don't search when we're returning */
	if (saction != ACT_NONE)
		return;

	/*
	 * prepare for the new search by setting the cursor
	 * position to point we want to search from; if possible
	 * search from the last found position instead of searching
	 * from the beginning.
	 */
	if (searchfromsave) {
		search_line = save_line;
		search_linenum = save_linenum;
		search_col = save_col;
	} else {
		search_line = find_line;
		search_linenum = find_linenum;
		search_col = find_col;

		/*
		 * when searching backwards, go to the front of the
		 * currently found search term so that additions to
		 * it will continue to find it.
		 */
		if (!sfwd && !sfirst)
			search_col += find_idx;
	}

	/*
	 * if a search term exists, try to find it in the
	 * appropriate direction
	 */
	if (sidx) {
		if (sfwd)
			r = search_next(minibuff_line->text,
				minibuff_line->length, &search_line,
				&search_linenum, &search_col);
		else
			r = search_prev(minibuff_line->text,
				minibuff_line->length, &search_line,
				&search_linenum, &search_col);
	} else
		r = switched ? 0 : -1;

	/*
	 * check the success of our search; failure to find a term
	 * requires us to make some noise, finding it means we
	 * should set the find_ and view cursor positions.
	 */
	if (r == 0) {
		sview->cursor.linenum = search_linenum;
		sview->cursor.line = search_line;
		sview->cursor.col = search_col;
		find_linenum = search_linenum;
		find_line = search_line;
		find_col = search_col;
		find_idx = sidx;
		if (sfwd)
			mvfwd(find_idx, &sview->cursor.line,
				&sview->cursor.linenum, &sview->cursor.col);
	} else if (!sfailed)
		screen_bell();
	sfailed = r == -1;
	sfirst = 0;
	searchfromsave = 0;

	vdefault_setscreencol(sview);
	vdefault_checkscroll(sview);
	sview->redraw_header = 1;
}

/* recursive loop for incremental searching */
static int isearchloop() {
	line_t *prev_save_line, *prev_find_line, *prev_cursor_line;
	int prev_save_linenum, prev_save_col;
	int prev_failed, prev_previdx;
	int prev_fwd, prev_switched;
	int prev_cursor_linenum, prev_cursor_col;
	int prev_find_linenum, prev_find_col, prev_find_idx;

	/* save old data on the stack */
	prev_save_line = save_line;
	prev_save_linenum = save_linenum;
	prev_save_col = save_col;
	prev_failed = sfailed;

	/*
	 * save the starting cursor position and set the initial
	 * find position to it
	 */
	find_idx = sidx;
	find_line = save_line = sview->cursor.line;
	find_linenum = save_linenum = sview->cursor.linenum;
	find_col = save_col = sview->cursor.col;
	searchfromsave = 1;
	saction = ACT_NONE;

	/* perform the first search */
	sfirst = 1;
	searchpostcb(NULL, 0);
	sfirst = 0;
	if (sfailed && !switched) {
		saction = ACT_BACKUP;
		goto restore;
	}

	/* enter the minibuffer and do some searching */
search:
	minibuff_rawenter();

	switch (saction) {
	case ACT_BACKUP:
		goto restore;
	case ACT_ACCEPT:
	case ACT_CANCEL:
		return saction;
	case ACT_RECISEARCH:
	case ACT_RECRISEARCH:
		prev_fwd = sfwd;
		prev_switched = switched;
		prev_find_line = find_line;
		prev_find_linenum = find_linenum;
		prev_find_col = find_col;
		prev_find_idx = find_idx;
		prev_previdx = sprevidx;
		prev_cursor_line = sview->cursor.line;
		prev_cursor_linenum = sview->cursor.linenum;
		prev_cursor_col = sview->cursor.col;

		/* prepare new searching vars and recurse */
		sfwd = saction == ACT_RECISEARCH;
		switched = sfwd != prev_fwd;
		sprevidx = sidx;
		if (switched)
			minibuff_setprompt(sfwd ? "isearch" : "risearch");
		saction = isearchloop();
		if (saction == ACT_CANCEL || saction == ACT_ACCEPT)
			return saction;

		/* restore some saved vars */
		sprevidx = prev_previdx;
		sfwd = prev_fwd;
		if (switched)
			minibuff_setprompt(sfwd ? "isearch" : "risearch");
		switched = prev_switched;
		sfailed = prev_failed;

		/* grab the saved find off the stack */
		find_line = prev_find_line;
		find_linenum = prev_find_linenum;
		find_col = prev_find_col;
		find_idx = prev_find_idx;
		sview->cursor.line = prev_cursor_line;
		sview->cursor.linenum = prev_cursor_linenum;
		sview->cursor.col = prev_cursor_col;

		/* we moved the cursor, so set redraw flags */
		vdefault_setscreencol(sview);
		vdefault_checkscroll(sview);
		sview->redraw_header = 1;
		searchfromsave = 0;
		goto restore;
	case ACT_NONE:
	default:
		abort();
	}

restore:
	/* restore previous searching values */
	save_line = prev_save_line;
	save_linenum = prev_save_linenum;
	save_col = prev_save_col;
	sfailed = prev_failed;

	/* don't quit if we're coming out of a recursion */
	if (saction != ACT_BACKUP)
		goto search;

	return ACT_NONE;
}

/*
 * base for launching into the recursive isearch loop
 */
static void isearchbase(vdefault_t *v, int fwd) {
	line_t *start_line;
	int start_linenum, start_col;
	int ret;

	/* save in case of cancels */
	start_line = save_line = v->cursor.line;
	start_linenum = save_linenum = v->cursor.linenum;
	start_col = save_col = v->cursor.col;

	/* prepare the minibuffer */
	minibuff_rawinit(fwd ? "isearch" : "risearch", NULL,
		searchcback, searchpostcb);

	/* set up some variables */
	switched = 1;
	sfwd = fwd;
	sview = v;
	sprevidx = sidx = 0;
	sfailed = 0;

	/* keep searching until the user accepts or cancels */
	while ((ret = isearchloop()) == ACT_NONE)
		screen_bell();

	/*
	 * if the user canceled, move the cursor back to its old
	 * position; otherwise leave it where it is and push a
	 * mark where the cursor used to be, and save the search
	 * string for the next invocation.
	 */
	if (ret == ACT_CANCEL) {
		v->cursor.line = start_line;
		v->cursor.linenum = start_linenum;
		v->cursor.col = start_col;
		vdefault_setscreencol(v);
		vdefault_checkscroll(v);
		v->redraw_header = 1;
	} else {
		if (last_text)
			free(last_text);
		if (minibuff_line->length) {
			last_len = minibuff_line->length;
			last_text = ckmalloc(last_len);
			memcpy(last_text, minibuff_line->text, last_len);
		} else
			last_text = NULL;

		buffer_pushmark(v->hdr.buffer, start_line, start_linenum,
			start_col);
		v->hdr.buffer->mark_implicit = 1;
		minibuff_printf("Mark saved where search started.");
	}
}

/*
 * try to find text incrementally in the forward direction as the
 * user types the characters.
 */
void command_isearch(vdefault_t *v) {
	isearchbase(v, 1);
}

/* same but start backwards */
void command_risearch(vdefault_t *v) {
	isearchbase(v, 0);
}

/* prompt the user for line number and go to it */
void command_gotoln(vdefault_t *v) {
	char *str;
	int i;

	/* get the line number from the user */
	str = minibuff_prompt("Line Number", NULL);
	if (!str)
		return;
	i = atoi(str) - 1;
	free(str);
	if (i < 0)
		return;

	/* set the cursor line to the typed line num */
	v->cursor.linenum = 0;
	v->cursor.line = buffer_lineat(v->hdr.buffer, i, &v->cursor.linenum);
	v->view.col = v->cursor.col = 0;
	v->view.line = v->cursor.line;

	/* put the view back half the view height */
	v->view.linenum = v->cursor.linenum;
	i = v->hdr.geom.height / 2;
	while (i-- && TAILQ_PREV(v->view.line, linelist, l_list)) {
		v->view.line = TAILQ_PREV(v->view.line, linelist, l_list);
		v->view.linenum--;
	}

	vdefault_setscreencol(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/* find and replace, prompting the user on each find */
void command_replace(vdefault_t *v) {
	line_t *repl_line;
	u_char *old, *new;
	int repl_linenum, repl_col;
	int count, oldlen, newlen, key;
	int all;

	old = minibuff_prompt("Replace", NULL);
	if (!old)
		return;
	new = minibuff_prompt("Replace with", NULL);
	if (!new)
		goto free;

	v->rm_protect = PROTECT_CURSOR;
	undo_beginblock(v->hdr.buffer);
	repl_line = v->cursor.line;
	repl_linenum = v->cursor.linenum;
	repl_col = v->cursor.col;
	buffer_pushmark(v->hdr.buffer, v->cursor.line, v->cursor.linenum,
		v->cursor.col);
	v->hdr.buffer->mark_implicit = 1;

	/*
	 * find each occurance of the search term, prompt the user
	 * for what to do about each one
	 */
	all = 0;
	oldlen = strlen(old);
	newlen = strlen(new);
	for (count = 0;; count++) {
		if (search_next(old, oldlen, &repl_line, &repl_linenum,
				&repl_col))
			break;
		if (all)
			goto replace;

		v->cursor.line = repl_line;
		v->cursor.linenum = repl_linenum;
		v->cursor.col = repl_col;
		mvfwd(oldlen, &v->cursor.line, &v->cursor.linenum,
			&v->cursor.col);
		vdefault_setscreencol(v);
		vdefault_checkscroll(v);
		key = minibuff_chprompt("Replace this instance", "ynaq", 1);

		switch (key) {
		case KEYSYM_CANCEL:
		case 'q':
			goto done;
		case 'a':
			all = 1;
			break;
		case 'y':
replace:
			/*
			 * XXX: if the search or repl terms contain ^Js this
			 * will break; a better implementation is neccesary.
			 */
			buffer_rmstr(v->hdr.buffer, 1, repl_line,
				repl_linenum, repl_col, oldlen);
			buffer_nputs(v->hdr.buffer, 1, repl_line,
				repl_linenum, repl_col, new, newlen);
			mvfwd(newlen, &repl_line, &repl_linenum, &repl_col);
			break;
		case 'n':
			mvfwd(oldlen, &repl_line, &repl_linenum, &repl_col);
			break;
		}
	}
done:
	v->cursor.line = repl_line;
	v->cursor.linenum = repl_linenum;
	v->cursor.col = repl_col;
	minibuff_printf("Replaced %d occurances.", count);
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);

	undo_endblock(v->hdr.buffer);

	free(new);
free:
	free(old);
}
