/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/*
 * transposition commands; tranchars, tranwords, and tranlines.
 */

/* transpose characters around the cursor */
void command_tranchars(vdefault_t *v) {
	line_t *prev;
	int i;
	u_char c;

	/* prevent changes to cursor and view on removes */
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/* handle cursor advancement */
	if (v->cursor.col == 0)
		i = v->cursor.col = 0;
	else if (v->cursor.col < v->cursor.line->length)
		i = v->cursor.col++;
	else
		i = v->cursor.col - 1;

	/*
	 * perform transposition depending on old position; if the cursor
	 * was at the end of the line we swap the two characters in frot
	 * of the cursor, otherwise swap the two characters around the
	 * cursor.  in column 0 swap the char across lines.
	 */
	if (!i) {
		prev = TAILQ_PREV(v->cursor.line, linelist, l_list);
		if (!prev)
			return;
		if (v->cursor.line->length) {
			undo_beginblock(v->hdr.buffer);
			buffer_putc(v->hdr.buffer, 1, prev,
				v->cursor.linenum - 1, prev->length,
				v->cursor.line->text[0]);
			buffer_rmch(v->hdr.buffer, 1, v->cursor.line,
				v->cursor.linenum, 0);
			undo_endblock(v->hdr.buffer);

			v->cursor.col = 0;
		} else {
			if (!prev->length)
				return;
			undo_beginblock(v->hdr.buffer);
			buffer_putc(v->hdr.buffer, 1, v->cursor.line,
				v->cursor.linenum, 0,
				prev->text[prev->length - 1]);
			buffer_rmch(v->hdr.buffer, 1, prev,
				v->cursor.linenum - 1, prev->length - 1);
			undo_endblock(v->hdr.buffer);

			v->cursor.col++;
		}
	} else {
		undo_beginblock(v->hdr.buffer);
		c = v->cursor.line->text[i];
		buffer_rmch(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, i);
		buffer_putc(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, i - 1, c);
		undo_endblock(v->hdr.buffer);
	}

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/* transpose words */
void command_tranwords(vdefault_t *v) {
	u_char *w1, *w2;
	int w1start, w1end, w2start, w2end;
	int w1len, w2len;

	/* prevent changes to cursor and view on removes */
	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/* find word boundaries */
	if (v->cursor.col == v->cursor.line->length)
		return;
	w2end = word_next(v->cursor.line->text, v->cursor.line->length,
		v->cursor.col);
	if (w2end <= 1)
		return;
	w2start = word_prev(v->cursor.line->text, w2end);
	if (w2start <= 1)
		return;
	w1start = word_prev(v->cursor.line->text, w2start);
	if (w1start == v->cursor.line->length)
		return;
	w1end = word_next(v->cursor.line->text, v->cursor.line->length,
		w1start);

	if (w1start == w2start || w2start < w1end) {
		screen_bell();
		return;
	}
	w1len = w1end - w1start;
	w2len = w2end - w2start;

	/* get a copy of each word */
	w1 = ckmalloc(w1len);
	w2 = ckmalloc(w2len);
	memcpy(w1, v->cursor.line->text + w1start, w1len);
	memcpy(w2, v->cursor.line->text + w2start, w2len);

	/* swap the two words in the buffer */
	undo_beginblock(v->hdr.buffer);
	buffer_rmstr(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		w1start, w1len);
	w2start -= w1len;
	buffer_rmstr(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		w2start, w2len);
	buffer_nputs(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		w1start, w2, w2len);
	w2start += w2len;
	buffer_nputs(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		w2start, w1, w1len);
	undo_endblock(v->hdr.buffer);

	/* move the cursor after the first word (moved into w2start) */
	v->cursor.col = w2start + w1len;

	free(w1);
	free(w2);
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/* transpose the current line with the one before it */
void command_tranlines(vdefault_t *v) {
	line_t *line;
	int linenum;
	int curlinelen;

	v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;

	/* find the line to transpose with */
	line = TAILQ_PREV(v->cursor.line, linelist, l_list);
	if (!line) {
		line = TAILQ_NEXT(v->cursor.line, l_list);
		if (!line) {
			minibuff_printf("Don't have two lines to transpose.");
			return;
		} else
			linenum = v->cursor.linenum + 1;
	} else
		linenum = v->cursor.linenum - 1;

	/* swap the contents of the two lines */
	curlinelen = v->cursor.line->length;
	undo_beginblock(v->hdr.buffer);
	buffer_nputs(v->hdr.buffer, 1, line, linenum, 0,
		v->cursor.line->text, curlinelen);
	buffer_rmstr(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		0, curlinelen);
	buffer_nputs(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
		0, &line->text[curlinelen], line->length - curlinelen);
	buffer_rmstr(v->hdr.buffer, 1, line, linenum, curlinelen,
		line->length - curlinelen);
	undo_endblock(v->hdr.buffer);

	/* place the cursor after the second line */
	if (linenum > v->cursor.linenum) {
		v->cursor.line = line;
		v->cursor.linenum = linenum;
	}
	if (TAILQ_NEXT(v->cursor.line, l_list)) {
		v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
		v->cursor.linenum++;
	}
	v->cursor.col = 0;

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}
