/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/*
 * basic typing commands; insertion of text and newlines.
 */

static void do_addnl(vdefault_t *v) {
	buffer_addline(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum);

	/* move any excess text to the new line */
	if (v->cursor.col < v->cursor.line->length) {
		buffer_nputs(v->hdr.buffer, 1,
			TAILQ_NEXT(v->cursor.line, l_list),
			v->cursor.linenum + 1, 0,
			&v->cursor.line->text[v->cursor.col],
			v->cursor.line->length - v->cursor.col);
		v->rm_protect = PROTECT_CURSOR | PROTECT_VIEW;
		buffer_rmstr(v->hdr.buffer, 1, v->cursor.line,
			v->cursor.linenum, v->cursor.col,
			v->cursor.line->length - v->cursor.col);
	}
}

void command_enter(vdefault_t *v) {
	line_t *prev;
	int i;

	undo_beginblock(v->hdr.buffer);
	do_addnl(v);
	prev = v->cursor.line;
	v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
	v->cursor.linenum++;
	v->cursor.col = 0;

	/*
	 * maintain the same indententation level as the last line
	 * the auto indent option is set; also if the previous line
	 * contains only whitespace, truncate it to 0.
	 */
	if (options.autoindent) {
		i = 0;
		while (i < prev->length) {
			if (prev->text[i] != ' ' && prev->text[i] != '\t')
				break;
			i++;
		}
		if (i) {
			buffer_nputs(v->hdr.buffer, 1, v->cursor.line,
				v->cursor.linenum, 0, prev->text, i);
			v->cursor.col += i;
			if (i == prev->length)
				buffer_rmstr(v->hdr.buffer, 1, prev,
					v->cursor.linenum - 1, 0, i);
		}
	}

	undo_endblock(v->hdr.buffer);

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_openline(vdefault_t *v) {
	undo_beginblock(v->hdr.buffer);
	do_addnl(v);
	undo_endblock(v->hdr.buffer);

	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

/* insert text input into a buffer */
void command_type(vdefault_t *v, int key) {
	line_t *save_line;
	int save_screencol, save_col, save_linenum;
	int save_lastcol;
	int parennest;
	u_char ch;

	/* add the character they typed */
	buffer_putc(v->hdr.buffer, 1, v->cursor.line, v->cursor.linenum,
			v->cursor.col, key);

	/*
	 * perform paren flashing if the option is set and there
	 * are no keys pending; otherwise just put the cursor in
	 * its new position and we're done.
	 */
	if (options.parenflash && !input_nodraw && !v->inhibit_parenflash &&
			(key == ')' || key == '}') && !input_key_pending(0)) {
		save_line = v->cursor.line;
		save_screencol = v->cursor.screencol +
			VIEWCALLI(v->hdr.top, viswid, v->cursor.screencol,
			v->cursor.line->text[v->cursor.col]);
		save_col = ++v->cursor.col;
		save_linenum = v->cursor.linenum;
		save_lastcol = v->cursor.screencol;

		/* find the matching { or ( */
		parennest = 0;
		for (;;) {
			while (v->cursor.col == 0) {
				v->cursor.line = TAILQ_PREV(v->cursor.line,
					linelist, l_list);
				if (!v->cursor.line)
					goto nomatch;
				v->cursor.col = v->cursor.line->length;
				v->cursor.linenum--;
				if (v->cursor.linenum < v->view.linenum)
					goto nomatch;
			}

			/* check this position */
			v->cursor.col--;
			ch = v->cursor.line->text[v->cursor.col];
			if (ch != key) {
				if (key == ')' && ch == '(')
					parennest--;
				else if (key == '}' && ch == '{')
					parennest--;
			} else
				parennest++;

			/* if nest is 0 we found the matching one */
			if (parennest == 0) {
				vdefault_setscreencol(v);
				if (v->cursor.screencol < v->view.col)
					goto nomatch;
				draw_screen();
				screen_refresh();
				screen_key_pending(FLASH_DURATION);
				break;
			}
		}

nomatch:
		v->cursor.line = save_line;
		v->cursor.col = save_col;
		v->cursor.screencol = save_screencol;
		v->cursor.linenum = save_linenum;
	} else
		v->cursor.screencol += VIEWCALLI(v->hdr.top, viswid,
			v->cursor.screencol,
			v->cursor.line->text[v->cursor.col++]);

	v->cursor.lastcol = v->cursor.screencol;

	vdefault_checkscroll(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}
