/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/*
 * commands for moving the view of the buffer (most of these will affect
 * the cursor position as well).
 */

static void viewmvup(vdefault_t *v, int amt) {
	while (amt--) {
		if (!TAILQ_PREV(v->cursor.line, linelist, l_list))
			break;
		if (TAILQ_PREV(v->view.line, linelist, l_list)) {
			v->view.line = TAILQ_PREV(v->view.line, linelist,
				l_list);
			v->view.linenum--;
		}
		v->cursor.line = TAILQ_PREV(v->cursor.line, linelist, l_list);
		v->cursor.linenum--;
	}
}

static void viewmvdn(vdefault_t *v, int amt) {
	int linediff, i;

	linediff = v->cursor.linenum - v->view.linenum;

	i = amt;
	while (i--) {
		if (!TAILQ_NEXT(v->cursor.line, l_list))
			break;
		v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
		v->cursor.linenum++;
	}

	/* travel back from the new cursor to maintain view linediff */
	if (v->cursor.linenum - v->view.linenum < amt)
		return;
	v->view.line = v->cursor.line;
	v->view.linenum = v->cursor.linenum;
	while (linediff--) {
		if (!TAILQ_PREV(v->view.line, linelist, l_list))
			break;
		v->view.line = TAILQ_PREV(v->view.line, linelist, l_list);
		v->view.linenum--;
	}
}

void command_pgup(vdefault_t *v) {
	viewmvup(v, v->hdr.geom.height);

	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_pgdn(vdefault_t *v) {
	viewmvdn(v, v->hdr.geom.height);

	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_halfpgup(vdefault_t *v) {
	viewmvup(v, v->hdr.geom.height / 2);

	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_halfpgdn(vdefault_t *v) {
	viewmvdn(v, v->hdr.geom.height / 2);

	vdefault_trylastcol(v);
	vdefault_checkscroll(v);
	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_lineup(vdefault_t *v) {
	if (!TAILQ_PREV(v->view.line, linelist, l_list)) {
		screen_bell();
		return;
	}

	v->view.line = TAILQ_PREV(v->view.line, linelist, l_list);
	v->view.linenum--;
	if (v->cursor.linenum - v->view.linenum > v->hdr.geom.height - 1) {
		v->cursor.line = TAILQ_PREV(v->cursor.line, linelist, l_list);
		v->cursor.linenum--;

		vdefault_trylastcol(v);
		vdefault_checkscroll(v);
	}

	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_linedn(vdefault_t *v) {
	if (!TAILQ_NEXT(v->view.line, l_list)) {
		screen_bell();
		return;
	}

	v->view.line = TAILQ_NEXT(v->view.line, l_list);
	v->view.linenum++;
	if (v->cursor.linenum - v->view.linenum < 0) {
		v->cursor.line = v->view.line;
		v->cursor.linenum = v->view.linenum;

		vdefault_trylastcol(v);
		vdefault_checkscroll(v);
	}

	v->redraw_text = 1;
	v->redraw_header = 1;
}

void command_firstline(vdefault_t *v) {
	buffer_pushmark(v->hdr.buffer, v->cursor.line,
		v->cursor.linenum, v->cursor.col);
	v->hdr.buffer->mark_implicit = 1;
	minibuff_printf("Mark set.");
	v->cursor.line = TAILQ_FIRST(&v->hdr.buffer->lines);
	v->cursor.linenum = v->cursor.col = 0;
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_lastline(vdefault_t *v) {
	buffer_pushmark(v->hdr.buffer, v->cursor.line,
		v->cursor.linenum, v->cursor.col);
	v->hdr.buffer->mark_implicit = 1;
	minibuff_printf("Mark set.");
	v->cursor.line = TAILQ_LAST(&v->hdr.buffer->lines, linelist);
	v->cursor.linenum = v->hdr.buffer->linecnt - 1;
	v->cursor.col = 0;
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
}

void command_center(vdefault_t *v) {
	int i;

	v->view.line = v->cursor.line;
	v->view.linenum = v->cursor.linenum;
	i = v->cursor.linenum - screen_height / 2;
	if (i < 0)
		i = 0;
	while (v->view.linenum > i) {
		v->view.line = TAILQ_PREV(v->view.line, linelist, l_list);
		v->view.linenum--;
	}
	v->redraw_text = 1;
	v->redraw_header = 1;
}
