/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* map color names to color identifiers */
struct color {
	LIST_ENTRY(color) c_list;

	u_char	*name;
	int	ident;
};
static LIST_HEAD(, color) *color_hash;
#define COLOR_BUCKETS	12

/* add an entry in the color map */
static __inline void addclrmap(u_char *name, int ident) {
	struct color *c;
	int idx;

	c = ckmalloc(sizeof(struct color));
	c->name = ckstrdup(name);
	c->ident = ident;
	idx = hash_string(COLOR_BUCKETS, name);
	LIST_INSERT_HEAD(&color_hash[idx], c, c_list);
}

/*
 * define the standard colors
 */
void color_init() {
	color_hash = hash_alloc("color_hash", COLOR_BUCKETS,
		offsetof(struct color, c_list));

	/* set up the default colors */
	screen_defcolor(COLOR_DEFAULT, "default", "default",
		SCREEN_A_NORM, SCREEN_A_NORM);
	screen_defcolor(COLOR_HEADING, "white", "blue",
		SCREEN_A_NORM, SCREEN_A_STAND);
	screen_defcolor(COLOR_MARKEDTEXT, "white", "magenta",
		SCREEN_A_NORM, SCREEN_A_STAND);
	screen_defcolor(COLOR_SPECIAL, "brown", "default",
		SCREEN_A_NORM, SCREEN_A_NORM);

	/* place them into the color map */
	addclrmap("default", COLOR_DEFAULT);
	addclrmap("heading", COLOR_HEADING);
	addclrmap("markedtext", COLOR_MARKEDTEXT);
	addclrmap("special", COLOR_SPECIAL);
}

void color_shutdown() {
	struct color *c, *n;
	int idx;

	for (idx = 0; idx < COLOR_BUCKETS; idx++) {
		c = LIST_FIRST(&color_hash[idx]);
		while (c) {
			n = LIST_NEXT(c, c_list);
			free(c->name);
			free(c);
			c = n;
		}
	}
	hash_free(color_hash);
}

int color_lookup(u_char *name) {
	struct color *c;
	int idx;

	idx = hash_string(COLOR_BUCKETS, name);
	LIST_FOREACH(c, &color_hash[idx], c_list)
		if (!strcmp(c->name, name))
			return c->ident;
	return COLOR_NULL;
}

int color_add(u_char *name, u_char *fg, u_char *bg, int attr, int bwattr) {
	static int next = COLOR_LAST;
	int ident;

	if (name) {
		if ((ident = color_lookup(name)) == COLOR_NULL) {
			ident = next++;
			addclrmap(name, ident);
		}
	} else
		ident = next++;
	screen_defcolor(ident, fg, bg, attr, bwattr);

	return ident;
}

static __inline int mkattr(u_char *c) {
	int attr;

	attr = 0;
	for (;;)
		switch (tolower(*c++)) {
		case '\0':
			return attr;
		case 's':
			attr |= SCREEN_A_STAND;
			break;
		case 'u':
			attr |= SCREEN_A_UL;
			break;
		case 'r':
			attr |= SCREEN_A_RV;
			break;
		case 'l':
			attr |= SCREEN_A_BLINK;
			break;
		case 'd':
			attr |= SCREEN_A_DIM;
			break;
		case 'b':
			attr |= SCREEN_A_BOLD;
			break;
		case 'n':
		default:
			break;
		}
}

/*
 * color configuration file is extreemly simple.  records are
 * seperated by whitespace with each statement ending on a ';'.
 * any #.*\n is considered a comment.
 *
 * the file has colors in the format
 *
 *		name fg bg attr bwattr ';'
 *
 * which comprise the arguments to pass to color_add.
 */
static __inline void doparse(FILE *fp) {
	u_char items[5][256];
	int attr, bwattr;
	int loc, idx, c;

	idx = loc = 0;
	for (;;) {
		if ((c = getc(fp)) == EOF)
			break;
		if (c == ' ' || c == '\t' || c == '\n') {
			if (idx)
				idx = 0, loc++;
			continue;
		}
		if (c == '#') {
			if (idx)
				idx = 0, loc++;
			while ((c = getc(fp)) != '\n')
				if (c == EOF)
					goto out;
			continue;
		}
		if (c == ';') {
			if (idx)
				loc++;
			if (loc == 5) {
				attr = mkattr(items[3]);
				bwattr = mkattr(items[4]);
				color_add(items[0], items[1], items[2],
					attr, bwattr);
			}
			idx = loc = 0;
			continue;
		}

		/* add character to the current item */
		if (loc > 4)
			continue;
		items[loc][idx++] = c;
		items[loc][idx] = '\0';
		if (idx + 1 >= sizeof(items[0])) {
			warnx("item overflow in doparse");
			idx = 0, loc++;
		}
	}
out:
}

/* read in color configuration file */
void color_parse() {
	char fullpath[PATH_MAX];
	char *colorpath, *tmp;
	char *path, *expath;
	FILE *fp;

	/* make a local copy for strsep */
	colorpath = ckstrdup(COLOR_PATH);
	tmp = colorpath;

	/* loop through the path */
	fp = NULL;
	while ((path = strsep(&tmp, ":")) != NULL && !fp) {
		expath = tilde_expand(path);

		/* build the full path */
		if (snprintf(fullpath, sizeof(fullpath), "%s/color",
				expath) >= sizeof(fullpath)) {
			warnx("path to color definition exceeds PATH_MAX");
			goto next;
		}

		fp = fopen(fullpath, "r");
next:		free(expath);
	}
	free(colorpath);
	if (!fp)
		return;

	doparse(fp);
	fclose(fp);
}
