/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* completion routine for filenames */
u_char **complete_filename(vminibuff_t *v, int *startidx) {
	u_char path[PATH_MAX];
	struct stat sb;
	u_char **comps;
	struct dirent *de;
	u_char *sofar, *exsofar;
	line_t *line;
	DIR *dir;
	int i, n, compsize;

	input_begincancel();
	line = TAILQ_FIRST(&v->hdr.buffer->lines);

	/* find the start of the part of a path */
	if (line->length) {
		for (*startidx = line->length; *startidx > 0; (*startidx)--)
			if (line->text[*startidx - 1] == '/')
				break;
	} else
		*startidx = 0;

	/* get the amount we know so far */
	if (*startidx) {
		sofar = ckmalloc(*startidx + 1);
		memcpy(sofar, line->text, *startidx);
		sofar[*startidx] = '\0';
	} else
		sofar = ckstrdup(".");
	exsofar = tilde_expand(sofar);

	/* start the possible completions array */
	compsize = 10;
	comps = ckmalloc(compsize * sizeof(u_char *));

	/* enumerate all the files in the directory */
	i = 0;
	dir = opendir(exsofar);
	if (!dir)
		goto none;
	while ((de = readdir(dir)) != NULL) {
		for (n = *startidx; n < line->length; n++) {
			if (line->text[n] != de->d_name[n - *startidx])
				goto next;
			if (!de->d_name[n - *startidx])
				break;
		}

		/* for directories, append a '/' */
		if (snprintf(path, sizeof(path), "%s/%s", exsofar,
				de->d_name) > sizeof(path))
			continue;
		if (stat(path, &sb))
			continue;
		if (sb.st_mode & S_IFDIR) {
			comps[i] = ckmalloc(strlen(de->d_name) + 2);
			strcpy(comps[i], de->d_name);
			strcat(comps[i++], "/");
		} else
			comps[i++] = ckstrdup(de->d_name);

		/* resize the array if neccesary */
		if (i >= compsize - 1) {
			compsize += 10;
			comps = ckrealloc(comps, compsize * sizeof(u_char *));
		}
next:
		if (input_checkcancel())
			break;
	}
	closedir(dir);

none:
	input_endcancel();

	free(sofar);
	free(exsofar);
	comps[i] = NULL;
	return comps;
}

/*
 * complete by getting names of items from a hash, offset of the
 * list entry and name must be provided.
 */
u_char **complete_fromhash(vminibuff_t *v, int *startidx, void *hash,
		int buckets, int off_name, int off_list) {
	struct dummy {};
	LIST_HEAD(, dummy) *hashhead;
	char *entry, *next;
	u_char **comps, *name;
	line_t *line;
	int compsize, i, n, idx;

	hashhead = hash;
	*startidx = 0;
	compsize = 10;
	comps = ckmalloc(compsize * sizeof(u_char *));
	line = TAILQ_FIRST(&v->hdr.buffer->lines);

	/* list everything in the hashtable */
	for (i = 0, idx = 0; idx < buckets; idx++)
		for (entry = (void *) LIST_FIRST(&hashhead[idx]);
				entry != NULL;) {
			name = *((u_char **) (entry + off_name));
			next = *((char **) (entry + off_list));

			for (n = 0; n < line->length; n++) {
				if (line->text[n] != name[n])
					goto next;
				if (!name[n])
					break;
			}
			comps[i++] = ckstrdup(name);
			if (i >= compsize - 1) {
				compsize *= 2;
				comps = ckrealloc(comps, compsize *
					sizeof(u_char *));
			}
next:
			entry = next;
		}
	comps[i] = NULL;

	return comps;
}

u_char **complete_buff(vminibuff_t *v, int *startidx) {
	u_char **comps;
	buffer_t *buffer;
	line_t *line;
	int compsize, i, n;

	i = 0;
	*startidx = 0;
	compsize = 10;
	comps = ckmalloc(compsize * sizeof(u_char *));
	line = TAILQ_FIRST(&v->hdr.buffer->lines);

	TAILQ_FOREACH(buffer, &buffer_list, b_list) {
		if (buffer->internal)
			continue;
		for (n = 0; n < line->length; n++) {
			if (line->text[n] != buffer->dispname[n])
				goto next;
			if (!buffer->dispname[n])
				break;
		}
		comps[i++] = ckstrdup(buffer->dispname);
		if (i >= compsize - 1) {
			compsize *= 2;
			comps = ckrealloc(comps, compsize *
				sizeof(u_char *));
		}
next:
	}
	comps[i] = NULL;

	return comps;
}
