/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* max chars a visified string can be */
#define MAX_VISCTRL	5

/* draw everything */
void draw_screen() {
	if (minibuff_active) {
		frame_draw(frame_active);
		minibuff_draw();
	} else {
		minibuff_draw();
		frame_draw(frame_active);
	}
}

/* draw view header for a vdefault */
void draw_header(vdefault_t *v, int x, int y, int width) {
	int i, len, len2;
	int pct, tot;

	/* we assume width is at least 2 for the buffer name part */
	if (width < 2)
		return;

	/* print the cursor position */
	screen_setcolor(COLOR_HEADING);
	screen_gotoxy(x, y);
	len = screen_nprintf(width, " L%d C%d",
		v->cursor.linenum + 1, v->cursor.screencol + 1);

	/* right align the buffer name */
	len2 = strlen(v->hdr.buffer->dispname) + 1;
	if (v->hdr.buffer->modified)
		len2++;
	len2 += v->hdr.buffer->nlstyle == NLSTYLE_LF ? 5 : 7;
	len2 += 5;		/* %2d%% for the line pct */
	len2 = min(len2, width);

	/* put spaces out for the rest of the width */
	screen_gotoxy(x + len, y);
	for (i = len; i < width - len2; i++)
		screen_putc(' ');
	if (i > width - len2) {
		for (; i < width; i++)
			screen_putc(' ');
		return;
	}

	/* get the percent of the buffer above the cursor line */
	tot = v->hdr.buffer->linecnt - 1;
	if (tot > 0)
		pct = ((double) v->cursor.linenum / (double) tot) * 100;
	else
		pct = 0;

	/* print the buffer name and some other info */
	screen_gotoxy(x + width - len2, y);
	if (v->hdr.buffer->modified) {
		screen_nputs(v->hdr.buffer->dispname, len2 - 1);
		screen_putc('*');
	} else
		screen_nputs(v->hdr.buffer->dispname, len2);
	screen_puts(v->hdr.buffer->nlstyle == NLSTYLE_LF ? " <lf>" :
		" <crlf>");
	screen_nprintf(5, " %3d%%", pct);
	screen_putc(' ');
}

/* make a visible representation of a control character */
void draw_visify(u_char *vis, u_char c) {
	/*
	 * we do quoting in the same way that vis(3) does it,
	 * but without the leading '\'.
	 */
	if (c < 32) {
		vis[0] = '^';
		vis[1] = '@' + c;
		vis[2] = '\0';
	} else if (c == '\177') {
		vis[0] = '^';
		vis[1] = '?';
		vis[2] = '\0';
	} else
		/*
		 * for anything else just do a represenation of if
		 * as a 2-character hex string with a \x prefix.
		 */
		snprintf(vis, MAX_VISCTRL, "\\x%2x", c);
}

/*
 * output a string of text on the screen in the current color; clip by
 * the scrollcol and don't output more than max screen columns.  the
 * base field is the screen column used as the base for tabstop
 * computation.  the viewcol is used to determine the screen column to
 * start printing at, and to return the number of columns printed.
 */
void draw_string(u_char *text, int max, int len, int scrollcol,
		int *viewcol) {
	u_char vis[MAX_VISCTRL];
	int i, n, tabstop;

	/*
	 * output all the characters in text, but no more than
	 * max.  if tabs or nonvisible characters are found,
	 * modify them to a visible representation and print
	 * that.
	 */
	for (i = 0; i < len; i++) {
		if (text[i] == '\t') {
			tabstop = draw_next_tabstop(*viewcol);
			for (n = 0; n < tabstop; n++) {
				if (*viewcol + 1 - scrollcol >= max)
					goto maxed;
				if (*viewcol >= scrollcol)
					screen_putc(' ');
				(*viewcol)++;
			}
		} else if (input_isctrl(text[i])) {
			draw_visify(vis, text[i]);
			for (n = 0; vis[n] != '\0'; n++) {
				if (*viewcol + 1 - scrollcol >= max)
					goto maxed;
				if (*viewcol >= scrollcol)
					screen_putc(vis[n]);
				(*viewcol)++;
			}
		} else {
			if (*viewcol + 1 - scrollcol >= max)
				goto maxed;
			if (*viewcol >= scrollcol)
				screen_putc(text[i]);
			(*viewcol)++;
		}
	}
	return;

	/*
	 * overflowed the maximum characters, so to let them know push
	 * viewcol over the max value before returning.
	 */
maxed:
	(*viewcol)++;
}

/*
 * put a '$' on the end of long lines, and for other lines,
 * clear with spaces to the end of the line.  spaces are drawn in
 * the currently active color.
 */
void draw_eol(vdefault_t *v, int viewcol, int y) {
	if (viewcol - v->view.col > v->hdr.geom.width - 1) {
		screen_gotoxy(v->hdr.geom.x + v->hdr.geom.width - 1,
			v->hdr.geom.y + y);
		screen_setcolor(COLOR_SPECIAL);
		screen_putc('$');
	} else {
		for (; viewcol - v->view.col < v->hdr.geom.width; viewcol++)
			if (viewcol >= v->view.col)
				screen_putc(' ');
	}
}

/*
 * draw the empty line character for any additional
 * lines that we are displaying, but have no buffer
 * data to show in.
 */
void draw_emptylines(viewhdr_t *v, int y) {
	int i;

	screen_setcolor(COLOR_SPECIAL);
	for (; y < v->geom.height; y++) {
		screen_gotoxy(v->geom.x, v->geom.y + y);
		screen_putc(options.blanklinechar);
		for (i = 1; i < v->geom.width; i++)
			screen_putc(' ');
	}
}

static __inline void draw_marked(vdefault_t *v, region_t *r,
		line_t **line, int *linenum, int *col, int *y, int *viewcol) {
	int len, done;

	/* don't draw if past it */
	if (*linenum > r->last.linenum || (*linenum == r->last.linenum &&
			*col >= r->last.col))
		return;

	/* draw the marked region in the appropriate color */
	done = 0;
	screen_setcolor(COLOR_MARKEDTEXT);
	while (*y < v->hdr.geom.height) {
		if (*line == r->last.line) {
			len = r->last.col - *col;
			if (*col + len >= r->last.col)
				done = 1;
		} else
			len = (*line)->length - *col;
		draw_setxy(v, *viewcol, *y);
		if (len) {
			draw_string(&(*line)->text[*col], v->hdr.geom.width,
				len, v->view.col, viewcol);
			*col += len;
		}

		if (*col >= (*line)->length) {
			if (done)
				screen_setcolor(COLOR_DEFAULT);
			draw_eol(v, *viewcol, *y);
			screen_setcolor(COLOR_MARKEDTEXT);
			*line = TAILQ_NEXT(*line, l_list);
			(*col) = (*viewcol) = 0;
			(*linenum)++;
			(*y)++;
		}
		if (done)
			break;
	}
}

void draw_text(viewhdr_t *v, vdefault_t *vdef) {
	region_t r;
	line_t *line;
	int linenum, col, viewcol;
	int y, havemark;

	line = vdef->view.line;
	linenum = vdef->view.linenum;
	y = col = viewcol = 0;


	/* if there's a mark we need to draw it first */
	if (!v->buffer->mark_implicit && v->buffer->mark) {
		havemark = 1;
		vdefault_getmarkrgn(vdef, &r);
	} else
		havemark = 0;

	/*
	 * draw as many lines in the buffer are visible in the
	 * portion of the buffer that the view can show, call out
	 * to the drawpart viewop to allow the view to control how
	 * exactly it looks.
	 */
	while (y < v->geom.height) {
		if (havemark && linenum >= r.first.linenum &&
				linenum <= r.last.linenum) {
			while (line == r.first.line && col < r.first.col)
				VIEWCALL(v, drawpart, &line, &linenum, &col,
					r.first.col - col, &y, &viewcol);
			draw_marked(vdef, &r, &line, &linenum, &col,
				&y, &viewcol);
			havemark = 0;
			if (!line)
				break;
			continue;
		}
		VIEWCALL(v, drawpart, &line, &linenum, &col,
			line->length - col, &y, &viewcol);

		/* move to the next line */
		if (col >= line->length) {
			screen_setcolor(COLOR_DEFAULT);
			draw_eol(vdef, viewcol, y);
			line = TAILQ_NEXT(line, l_list);
			viewcol = col = 0;
			linenum++, y++;
		}
		if (!line)
			break;
	}

	draw_emptylines((viewhdr_t *) v, y);
}

void draw_standard(viewhdr_t *v, vdefault_t *vdef) {
	if (v->buffer->dirty || vdef->redraw_text ||
			(v->buffer->mark &&
			!v->buffer->mark_implicit &&
			vdef->redraw_header)) {
		draw_text(v, vdef);
		vdef->redraw_text = 0;
	}
	vdef->redraw_header = 0;
}
