/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* open a file and create a new buffer out of it */
buffer_t *file_read(u_char *fn) {
	buffer_t *buffer;
	line_t *line;
	FILE *fp;
	int linenum, col;

	/* attempt to open the file */
	fp = fopen(fn, "rb");
	if (fp == NULL)
		return NULL;
	buffer = buffer_create(fn, 0);

	/* read the file into the buffer */
	buffer->nlstyle = NLSTYLE_UNKNOWN;
	line = TAILQ_FIRST(&buffer->lines);
	linenum = 0;
	col = 0;
	if (file_fread(buffer, 0, 0, &line, &linenum, &col, fp)) {
		buffer_destroy(buffer);
		buffer = NULL;
		goto exit;
	}
	buffer->modified = 0;

exit:
	fclose(fp);
	return buffer;
}

/* open a file and write the buffer out to it */
int file_write(buffer_t *buffer, u_char *fn) {
	int ret = 0;
	FILE *fp;

	fp = fopen(fn, "wb");
	if (!fp)
		return -1;
	if (file_fwrite(buffer, fp))
		ret = -1;
	fclose(fp);

	return ret;
}

/*
 * read from a FILE and add them to line, creating new
 * lines and attaching them as neccesary; returns a nonzero if an
 * error occured while reading
 */
int file_fread(buffer_t *buffer, int undoable, int cancelable, line_t **line,
		int *linenum, int *col, FILE *fp) {
	u_char linetxt[4096];
	line_t *next;
	int txtidx, c;

	if (undoable)
		undo_beginblock(buffer);

	/*
	 * read a character at a time from the file, placing the
	 * non newlines into linetxt[].  When we get newlines or
	 * when the linetxt is full it is flushed with out to
	 * the line.
	 */
	txtidx = 0;
	for (;;) {
		c = fgetc(fp);

		/* check cancellation, if supposed to */
		if (cancelable && input_checkcancel()) {
			errno = EINTR;
			goto error;
		}

		/* check if our buffer is full */
		if (txtidx == sizeof(linetxt) || c == EOF)
			if (txtidx) {
				buffer_nputs(buffer, undoable, *line, *linenum,
					*col, linetxt, txtidx);
				*col += txtidx;
				txtidx = 0;
			}

		/* check if we've finished */
		if (c == EOF) {
			if (ferror(fp))
				goto error;
			break;
		}

		/* insertions of characters or newlines */
		if (c == '\r' && buffer->nlstyle != NLSTYLE_LF) {
			c = fgetc(fp);
			if (c != '\n') {
				ungetc(c, fp);
				c = '\r';
			} else {
				buffer->nlstyle = NLSTYLE_CRLF;
				goto newline;
			}
		} else if (c == '\n') {
			if (buffer->nlstyle == NLSTYLE_UNKNOWN)
				buffer->nlstyle = NLSTYLE_LF;
			if (buffer->nlstyle == NLSTYLE_LF)
				goto newline;
		}
		linetxt[txtidx++] = c;
		continue;

		/* insert new lines into the buffer when we get them */
newline:
		if (txtidx) {
			buffer_nputs(buffer, undoable, *line, *linenum,
				*col, linetxt, txtidx);
			*col += txtidx;
			txtidx = 0;
		}
		next = buffer_addline(buffer, undoable, *line,
			*linenum);
		if (*col < (*line)->length) {
			buffer_nputs(buffer, undoable, next,
				*linenum + 1, 0, &(*line)->text[*col],
				(*line)->length - *col);
			buffer_rmstr(buffer, undoable, *line, *linenum,
				*col, (*line)->length - *col);
		}
		*line = next;
		(*linenum)++;
		*col = 0;
	}

	if (undoable)
		undo_endblock(buffer);
	return 0;

error:
	if (undoable)
		undo_endblock(buffer);
	return -1;
}

/* write the lines in a buffer out to a FILE */
int file_fwrite(buffer_t *buffer, FILE *fp) {
	line_t *line;
	int ret;

	TAILQ_FOREACH(line, &buffer->lines, l_list) {
		if (line->length) {
			ret = fwrite(line->text, 1, line->length, fp);
			if (ret < line->length)
				return -1;
		}
		if (!TAILQ_NEXT(line, l_list)) {
			if (options.forcefilenl && line->length)
				goto forcenl;
			break;
		}

forcenl:
		if (buffer->nlstyle == NLSTYLE_CRLF)
			putc('\r', fp);
		putc('\n', fp);
	}

	return 0;
}
