/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/*
 * minimum view sizes; these are used when initially splitting views; when
 * views are being resized through grow/shrink keys or a window size change,
 * the minimums are treated as 1 in either dimension.
 */
#define MIN_VIEW_WIDTH		5
#define MIN_VIEW_HEIGHT		3

/* list of all frames in existence */
framelist_t	frame_list	= TAILQ_HEAD_INITIALIZER(frame_list);

/* destroy all frames */
void frame_shutdown() {
	while (!TAILQ_EMPTY(&frame_list))
		frame_destroy(TAILQ_FIRST(&frame_list));
}

/*
 * allocate a new frame, with a single view in it.
 */
frame_t *frame_create(viewhdr_t *v, buffer_t *buffer) {
	frame_t *frame;

	frame = ckmalloc(sizeof(frame_t));
	frame->node = ckmalloc(sizeof(framenode_t));
	frame->node->type = FRAME_UNSPLIT;
	frame->node->d.v = v;
	frame->node->above = NULL;
	v->node = frame->node;
	TAILQ_INIT(&frame->views);
	TAILQ_INSERT_HEAD(&frame_list, frame, f_list);
	TAILQ_INSERT_TAIL(&frame->views, v, v_list);
	frame->width = screen_width;
	frame->height = screen_height;
	view_mapstack(v, buffer, 0, 1, screen_width, screen_height - 2);
	frame->redraw_hsplits = 0;

	return frame;
}

/*
 * recursive routine to free a tree of frame nodes, the views (and
 * their node pointers) are untouched.  frames aren't removed from
 * their parent nodes.
 */
static void ndfreetree(framenode_t *node) {
	int i;

	if (node->type != FRAME_UNSPLIT) {
		for (i = 0; i < node->d.s.count; i++) {
			ndfreetree(node->d.s.f[i]);
			free(node->d.s.f[i]);
		}
		free(node->d.s.f);
	}
}

/* destroy all the views in frame, free allocated memory */
void frame_destroy(frame_t *frame) {
	viewhdr_t *v;

	ndfreetree(frame->node);
	free(frame->node);
	while (!TAILQ_EMPTY(&frame->views)) {
		v = TAILQ_FIRST(&frame->views);
		TAILQ_REMOVE(&frame->views, v, v_list);
		while ((v = view_destroy(v)) != NULL)
			;
	}
	TAILQ_REMOVE(&frame_list, frame, f_list);
	free(frame);
}

/* redraw all the views and their headers */
void frame_draw(frame_t *frame) {
	viewhdr_t *v;
	int i;

	/* currently shouldn't ever draw anything other than active */
	assert(frame == frame_active);

	/* draw splitters if neccesary */
	if (!frame->redraw_hsplits)
		goto drawviews;
	frame->redraw_hsplits = 0;
	screen_setcolor(COLOR_HEADING);
	TAILQ_FOREACH(v, &frame->views, v_list) {
		if (v->geom.x == 0)
			continue;
		for (i = v->geom.y - 1; i < v->geom.y +
				v->geom.height; i++) {
			screen_gotoxy(v->geom.x - 1, i);
			screen_putc(' ');
		}
	}

drawviews:
	/* call draw and drawhdr for every view */
	TAILQ_FOREACH(v, &frame->views, v_list) {
		VIEWCALL(v, drawhdr, v->geom.x, v->geom.y - 1, v->geom.width);
		VIEWCALL(v, draw);
	}
}

/*
 * tell all the views in this frame that they need to be fully
 * redrawn next time.
 */
void frame_forcedraw(frame_t *frame) {
	viewhdr_t *v;

	TAILQ_FOREACH(v, &frame_active->views, v_list)
		VIEWCALL(v, activate);

	/* flag as needing a splitter redraw */
	frame_active->redraw_hsplits = 1;
}

/*
 * make a new framenode by splitting an existing one (that is already
 * a split node -- the first split must be done seperately); the
 * returned node's view should be set up by the caller.
 */
static __inline framenode_t *splitnode(framenode_t *node, int idx) {
	int i;

	assert(node->type != FRAME_UNSPLIT && node->d.s.count >= 2);
	node->d.s.count++;
	node->d.s.f = ckrealloc(node->d.s.f, sizeof(framenode_t *) *
		node->d.s.count);
	idx++;
	memmove(&node->d.s.f[idx + 1], &node->d.s.f[idx], (node->d.s.count -
		(idx + 1)) * sizeof(framenode_t *));
	for (i = idx + 1; i < node->d.s.count; i++)
		node->d.s.f[i]->aboveidx++;
	node->d.s.f[idx] = ckmalloc(sizeof(framenode_t));
	node->d.s.f[idx]->type = FRAME_UNSPLIT;
	node->d.s.f[idx]->above = node;
	node->d.s.f[idx]->aboveidx = idx;

	return node->d.s.f[idx];
}

static framenode_t *typedsplit(framenode_t *node, int type) {
	framenode_t *new;

	assert(node->type == FRAME_UNSPLIT);
	if (node->above && node->above->type == type)
		return splitnode(node->above, node->aboveidx);

	/*
	 * change the node into a node of type, move the view for
	 * this node into a child, and then do the split.
	 */
	new = ckmalloc(sizeof(framenode_t));
	new->type = FRAME_UNSPLIT;
	new->above = node;
	new->aboveidx = 0;
	new->d.v = node->d.v;
	new->d.v->node = new;

	/* set up the parent */
	node->type = type;
	node->d.s.count = 2;
	node->d.s.f = ckmalloc(2 * sizeof(framenode_t *));
	node->d.s.f[0] = new;

	/*
	 * the caller is responsible for setting up the new child.
	 */
	new = ckmalloc(sizeof(framenode_t));
	new->type = FRAME_UNSPLIT;
	new->above = node;
	new->aboveidx = 1;
	node->d.s.f[1] = new;

	return new;
}

void frame_vertsplit(frame_t *frame, viewhdr_t *view) {
	framenode_t *newnode;
	viewhdr_t *newview;
	int r, h;

	/* don't make unreasonably sized frames */
	r = view->geom.height / 2 - 1;
	h = view->geom.height - r - 1;
	if (r < MIN_VIEW_HEIGHT || h < MIN_VIEW_HEIGHT) {
		minibuff_printf("Not enough room to split vertical.");
		return;
	}

	/*
	 * split and add a new view in the new branch of the
	 * tree of framenodes.
	 */
	newnode = typedsplit(view->node, FRAME_VSPLIT);
	newview = view_duplicate(view);
	newview->node = newnode;
	newnode->d.v = newview;
	view_mapstack(newview, view->buffer, view->geom.x,
		view->geom.y + r + 1, view->geom.width, h);
	view_resize(view, view->geom.x, view->geom.y, view->geom.width, r);

	/* add the new one to the frame list */
	TAILQ_INSERT_TAIL(&frame->views, newview, v_list);
}

void frame_horizsplit(frame_t *frame, viewhdr_t *view) {
	framenode_t *newnode;
	viewhdr_t *newview;
	int c, w;

	/* don't make unreasonably sized frames */
	c = view->geom.width / 2 - 1;
	w = view->geom.width - c - 1;
	if (c < MIN_VIEW_WIDTH || w < MIN_VIEW_WIDTH) {
		minibuff_printf("Not enough room to split horizontal.");
		return;
	}

	/* split and add the new view */
	newnode = typedsplit(view->node, FRAME_HSPLIT);
	newview = view_duplicate(view);
	newview->node = newnode;
	newnode->d.v = newview;
	view_mapstack(newview, view->buffer, view->geom.x + c + 1,
		view->geom.y, w, view->geom.height);
	view_resize(view, view->geom.x, view->geom.y, c, view->geom.height);

	/* add the new one to the frame list */
	TAILQ_INSERT_TAIL(&frame->views, newview, v_list);

	/* we're going to need to draw splitters */
	frame->redraw_hsplits = 1;
}

/*
 * compute the screen dimensions of a framenode by recursing
 * down till we find views, and adding for splitters.
 */
static void nodedims(framenode_t *node, int *width, int *height) {
	int i;

	switch (node->type) {
	case FRAME_UNSPLIT:
		*width += node->d.v->geom.width;
		*height += node->d.v->geom.width;
		return;
	case FRAME_VSPLIT:
		for (i = 1; i < node->d.s.count; i++)
			(*height)++;
		break;
	case FRAME_HSPLIT:
		for (i = 1; i < node->d.s.count; i++)
			(*width)++;
		break;
	}

	for (i = 0; i < node->d.s.count; i++)
		nodedims(node->d.s.f[i], width, height);
}

/* merge a frame node with less than 2 entries upwards */
static int mergeup(frame_t *frame, framenode_t *node) {
	int ret, i;

	assert(node->type != FRAME_UNSPLIT && node->d.s.count < 2);

	/*
	 * if the node is empty, instead of merging it up, we just need
	 * to delete it's place in the parent node.  (if node->above isn't
	 * set here, there's something wrong..)
	 */
	if (node->d.s.count == 0) {
		assert(node->above);
		for (i = node->aboveidx + 1; i < node->above->d.s.count; i++)
			node->above->d.s.f[i]->aboveidx--;
		memmove(&node->above->d.s.f[node->aboveidx],
			&node->above->d.s.f[node->aboveidx + 1],
			(node->above->d.s.count - (node->aboveidx + 1)) *
			sizeof(framenode_t *));
		node->above->d.s.count--;
		ret = 1;
		goto free;
	}

	/*
	 * if there is no node above this, we can't merge up; instead
	 * just set the frame's top framenode to the only child of this
	 * one.
	 */
	ret = 0;
	if (!node->above) {
		assert(frame->node == node);
		frame->node = node->d.s.f[0];
		frame->node->above = NULL;
	} else {
		node->above->d.s.f[node->aboveidx] = node->d.s.f[0];
		node->d.s.f[0]->above = node->above;
		node->d.s.f[0]->aboveidx = node->aboveidx;
	}

free:
	free(node->d.s.f);
	free(node);

	return ret;
}

/* forward declaration */
static int recsz(frame_t *frame, framenode_t *node, int type,
		int idx, int doadjust, int move, int cnt);

/*
 * handle adjustment for the size of a removed frame (we need to correct
 * the amount of space that was being used for the splitter bar).  the
 * idx passed in is the idx of the just-removed node.
 */
static __inline void adjrmed(frame_t *frame, framenode_t *node, int type,
		int idx, int amt) {
	assert(node->type != FRAME_UNSPLIT);

	/* empty nodes is a valid call for convenience */
	if (node->d.s.count == 0)
		abort();

	/*
	 * it shouldn't be possible for either of these calls to
	 * recsz to result in node removal.
	 */
	if (idx > 0) {
		if (recsz(frame, node->d.s.f[idx - 1], type, 0, 0, 0, amt))
			assert(0);
	} else if (idx < node->d.s.count) {
		if (recsz(frame, node->d.s.f[idx], type, 0, 0, -amt, amt))
			assert(0);
	}
}

/*
 * recurse down a framenode tree and size in the specified manner; no
 * checking is done to make sure there is room for the root node to
 * size; that must be done from outside this routine.  if it is
 * impossible to handle making something under this tree size in the
 * specified manner, the view will be removed and it's frame unsplit.
 * this should only happen when making something smaller.
 *
 * the return value is used to communicate to the caller whether the
 * frame node needed to be destroyed because it couldn't be sized in
 * the requested manner.  nonzero indicates it was destroyed.  the
 * calling routine (should be this routine) should then handle sizing
 * adjacent nodes to occupy the space that was in use for the splitter
 * bar (using the adjrmed routine).
 *
 * rather than try to obfuscate the recsz logic to support magnitudes of
 * cnt > 1, it's easier (although perhaps less efficent) to just require
 * that it be called multiple times.  this restriction isn't placed on
 * situations that can be certain that there is room (such as adjrmed).
 *
 * doadjust indicates whether another child node on idx should be sized
 * opposite cnt to accommodate the change at idx
 *
 * XXX: when removing impossible views, if the buffer is modified, data
 * will currently be lost.  the buffer needs to be detached; this isn't
 * yet possible because buffer detaching isn't implemented yet.
 */
static int recsz(frame_t *frame, framenode_t *node, int type,
		int idx, int doadjust, int move, int cnt) {
	viewhdr_t *v;
	int x, y, width, height;
	int szidx, low, high, i;

	/*
	 * for unsplit nodes just size/move the view according to the
	 * parameters that were passed in.
	 */
	if (node->type == FRAME_UNSPLIT) {
		x = node->d.v->geom.x;
		y = node->d.v->geom.y;
		width = node->d.v->geom.width;
		height = node->d.v->geom.height;

		switch (type) {
		case FRAME_VSPLIT:
			y += move;
			height += cnt;
			if (height < 1)
				goto destroy;
			break;
		case FRAME_HSPLIT:
			x += move;
			width += cnt;
			if (width < 1)
				goto destroy;
			break;
		}

		view_resize(node->d.v, x, y, width, height);
		return 0;

destroy:
		/*
		 * XXX: need to make the buffer 'backgrounded' if this
		 * will rm the last reference to it.
		 */
		if (!node->above)
			abort();
		v = node->d.v;
		TAILQ_REMOVE(&frame->views, v, v_list);
		while ((v = view_destroy(v)) != NULL)
			;
		for (i = node->aboveidx + 1; i < node->above->d.s.count; i++)
			node->above->d.s.f[i]->aboveidx--;
		memmove(&node->above->d.s.f[node->aboveidx],
			&node->above->d.s.f[node->aboveidx + 1],
			(node->above->d.s.count - (node->aboveidx + 1)) *
			sizeof(framenode_t *));
		node->above->d.s.count--;
		free(node);
		return 1;
	}

	/*
	 * if this node is of a different split type, just propagate
	 * to all child nodes and handle merging.
	 */
	if (node->type != type) {
		for (i = 0; i < node->d.s.count;) {
			if (!recsz(frame, node->d.s.f[i], type, 0, 0,
					move, cnt))
				i++;
		}
		if (node->d.s.count < 2) {
			assert(node->d.s.count == 0);
			return mergeup(frame, node);
		}
		return 0;
	}

	/*
	 * if we don't need to adjust a sibling to make room, just
	 * size the idx we were handed.
	 */
	if (!doadjust) {
		/*
		 * this probably only works with idx == 0, and there is
		 * currently no reason to make it work for other idx yet.
		 */
		assert(idx == 0);

		/*
		 * the logic here is slightly different if the node isn't
		 * supposed to be moved, but only sized, or isn't supposed
		 * to be sized, but only moved.  the simpler case is when
		 * it is both moved and sized, because all of the changes
		 * may be made to the first child node (plus any removal
		 * correction).
		 */
		if (!move || !cnt) {
			while (recsz(frame, node->d.s.f[idx], type, 0, 0,
					move, cnt))
				adjrmed(frame, node, type, idx, 2);

			/*
			 * move the ones past the one that was sized to
			 * accomodate the same space.  it should be impossible
			 * for these recszes to destroy a node.
			 */
			for (i = idx + 1; i < node->d.s.count; i++) {
				if (recsz(frame, node->d.s.f[i], type, 0, 0,
						move + cnt, 0))
					assert(0);
			}
		} else {
			if (recsz(frame, node->d.s.f[idx], type, 0, 0,
					move, cnt))
				adjrmed(frame, node, type, idx, 1);
		}

		if (node->d.s.count < 2)
			return mergeup(frame, node);
		return 0;
	}

	/*
	 * determine the idx of the node that should be corrected for
	 * the change of cnt in the node at idx.
	 */
	szidx = idx + 1;
	if (szidx == node->d.s.count)
		szidx = 0;

	/*
	 * adjust node at szidx to accommodate the change in the node
	 * at idx; the nodes in between to must be moved, but their
	 * sizes wont change.
	 */
	low = min(szidx, idx);
	high = max(szidx, idx);
	if (recsz(frame, node->d.s.f[low], type, 0, 0, move,
			low == idx ? cnt : -cnt)) {
		adjrmed(frame, node, type, low, 1);
		low--, high--, idx--;
	}
	if (recsz(frame, node->d.s.f[high], type, 0, 0, move +
			(high == idx ? -cnt : cnt),
			high == idx ? cnt : -cnt))
		adjrmed(frame, node, type, high, 1);
	for (i = low + 1; i < high; i++)
		if (recsz(frame, node->d.s.f[i], type, 0, 0, move + -cnt, 0))
			/*
			 * this condition should never happen.  unless I
			 * misthought this...
			 */
			assert(0);
	if (node->d.s.count < 2)
		return mergeup(frame, node);

	return 0;
}

/*
 * return the first node encountered of the specified type while
 * traversing upwards.  the index of the branch used to travel
 * up to this node is passed out as well.
 */
static framenode_t *topnode(framenode_t *node, int type, int *idx) {
	while (node->above) {
		if (node->above->type == type) {
			*idx = node->aboveidx;
			return node->above;
		}

		node = node->above;
	}

	return NULL;
}

int frame_size(frame_t *frame, viewhdr_t *v, int type, int cnt) {
	framenode_t *node;
	int idx;

	node = topnode(v->node, type, &idx);
	if (!node)
		return -1;
	recsz(frame, node, type, idx, 1, 0, cnt);
	frame->redraw_hsplits = 1;

	return 0;
}

/*
 * resize all the frames to the specified size (should probably be
 * the new screen size).
 */
void frame_resize(int width, int height) {
	frame_t *frame;
	int dir;

	/*
	 * size every frame according to the change in size.  see
	 * the explaination at recsz() for why the recsz routine
	 * is called multiple times.
	 */
	TAILQ_FOREACH(frame, &frame_list, f_list) {
		dir = width - frame->width > 0 ? 1 : -1;
		while (frame->width != width) {
			recsz(frame, frame->node, FRAME_HSPLIT, 0, 0, 0, dir);
			frame->width += dir;
		}
		dir = height - frame->height > 0 ? 1 : -1;
		while (frame->height != height) {
			recsz(frame, frame->node, FRAME_VSPLIT, 0, 0, 0, dir);
			frame->height += dir;
		}
		frame->redraw_hsplits = 1;
	}
}

int frame_rmview(frame_t *frame, viewhdr_t *v) {
	framenode_t *above;
	int useidx, amt, dir;

	/* don't remove the last view */
	if (TAILQ_FIRST(&frame->views) == TAILQ_LAST(&frame->views, viewlist))
		return -1;

	/*
	 * the simplest way to do this is to just call the sizing routine to
	 * shrink the node till it needs to be destroyed.  which idx we use
	 * it on is obviously closely tied to the choice of szidx in recsz:
	 * if recsz changes, changes will be required here.
	 */
	assert(v->node->above);
	above = v->node->above;
	useidx = v->node->aboveidx - 1;
	if (useidx < 0) {
		useidx = 0;
		dir = -1;
	} else
		dir = 1;
	amt = above->type == FRAME_VSPLIT ? v->geom.height : v->geom.width;
	while (amt--)
		if (recsz(frame, above, above->type,
				useidx, 1, 0, dir))
			assert(amt == -1);
	frame->redraw_hsplits = 1;

	return 0;
}

int frame_oneview(frame_t *frame, viewhdr_t *oneview) {
	viewhdr_t *v;

	/* don't do it if there's already only one view */
	if (!oneview->node->above)
		return -1;

	/* free old views and the tree of nodes */
	assert(frame->node->type != FRAME_UNSPLIT);
	ndfreetree(frame->node);
	while (!TAILQ_EMPTY(&frame->views)) {
		v = TAILQ_FIRST(&frame->views);
		TAILQ_REMOVE(&frame->views, v, v_list);
		if (v != oneview)
			while ((v = view_destroy(v)) != NULL)
				;
	}

	/* revert the root node to an unsplit node */
	frame->node->type = FRAME_UNSPLIT;
	frame->node->d.v = oneview;
	oneview->node = frame->node;

	TAILQ_INSERT_TAIL(&frame->views, oneview, v_list);
	view_resize(oneview, 0, 1, screen_width, screen_height - 2);

	return 0;
}
