/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef FRAME_H
#define FRAME_H

/* types of framenodes */
enum {
	FRAME_UNSPLIT,
	FRAME_VSPLIT,
	FRAME_HSPLIT
};

/*
 * node in the tree of views on a frame; each node can either be
 * a horizontal or vertical split, or an unsplit view.
 */
struct framenode {
	int		type;

	/*
	 * back pointer to parent, and the index that this frame occupies
	 * in the parent's array of children.
	 */
	framenode_t	*above;
	int		aboveidx;

	/* type specific data */
	union {
		viewhdr_t *v;

		/* for split nodes */
		struct {
			int		count;
			framenode_t	**f;
		} s;
	} d;
};

/*
 * a frame is a collection of views; on a tty there will only be
 * a single visible frame at a time, and the user can switch which
 * one is visible using keyboard commands.  all frames have an
 * active view, which is the view to recieve input when the frame
 * is active; the active view is the last on the views list, and
 * should be accessed using the macro view_current (from view.h)
 */
struct frame {
	TAILQ_ENTRY(frame) f_list;

	framenode_t	*node;		/* main frame node */
	viewlist_t	views;		/* views displayed on the frame */
	int		redraw_hsplits;	/* flag if we need to redraw */
	int		width, height;
};

TAILQ_HEAD(framelist, frame);

/*
 * macro to obtain the currently active frame
 */
#define frame_active	TAILQ_FIRST(&frame_list)

extern framelist_t	frame_list;

void	frame_shutdown();
frame_t	*frame_create(viewhdr_t *, buffer_t *);
void	frame_destroy(frame_t *);
void	frame_draw(frame_t *);
void	frame_forcedraw(frame_t *);
void	frame_vertsplit(frame_t *, viewhdr_t *);
void	frame_horizsplit(frame_t *, viewhdr_t *);
int	frame_size(frame_t *, viewhdr_t *, int type, int cnt);
void	frame_resize(int, int);
int	frame_rmview(frame_t *, viewhdr_t *);
int	frame_oneview(frame_t *, viewhdr_t *);

static __inline void frame_activate(frame_t *frame) {
	TAILQ_REMOVE(&frame_list, frame, f_list);
	TAILQ_INSERT_HEAD(&frame_list, frame, f_list);
}

static __inline void frame_focview(frame_t *frame, viewhdr_t *v) {
	TAILQ_REMOVE(&frame->views, v, v_list);
	TAILQ_INSERT_TAIL(&frame->views, v, v_list);
}

#endif
