/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* flag set to cause input_loop to exit */
int	input_exitloop		= 0;

/* last keysym the user sent */
int	input_lastkeysym	= KEYSYM_NOP;

/*
 * draws can be protected by incrementing this variable.  to allow
 * different modules or subsystems (like macro) to concurrently
 * disable drawing, this variable should only be incremented and
 * decremented; not explicitly set to 1 or 0.
 */
int	input_nodraw		= 0;

/* this should eventually be moved to screen.c */
static struct termios		save_termios;

/*
 * initalize all default keysym definitions
 */
void input_init() {
	if (tcgetattr(STDIN_FILENO, &save_termios) == -1)
		err(1, "tcgetattr");

	/* basic keysyms */
	keybind_define(KEYSYM_UP,		"<up>");
	keybind_define(KEYSYM_DOWN,		"<down>");
	keybind_define(KEYSYM_LEFT,		"<left>");
	keybind_define(KEYSYM_RIGHT,		"<right>");
	keybind_define(KEYSYM_PGUP,		"<ppage>");
	keybind_define(KEYSYM_PGDN,		"<npage>");
	keybind_define(KEYSYM_PGUP,		"<previous>");
	keybind_define(KEYSYM_PGDN,		"<next>");
	keybind_define(KEYSYM_HOME,		"<home>");
	keybind_define(KEYSYM_END,		"<end>");
	keybind_define(KEYSYM_BACKSPACE,	"<backspace>");
	keybind_define(KEYSYM_DELETE,		"<dc>");
	keybind_define(KEYSYM_ENTER,		"<enter>");

	/* main keys */
	keybind_define(KEYSYM_QUIT,		"\\^X\\^C");
	keybind_define(KEYSYM_SUSPEND,		"\\^Z");
	keybind_define(KEYSYM_CANCEL,		"\\^C");
	keybind_define(KEYSYM_CANCEL,		"\\^G");
	keybind_define(KEYSYM_EXCMD,		"\\33x");
	keybind_define(KEYSYM_SWITCHBUFF,	"\\^Xb");
	keybind_define(KEYSYM_SETOPT,		"\\33oo");
	keybind_define(KEYSYM_OPENFILE,		"\\^X\\^F");
	keybind_define(KEYSYM_SAVE,		"\\^X\\^S");
	keybind_define(KEYSYM_SAVEAS,		"\\^X\\^W");
	keybind_define(KEYSYM_REDRAW,		"\\33l");
	keybind_define(KEYSYM_QUOTENEXT,	"\\33\\\\");
	keybind_define(KEYSYM_AGAIN,		"\\^\\");
	keybind_define(KEYSYM_VERTSPLIT,	"\\^_");
	keybind_define(KEYSYM_HORIZSPLIT,	"\\33-");
	keybind_define(KEYSYM_VERTSPLIT,	"\\^X2");
	keybind_define(KEYSYM_HORIZSPLIT,	"\\^X3");
	keybind_define(KEYSYM_RMVIEW,		"\\^X0");
	keybind_define(KEYSYM_RMOTHERVIEWS,	"\\^X1");
	keybind_define(KEYSYM_VIEWCYCLE,	"\\33v");
	keybind_define(KEYSYM_VIEWCYCLE,	"\\^Xo");
	keybind_define(KEYSYM_PUSHVIEW,		"\\0330");
	keybind_define(KEYSYM_POPVIEW,		"\\339");
	keybind_define(KEYSYM_VERTGROW,		"\\33.");
	keybind_define(KEYSYM_VERTSHRINK,	"\\33,");
	keybind_define(KEYSYM_HORIZGROW,	"\\33>");
	keybind_define(KEYSYM_HORIZSHRINK,	"\\33<");
	keybind_define(KEYSYM_FRAMECYCLE,	"\\33=");
	keybind_define(KEYSYM_NEWFRAME,		"\\^X52");
	keybind_define(KEYSYM_NEWFRAME,		"\\33+");
	keybind_define(KEYSYM_RMFRAME,		"\\^X=");
	keybind_define(KEYSYM_RMFRAME,		"\\^X50");

	/* keyboard macro keys */
	keybind_define(KEYSYM_RECORD,		"\\33r");
	keybind_define(KEYSYM_PLAYBACK,		"\\33p");
	keybind_define(KEYSYM_REPEAT,		"\\33g");

	/* vdefault keys */
	keybind_define(KEYSYM_HOME, 		"\\^A");
	keybind_define(KEYSYM_END,		"\\^E");
	keybind_define(KEYSYM_PGUP,		"\\^F");
	keybind_define(KEYSYM_PGDN,		"\\^V");
	keybind_define(KEYSYM_OPENLINE,		"\\^O");
	keybind_define(KEYSYM_TOGNLSTYLE,	"\\33on");
	keybind_define(KEYSYM_UNDO,		"\\^U");
	keybind_define(KEYSYM_REDO,		"\\33u");
	keybind_define(KEYSYM_NEXTWORD,		"\\33f");
	keybind_define(KEYSYM_PREVWORD,		"\\33b");
	keybind_define(KEYSYM_NEXTPARA,		"\\33}");
	keybind_define(KEYSYM_PREVPARA,		"\\33{");
	keybind_define(KEYSYM_VIEWTOP,		"\\33\\^F");
	keybind_define(KEYSYM_VIEWBOTTOM,	"\\33\\^V");
	keybind_define(KEYSYM_LINEUP,		"\\^Q");
	keybind_define(KEYSYM_LINEDN,		"\\^D");
	keybind_define(KEYSYM_FIRSTLINE,	"\\33\\^A");
	keybind_define(KEYSYM_LASTLINE,		"\\33\\^E");
	keybind_define(KEYSYM_HALFPGUP,		"\\33(");
	keybind_define(KEYSYM_HALFPGDN,		"\\33)");
	keybind_define(KEYSYM_CENTER,		"\\^L");
	keybind_define(KEYSYM_TRANCHARS,	"\\^T");
	keybind_define(KEYSYM_TRANWORDS,	"\\33t");
	keybind_define(KEYSYM_TRANLINES,	"\\^X\\^T");
	keybind_define(KEYSYM_SETMARK,		"\\000");
	keybind_define(KEYSYM_POPMARK,		"\\33 ");
	keybind_define(KEYSYM_PROMOTEMARK,	"\\^X ");
	keybind_define(KEYSYM_DEMOTEMARK,	"\\33@");
	keybind_define(KEYSYM_XCHGMARK,		"\\^X\\^X");
	keybind_define(KEYSYM_KILLLINE,		"\\^K");
	keybind_define(KEYSYM_KILLEOL,		"\\33k");
	keybind_define(KEYSYM_KILLWORD,		"\\33\b");
	keybind_define(KEYSYM_KILLWORD,		"\\^W");
	keybind_define(KEYSYM_FWDKILLWORD,	"\\33d");
	keybind_define(KEYSYM_APPENDKILL,	"\\33\\^W");
	keybind_define(KEYSYM_YANK,		"\\^Y");
	keybind_define(KEYSYM_YANKPOP,		"\\33y");
	keybind_define(KEYSYM_FWDYANKPOP,	"\\33\\^Y");
	keybind_define(KEYSYM_ISEARCH,		"\\^S");
	keybind_define(KEYSYM_RISEARCH,		"\\^R");
	keybind_define(KEYSYM_REPLACE,		"\\33%");
	keybind_define(KEYSYM_GOTOLN,		"\\^]");
	keybind_define(KEYSYM_FILTER,		"\\33|");
	keybind_define(KEYSYM_INSERTFILE,	"\\^Xi");
}

/*
 * allocate a range of keysym identifiers; used by views so that
 * potential keysym conflicts are prevented.
 */
int input_allocsyms(int count) {
	static int nextblk = KEYSYM_FIRSTVIEWSYM;
	int ret;

	ret = nextblk;
	nextblk += count;

	return ret;
}

/* get a key and handle resizes and macro playback */
int input_getkey() {
	int was_repeating;
	int c;

	was_repeating = macro_repeating;
	for (;;) {
		/*
		 * get a key, either from the macro recorder/repeater,
		 * or from the tty
		 */
		if (macro_playing || macro_repeating)
			c = macro_getkey();
		else if (macro_repeatplay) {
			c = KEYSYM_PLAYBACK;
			if (--macro_repeatplay == 0)
				input_nodraw--;
		} else
			c = keybind_getkey();

		/*
		 * handle window resizing; return a KEYSYM_NOP so
		 * the size changes get redrawn.
		 */
		if (c == KEYSYM_RESIZE) {
			frame_resize(screen_width, screen_height);
			minibuff_resize(screen_width, screen_height);
			return KEYSYM_NOP;
		}

		/*
		 * handle certain keys from here, so that they work
		 * anywhere.
		 */
		switch (c) {
		case KEYSYM_QUOTENEXT:
			input_lastkeysym = KEYSYM_QUOTENEXT;
			screen_nokeysyms(1);
			c = input_getkey();
			screen_nokeysyms(0);
			break;
		case KEYSYM_REDRAW:
			screen_redraw();
			continue;
		}

		/*
		 * record the key if the macro recorder is
		 * recording, repeated keys aren't recorded.
		 */
		if (macro_recording && !was_repeating)
			return macro_recordkey(c);
		else
			return c;
	}
}

/*
 * check for pending input and take into account macro playback and
 * the key repeater.  block for up to m millis.
 */
int input_key_pending(int m) {
	if (macro_playing || macro_repeating)
		return 1;
	return screen_key_pending(m);
}

/*
 * determine if a key is a control (non printable) character
 * or not.
 */
int input_isctrl(int keysym) {
	if (keysym >= SCREEN_FIRSTSYM)
		return 1;
	if ((u_char) keysym == '\t')
		return 0;
	if (!isprint((u_char) keysym))
		return 1;
	return 0;
}

/*
 * default keysym handling; vdefault drops keysyms to this
 * handler when they are app-level keysyms (save, quit, and
 * so forth).
 */
void input_default_handler(viewhdr_t *view, int keysym) {
	command_t *command;

	switch (keysym) {
	case KEYSYM_QUIT:
		input_exitloop = 1;
		break;
	case KEYSYM_SUSPEND:
		screen_suspend();
		kill(getpid(), SIGTSTP);
		screen_reinit();
		break;
	case KEYSYM_HASHSTATS:		command_hashstats();	break;
	case KEYSYM_SAVEBUFFERS:	command_savebuffers();	break;
	case KEYSYM_SWITCHBUFF:		command_switchbuff();	break;
	case KEYSYM_SETOPT:		command_setopt();	break;
	case KEYSYM_OPENFILE:		command_openfile();	break;
	case KEYSYM_SAVE:		command_save();		break;
	case KEYSYM_SAVEAS:		command_saveas();	break;
	case KEYSYM_VIEWCYCLE:		command_viewcycle();	break;
	case KEYSYM_PUSHVIEW:		command_pushview();	break;
	case KEYSYM_POPVIEW:		command_popview();	break;
	case KEYSYM_VERTSPLIT:		command_vertsplit();	break;
	case KEYSYM_HORIZSPLIT:		command_horizsplit();	break;
	case KEYSYM_RMVIEW:		command_rmview();	break;
	case KEYSYM_RMOTHERVIEWS:	command_rmotherviews();	break;
	case KEYSYM_VERTGROW:		command_vertgrow();	break;
	case KEYSYM_VERTSHRINK:		command_vertshrink();	break;
	case KEYSYM_HORIZGROW:		command_horizgrow();	break;
	case KEYSYM_HORIZSHRINK:	command_horizshrink();	break;
	case KEYSYM_FRAMECYCLE:		command_framecycle();	break;
	case KEYSYM_NEWFRAME:		command_newframe();	break;
	case KEYSYM_RMFRAME:		command_rmframe();	break;

	case KEYSYM_RECORD:		macro_record();		break;
	case KEYSYM_PLAYBACK:		macro_playback();	break;
	case KEYSYM_REPEAT:		macro_repeat();		break;

	default:
		/*
		 * try a function-pointer-based command (these are usually
		 * from a module).
		 */
		command = command_getcmd(keysym);
		if (command && command->type == COMMAND_FUNC) {
			command->handler(command);
			break;
		}

		/*
		 * express our displeasure
		 */
		screen_bell();
	}
}


/*
 * give the active view control over input; if we're
 * editing in the minibuffer, send it to the minibuffer
 * handler instead.
 */
void input_dispatch(int keysym) {
	if (minibuff_editing)
		minibuff_input(keysym);
	else
		VIEWCALL(view_current, input, keysym);
	input_lastkeysym = keysym;
}

/* main input reception loop */
void input_loop() {
	int c;

	/*
	 * loop getting keys from the tty and handing them off to
	 * the view's input handler.  drawing is done unless we
	 * are executing a macro or repeating a key.
	 */
	input_exitloop = 0;
	while (!input_exitloop) {
		if (!input_nodraw) {
			draw_screen();
			screen_refresh();

			/* remove buffer dirty flags */
			buffer_cleanse();
		}

		/* get a key to send to the view's input handler */
		c = input_getkey();
		while (c == KEYSYM_EXCMD)
			c = command_excmd();
		switch (c) {
		case KEYSYM_AGAIN:
			c = input_lastkeysym;
			break;
		case KEYSYM_NOP:
			continue;
		}

		input_dispatch(c);
	}
}

/*
 * to make cancel handling allow cancels when inside a OS call that
 * blocks, we use the tty's support for a key that will generate
 * SIGINT.
 */
static volatile int gotsigint = 0;
static void inthdlr(int sig) {
	gotsigint = 1;
}

/* cancellation variables */
static int *cancel_typeahead;
static int cancel_len;
static int cancel_idx;
static int canceled;

/* begin a cancelable operation */
void input_begincancel() {
	struct termios tios;

	canceled = cancel_idx = 0;
	cancel_len = 10;
	cancel_typeahead = ckmalloc(cancel_len * sizeof(int));

	/*
	 * enable and set the tty's interrupt key to the character specified
	 * by the user's options.
	 */
	signal(SIGINT, inthdlr);
	memcpy(&tios, &save_termios, sizeof(struct termios));
	tios.c_cc[VINTR] = options.intrchar;
	tios.c_cc[VQUIT] = _POSIX_VDISABLE;
	tios.c_cc[VSUSP] = _POSIX_VDISABLE;
	tios.c_lflag |= ISIG;
	if (tcsetattr(STDIN_FILENO, TCSANOW, &tios) == -1)
		err(1, "tcsetattr");
}

/* check to see if the operation was canceled */
int input_checkcancel() {
	int c;

	if (canceled)
		return 1;
	if (gotsigint)
		goto cancel;
	while (screen_key_pending(0)) {
		c = keybind_getkey();

		if (gotsigint)
			goto cancel;
		if (c == KEYSYM_CANCEL)
			goto cancel;
		cancel_typeahead[cancel_idx++] = c;
		if (cancel_idx >= cancel_len) {
			cancel_len += 10;
			cancel_typeahead = ckrealloc(cancel_typeahead,
				cancel_len);
		}
	}

	return 0;

cancel:
	/* trash keys in the tty */
	while (screen_key_pending(0))
		screen_getkey();
	canceled = 1;
	return 1;
}

/* this should be called at the end of a cancelable operation */
void input_endcancel() {
	int i;

	if (!canceled) {
		for (i = cancel_idx - 1; i >= 0; i--)
			keybind_ungetkey(cancel_typeahead[i]);
	}
	signal(SIGINT, SIG_IGN);
	if (tcsetattr(STDIN_FILENO, TCSANOW, &save_termios) == -1)
		err(1, "tcsetattr");
	gotsigint = 0;
	free(cancel_typeahead);
}
