/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* keybinding entry */
typedef struct keybind {
	LIST_ENTRY(keybind) kb_list;

	int	*keystr;
	int	len;
	int	keysym;		/* the keysym at this node */
	int	refcnt;		/* references */
	u_int	ambig:1;	/* this string is ambiguous */
	u_int	valid:1;	/* flags ambig placeholders w/o keysyms */
} keybind_t;
static LIST_HEAD(keybindlist, keybind)	*keybindhash;
#define KEYBIND_BUCKETS			128

/*
 * key recording buffer used for ungetkey and while reading
 * different keys in a keybinding
 */
static int *key_buf;
static int key_idx, key_size;

void keybind_init() {
	keybindhash = hash_alloc("keybindhash", KEYBIND_BUCKETS,
		offsetof(keybind_t, kb_list));
	key_size = 40;
	key_buf = ckmalloc(key_size * sizeof(int));
}

void keybind_shutdown() {
	keybind_clear();
	hash_free(keybindhash);
	free(key_buf);
}

void keybind_clear() {
	keybind_t *keybind;
	int idx;

	for (idx = 0; idx < KEYBIND_BUCKETS; idx++)
		while (!LIST_EMPTY(&keybindhash[idx])) {
			keybind = LIST_FIRST(&keybindhash[idx]);
			LIST_REMOVE(keybind, kb_list);
			free(keybind->keystr);
			free(keybind);
		}
}

/* lookup a string in the keybind hashtable */
static __inline keybind_t *getkb(int *s, int slen) {
	keybind_t *kb;
	int idx;

	idx = hash_mem(KEYBIND_BUCKETS, (u_char *) s, slen);
	LIST_FOREACH(kb, &keybindhash[idx], kb_list)
		if (kb->len == slen && !memcmp(kb->keystr, s,
				slen * sizeof(int)))
			return kb;
	return NULL;
}

/* add a keybinding into the keybind hashtable */
static __inline void addkb(keybind_t *kb) {
	int idx;

	idx = hash_mem(KEYBIND_BUCKETS, (u_char *) kb->keystr, kb->len);
	LIST_INSERT_HEAD(&keybindhash[idx], kb, kb_list);
}

/* remove a keybinding */
static void rmbind(keybind_t *kb) {
	keybind_t *back;

	if (kb->len > 1) {
		back = getkb(kb->keystr, kb->len - 1);
		if (!--back->refcnt)
			rmbind(back);
		else if (back->refcnt == 1 && back->valid)
			back->ambig = 0;
	}
	LIST_REMOVE(kb, kb_list);
	free(kb->keystr);
	free(kb);
}

/* unbind all the keys mapping to a keysym */
static __inline void unbindsym(int keysym) {
	keybind_t *kb, *next;
	int idx;

	for (idx = 0; idx < KEYBIND_BUCKETS; idx++) {
		kb = LIST_FIRST(&keybindhash[idx]);
		while (kb) {
			next = LIST_NEXT(kb, kb_list);
			if (kb->keysym == keysym && kb->valid) {
				if (--kb->refcnt == 0)
					rmbind(kb);
				else
					kb->valid = 0;
			}
			kb = next;
		}
	}
}

/* this function is generated and in another module */
extern int getnamedkey(u_char *);

/* return a key by name */
static __inline int namedkey(u_char *s, int slen, int *i) {
	u_char *str;
	u_char *end;
	int len, c;

	end = memchr(s, '>', slen);
	if (!end)
		return '<';
	len = end - (s + 1);
	*i += len + 1;
	if (!len)
		return '<';
	str = ckmalloc(len + 1);
	memcpy(str, s + 1, len);
	str[len] = '\0';
	c = getnamedkey(str);
	free(str);

	return c;
}

/*
 * used to turn a string that is being bound to a keysym into an array
 * of integers that will be returned from screen_getkey to indicate
 * this sequence of key presses.  the string can contain <keyname> sequences
 * which represent the KEY_ macros getch() can return.
 */
int *getkeystr(u_char *s, int slen, int *len) {
	int *keystr;
	int i, idx;

	*len = slen;
	keystr = ckmalloc(slen * sizeof(int));
	for (idx = i = 0; i < slen; i++)
		if (s[i] == '<')
			keystr[idx++] = namedkey(&s[i], slen - i, &i);
		else
			keystr[idx++] = s[i];
	*len = idx;

	return keystr;
}

/*
 * associate a key string with an identifier; if key is null, all
 * bindings to the specified keysym are removed.  if keysym is 0,
 * the keybinding specified by the string passed in is removed, if
 * it exists.
 *
 * the string is the a format specified by bsd vis(3).  the resultant
 * string (after unvising it) also may have <keyname> where keyname is
 * a format understood in screen.c; the <'s may be quoted to prevent
 * this step.
 */
int keybind_define(int keysym, u_char *key) {
	keybind_t *kb;
	u_char *unvisstr;
	int *keystr;
	int noambig;
	int len, i;

	/*
	 * if key is unset, remove all keybindings that map to the
	 * specified keysym.
	 */
	if (!key) {
		unbindsym(keysym);
		return 0;
	}

	unvisstr = ckmalloc(strlen(key) + 1);
	len = strunvis(unvisstr, key);
	if (len == -1 || len == 0) {
		free(unvisstr);
		return -1;
	}
	keystr = getkeystr(unvisstr, len, &len);
	free(unvisstr);

	/*
	 * if keysym is unset, remove the keybinding specified by the
	 * key string from the keybinding table.
	 */
	if (keysym == 0) {
		kb = getkb(keystr, len);
		if (!kb || !kb->valid)
			goto free;
		kb->valid = 0;
		if (!--kb->refcnt)
			rmbind(kb);
		goto free;
	}

	/*
	 * if the keybinding is already registered, just change
	 * what it means.  it's not neccesary to set up ambiguity
	 * entries because they already exist from when the old
	 * one was added.
	 */
	kb = getkb(keystr, len);
	if (!kb) {
		kb = ckmalloc(sizeof(keybind_t));
		kb->keystr = ckmalloc(len * sizeof(int));
		memcpy(kb->keystr, keystr, len * sizeof(int));
		kb->len = len;
		kb->ambig = 0;
		kb->refcnt = 1;
		noambig = 0;
		addkb(kb);
	} else {
		noambig = 1;
		kb->refcnt++;
	}
	kb->valid = 1;
	kb->keysym = keysym;
	if (noambig)
		goto free;

	/*
	 * add ambiguity entries for everything smaller than the
	 * full screen.
	 */
	for (i = 1; i < len; i++) {
		kb = getkb(keystr, i);
		if (!kb) {
			kb = ckmalloc(sizeof(keybind_t));
			kb->keystr = ckmalloc(i * sizeof(int));
			memcpy(kb->keystr, keystr, i * sizeof(int));
			kb->len = i;
			kb->valid = 0;
			kb->refcnt = 1;
			addkb(kb);
		} else
			kb->refcnt++;
		kb->ambig = 1;
	}

free:
	free(keystr);
	return 0;
}

static __inline void relockeybuf() {
	key_size += 40;
	key_buf = ckrealloc(key_buf, key_size * sizeof(int));
}

#define STRSZ		40

/*
 * return the ident of a pressed key, the identifiers for simple
 * characters are the value of the character ('a', 'b', and so forth)
 */
int keybind_getkey() {
	static int string[STRSZ];
	static int stridx = 0;
	keybind_t *kb;
	int c;

	/* stored keys get flushed out */
	if (key_idx)
		return key_buf[--key_idx];
	if (!screen_keysyms)				/* XXX */
		return screen_getkey();

	/* perform keybinding translation */
	assert(key_idx == 0 && stridx == 0);
again:
	c = screen_getkey();
	if (stridx > STRSZ - 2)
		abort();
	string[stridx++] = c;
	kb = getkb(string, stridx);
	if (!kb) {
		/*
		 * support terminals which set the high-bit of a character
		 * when the user has the meta key depressed.
		 */
		if (options.highbitmeta && c <= UCHAR_MAX && c > 0x80) {
			string[stridx - 1] = '\33';
			string[stridx] = c & ~0x80;
			kb = getkb(string, stridx + 1);
			if (!kb) {
				string[stridx - 1] = c;
				goto flush;
			}
			stridx++;
		} else
			goto flush;
	}
	if (kb->ambig) {
		/*
		 * give the user an amount of time to clarify an ambiguous
		 * keybinding; if no new keys are pressed in the amount of
		 * time, return the current sym if valid, or flush the string.
		 */
		if (screen_key_pending(500))
			goto again;
		if (!kb->valid)
			goto flush;
	}

	stridx = 0;
	return kb->keysym;

flush:
	/*
	 * didn't find anything so unget the entire string, and return
	 * one key to the user.
	 */
	assert(stridx != 0);
	while (stridx)
		keybind_ungetkey(string[--stridx]);
	assert(key_idx);
	return keybind_getkey();
}

void keybind_ungetkey(int k) {
	if (key_idx > key_size - 1)
		relockeybuf();
	key_buf[key_idx++] = k;
}
