/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef LINE_H
#define LINE_H

/*
 * size of a line when it is initially allocated; the line
 * size goes up by doubling its size each time it overflows.
 * a line can decrease in length, but we never lessen the
 * amount of memory allocated for it.
 */
#define LINE_STARTSIZE		10

/*
 * line structure
 */
struct line {
	TAILQ_ENTRY(line) l_list;

	int	length;	/* the length of the line */
	int	size;	/* amount of memory allocated for the line */
	u_char	*text;	/* string of text for the line; no null terminator */
};

TAILQ_HEAD(linelist, line);

void	line_putc(line_t *line, int loc, u_char c);
void	line_nputs(line_t *line, int loc, u_char *s, int slen);

/* add a line to a list of lines */
static __inline line_t *line_insert(linelist_t *head, line_t *line) {
	line_t *newln;

	/* allocate a new line structure */
	newln = ckmalloc(sizeof(line_t));
	newln->text = ckmalloc(LINE_STARTSIZE);

	/* fill it in and link it onto the list */
	newln->length = 0;
	newln->size = LINE_STARTSIZE;
	if (line)
		TAILQ_INSERT_AFTER(head, line, newln, l_list);
	else
		TAILQ_INSERT_HEAD(head, newln, l_list);

	return newln;
}

/* remove a line from a list of lines, and free it */
static __inline void line_remove(linelist_t *head, line_t *line) {
	TAILQ_REMOVE(head, line, l_list);
	free(line->text);
	free(line);
}

/* remove a section of a string from line */
static __inline void line_rmstr(line_t *line, int loc, int slen) {
	assert(loc + slen <= line->length && loc >= 0);

	if (line->length - (loc + slen))
		memmove(&line->text[loc], &line->text[loc + slen],
			line->length - (loc + slen));
	line->length -= slen;
}

/* remove a character from a line */
static __inline void line_rmch(line_t *line, int loc) {
	assert(loc < line->length && line->length && loc >= 0);

	if (line->length - (loc + 1))
		memmove(&line->text[loc], &line->text[loc + 1],
			line->length - (loc + 1));
	line->length--;
}

#endif
