/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* macro flags */
int	macro_playing;
int	macro_repeating;
int	macro_repeatplay;
int	macro_recording;

/* the macro recording buffer */
static int	rsize;
static int	rlength;
static int	rpos;
static int	*record;

/* key to be repeated by the key repeater */
static int	rkey;

/*
 * flag if the keyrepeater is being set up; used to
 * prevent infinite an infinite loop if people do
 * dumb stuff.
 */
static int	gettingrepkey;

/* initialize the macro recording stuff */
void macro_init() {
	record = ckmalloc(LINE_STARTSIZE * sizeof(int));
	rsize = LINE_STARTSIZE;
}

/* free memory */
void macro_shutdown() {
	free(record);
}

/*
 * add a key to the macro recording, and return the key that the
 * calling process should use.
 */
int macro_recordkey(int keysym) {
	if (keysym == KEYSYM_RECORD || keysym == KEYSYM_PLAYBACK) {
		minibuff_printf("Macro recording stopped.");
		macro_recording = 0;
		/*
		 * this is neccesary to prevent an infinite loop if
		 * someone tries to repeat a macro that ends with
		 * an unfinished repeat command.
		 */
		if (gettingrepkey)
			keysym = KEYSYM_CANCEL;
		else
			return KEYSYM_NOP;
	}

	if (rlength + 1 >= rsize) {
		rsize *= 2;
		record = ckrealloc(record, rsize * sizeof(int));
	}
	record[rlength++] = keysym;

	return keysym;
}

/*
 * return a key from the macro system; either from the key repeater
 * or from a playing keyboard macro.
 */
int macro_getkey() {
	int c;

	if (macro_repeating) {
		if (--macro_repeating == 0)
			input_nodraw--;
		return rkey;
	}

	assert(macro_playing && !macro_recording);
	c = record[rpos++];
	if (rpos >= rlength) {
		macro_playing = 0;
		input_nodraw--;
	}
	return c;
}

/* input callback for minibuff repeat prompt */
static int repcback(vminibuff_t *v, int keysym) {
	if (isdigit(keysym) || keysym == KEYSYM_CANCEL)
		return 0;
	else {
		rkey = keysym;
		vminibuff_exit(v);
	}

	return 1;
}

void macro_repeat() {
	u_char *repstr;

	/* figure out what to repeat and how many times */
	repstr = minibuff_cbprompt("Repeat", NULL, repcback, NULL);
	if (!repstr)
		return;

	/*
	 * because macros may want to use the key repeater, if you
	 * try to repeat a macro with the repeater it is neccesary
	 * to use a seperate counter to do it.
	 */
	if (rkey == KEYSYM_PLAYBACK)
		macro_repeatplay = atoi(repstr);
	else
		macro_repeating = atoi(repstr);
	if (macro_repeating || macro_repeatplay)
		input_nodraw++;
	free(repstr);
}

void macro_record() {
	minibuff_printf("Macro recording started.");
	rlength = 0;
	macro_recording = 1;
}

void macro_playback() {
	if (rlength) {
		rpos = 0;
		macro_playing = 1;
		input_nodraw++;
	} else
		minibuff_printf("No macro to play.");
}
