/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

char *argv0;

/* handle child death sigs that the filter command can cause */
static void sigchld_handler(int sig) {
	pid_t pid;
	int status;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0 ||
		(pid < 0 && errno == EINTR));
}

/*
 * entry point; parse arguments, create the first frame and make
 * the first buffer; then hand off to an input_loop invocation.
 */
int main(int argc, char *argv[]) {
	buffer_t *buffer;
	u_char *filename;
	int ch, first, jmpline;

	argv0 = argv[0];

	/* make sure we're on a tty */
	if (!isatty(STDIN_FILENO) || !isatty(STDOUT_FILENO))
		errx(1, "stdin and stdout must be ttys");

	/*
	 * support the trandition +linenumber stuff; this is only carried
	 * out for the first file on the command line.
	 */
	if (argc > 1 && argv[1][0] == '+') {
		jmpline = atoi(argv[1] + 1);
		if (--jmpline < 0)
			jmpline = 0;
		argc--, argv++;
	} else
		jmpline = 0;

	/* process command line arguments */
	while ((ch = getopt(argc, argv, "vh")) != -1)
		switch (ch) {
		case 'v':
			printf("led-" VERSION "\n"
				BUILD_INFO "\n"
				"Copyright (c) 2002 Jordan DeLong\n");
			exit(0);
		case 'h':
		default:
			printf("usage: led [+linenum] [-vh] filename\n");
			exit(ch != 'h');
		}
	argc -= optind;
	argv += optind;

	/*
	 * hook signals
	 */
	signal(SIGCHLD, sigchld_handler);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, SIG_IGN);

	/*
	 * use the proper locale settings
	 */
	setlocale(LC_ALL, "");

	/*
	 * initialize everything
	 */
	options_init();
	screen_init();
	err_set_exit(screen_shutdown);
	keybind_init();
	command_init();
	color_init();
	color_parse();
	view_init();
	minibuff_init();
	macro_init();
	staticmod_init();
	input_init();

	/*
	 * pull in the user's configuration settings
	 */
	rcfile_parse();

	/*
	 * for each filename specified create a frame containing a single
	 * view of a buffer.  if no args are specified, still do the loop
	 * once to get a blank buffer.
	 */
	for (first = 1; --argc >= 0 || first; first = 0) {
		if (first && argc < 0)
			filename = NULL;
		else
			filename = argv[argc];

		/*
		 * attempt to load the file; default to an empty buffer
		 * if the file can't be opened.
		 */
		if (filename) {
			buffer = file_read(filename);
			if (buffer == NULL)
				goto emptybuf;
		} else {
			filename = "newfile";
emptybuf:
			buffer = buffer_create(filename, 0);
		}

		/*
		 * create the frame, and make a view to put into it.
		 * stack each of the user's autoviews over the vdefault.
		 */
		{
			viewtype_t *type;
			viewhdr_t *v;
			char *s, *sep, *vt;

			v = view_create(view_findtype("vdefault"), NULL);
			if (options.autoviews) {
				sep = s = ckstrdup(options.autoviews);
				while ((vt = strsep(&sep, ":")) != NULL) {
					type = view_findtype(vt);
					if (!type)
						continue;
					v = view_create(type, v);
				}
				free(s);
			}
			frame_create(v, buffer);
		}
	}

	/*
	 * jump the cursor to the linenumber specified on the command
	 * line, if it was specified.
 	 */
	if (jmpline) {
		vdefault_t *vdef;

		vdef = view_findvdef(view_current);
		vdef->cursor.line = buffer_lineat(buffer,
			jmpline, &vdef->cursor.linenum);
		vdefault_checkscroll(vdef);
	}

	/*
	 * main program loop; when the user exits the loop, give them
	 * a chance to save any modified buffers.
	 */
	do
		input_loop();
	while (command_savebuffers());

	/*
	 * shutdown everything; these will remove the frames, views and
	 * buffer we created in this routine, along with any others
	 * that are created later.
	 */
	minibuff_shutdown();
	macro_shutdown();
	frame_shutdown();
	module_shutdown();
	staticmod_shutdown();
	view_shutdown();
	command_shutdown();
	keybind_shutdown();
	color_shutdown();
	screen_shutdown();
	options_shutdown();

	/* free the killring */
	while (!TAILQ_EMPTY(&killring.buffers)) {
		killbuffer_t *killbuff;

		killbuff = TAILQ_FIRST(&killring.buffers);
		while (!TAILQ_EMPTY(&killbuff->lines))
			line_remove(&killbuff->lines,
				TAILQ_FIRST(&killbuff->lines));
		TAILQ_REMOVE(&killring.buffers, killbuff, kb_list);
		free(killbuff);
	}

	/* check for buffer reference leaks */
	assert(TAILQ_EMPTY(&buffer_list));

	/*
	 * on terminals that don't support ti/te (smcup/rmcup) capabilities
	 * it is useful to print an extra \n to prevent messing up the
	 * user's shell prompt.
	 */
	if (options.printnlatend)
		printf("\n");

	return 0;
}
