/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* flag that input goes to the minibuffer */
int			minibuff_editing;

/*
 * flag indicating we are actively in the minibuffer; this, when set,
 * only affects the order in which things are drawn so the user sees
 * the terminal's cursor on the minibuffer line.  so for character-at-a-time
 * style minibuff prompts it shouldn't be set.
 */
int			minibuff_active;

/* the minibuffer line; useful for callback-based prompts */
line_t			*minibuff_line;

/* the minibuffer itself */
static vminibuff_t	*minibuff_view;
static buffer_t		*minibuff;

void minibuff_init() {
	viewhdr_t *v;

	/* create a view for it */
	minibuff = buffer_create("minibuffer", 1);
	minibuff_line = TAILQ_FIRST(&minibuff->lines);
	v = view_create(view_findtype("vdefault"), NULL);
	v = view_create(view_findtype("vminibuff"), v);
	view_mapstack(v, minibuff, 0, screen_height - 1, screen_width, 1);
	minibuff_view = (vminibuff_t *) v;
}

void minibuff_shutdown() {
	viewhdr_t *v;

	v = (viewhdr_t *) minibuff_view;
	while ((v = view_destroy(v)) != NULL)
		;
}

void minibuff_draw() {
	VIEWCALL((viewhdr_t *) minibuff_view, draw);
}

void minibuff_forcedraw() {
	if (minibuff_view->state == MINIBUFF_NORMAL)
		minibuff_view->state = MINIBUFF_NEEDREDRAW;
}

void minibuff_resize(int width, int height) {
	view_resize((viewhdr_t *) minibuff_view, 0, height - 1, width, 1);
}

void minibuff_input(int keysym) {
	VIEWCALL((viewhdr_t *) minibuff_view, input, keysym);
}

void minibuff_print(u_char *str) {
	line_t *line;

	line = TAILQ_FIRST(&minibuff_view->hdr.buffer->lines);
	line->length = 0;
	line_nputs(line, 0, str, strlen(str));
	minibuff_view->state = MINIBUFF_DRAWMSG;
}

void minibuff_printf(u_char *fmt, ...) {
	u_char buff[minibuff_view->hdr.geom.width];
	va_list args;

	/* don't do anything if the minibuff is prompting */
	if (minibuff_view->state == MINIBUFF_PROMPT)
		return;

	/* fill the buffer with the formatted output */
	va_start(args, fmt);
	vsnprintf(buff, sizeof(buff), fmt, args);
	va_end(args);

	minibuff_print(buff);
}

static u_char *doprompt(u_char *prompt, u_char *def) {
	u_char *s;

	/* enter the minibuffer */
	minibuff_editing = 1;
	vminibuff_prepare(minibuff_view, prompt, def);
	vminibuff_enter(minibuff_view);
	minibuff_editing = 0;
	minibuff_active = 0;
	if (minibuff_view->canceled) {
		minibuff_printf("Canceled.");
		return NULL;
	}

	/* allocate a copy of what was typed to return */
	s = ckmalloc(minibuff_line->length + 1);
	memcpy(s, minibuff_line->text, minibuff_line->length);
	s[minibuff_line->length] = '\0';

	return s;
}

u_char *minibuff_prompt(u_char *prompt, u_char *def) {
	minibuff_active = 1;
	return doprompt(prompt, def);
}

/* stuff for chcback */
static u_char	*chcback_str;
static int	chcback_ret;
static int	chcback_allowcancel;

/* input callback for chprompt */
static int chcback(vminibuff_t *v, int keysym) {
	if ((keysym <= UCHAR_MAX && strchr(chcback_str, keysym)) ||
			(chcback_allowcancel && keysym == KEYSYM_CANCEL)) {
		chcback_ret = keysym;
		vminibuff_exit(v);
		return 1;
	}
	screen_bell();

	return 1;
}

/*
 * prompt the user to press any of the characters in str,
 * and return the one they pressed.
 */
int minibuff_chprompt(u_char *prompt, u_char *str, int allowcancel) {
	u_char *fullprompt;

	fullprompt = ckmalloc(strlen(prompt) + strlen(str) + 5);
	strcpy(fullprompt, prompt);
	strcat(fullprompt, "? (");
	strcat(fullprompt, str);
	strcat(fullprompt, ")");

	chcback_allowcancel = allowcancel;
	chcback_str = str;
	free(minibuff_cbprompt(fullprompt, NULL, chcback, NULL));
	free(fullprompt);
	if (chcback_ret == KEYSYM_CANCEL)
		minibuff_printf("Canceled.");
	return chcback_ret;
}

/* prompting with an input handler callback */
u_char *minibuff_cbprompt(u_char *prompt, u_char *def,
		int (*callback)(vminibuff_t *, int),
		void (*postcallback)(vminibuff_t *, int)) {
	minibuff_view->inputcback = callback;
	minibuff_view->postinputcback = postcallback;
	return doprompt(prompt, def);
}

/* prepare for minibuff_rawenter calls */
void minibuff_rawinit(u_char *prompt, u_char *def,
		int (*callback)(vminibuff_t *, int),
		void (*postcallback)(vminibuff_t *, int)) {
	minibuff_view->inputcback = callback;
	minibuff_view->postinputcback = postcallback;
	vminibuff_prepare(minibuff_view, prompt, def);
}

/*
 * callback prompt but the current state of the line is accessable
 * in minibuff_line instead of as a return value.  this routine
 * uses minibuffer reentry, so it is neccesary to make a call to
 * minibuff_rawinit() before the first entry.
 */
void minibuff_rawenter() {
	vdefault_t *vdef;

	/*
	 * flag inhibit paren flashing; paren flashing just looks
	 * like an unexplained delay when the minbuffer isn't the
	 * active view and a ')' is sent to it's input operation.
	 */
	vdef = view_findvdef((viewhdr_t *) minibuff_view);
	vdef->inhibit_parenflash = 1;

	minibuff_editing = 1;
	vminibuff_enter(minibuff_view);
	minibuff_editing = 0;

	vdef->inhibit_parenflash = 0;
}

/* change the prompt in an active minibuffer (raw editing) */
void minibuff_setprompt(u_char *prompt) {
	minibuff_view->prompt = prompt;
	vminibuff_hidesz(minibuff_view);
}

u_char *minibuff_fileprompt(u_char *def) {
	minibuff_view->compfunc = complete_filename;
	return minibuff_prompt("Filename", def);
}

u_char *minibuff_cmdprompt(u_char *def) {
	minibuff_view->compfunc = complete_command;
	return minibuff_prompt("Command", def);
}

u_char *minibuff_vtypeprompt(u_char *def) {
	minibuff_view->compfunc = complete_vtype;
	return minibuff_prompt("Viewtype", def);
}

u_char *minibuff_optprompt(u_char *def) {
	minibuff_view->compfunc = complete_opt;
	return minibuff_prompt("Option", def);
}

u_char *minibuff_buffprompt(u_char *def) {
	minibuff_view->compfunc = complete_buff;
	return minibuff_prompt("Buffer", def);
}
