/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"
#include <dlfcn.h>

static LIST_HEAD(, module) module_list = LIST_HEAD_INITIALIZER(module_list);

static void module_unload(module_t *module) {
	u_char symname[512];
	void (*shutdown)();

	/* call the module's shutdown routine */
	if (strlcpy(symname, SYM_PREFIX, sizeof(symname)) >= sizeof(symname))
		goto toobad;
	if (strlcat(symname, module->name, sizeof(symname)) >= sizeof(symname))
		goto toobad;
	if (strlcat(symname, "_shutdown", sizeof(symname)) >= sizeof(symname))
		goto toobad;
	shutdown = dlsym(module->hnd, symname);
	if (shutdown)
		shutdown();

toobad:
#ifndef USE_DMALLOC
	dlclose(module->hnd);
#endif
	LIST_REMOVE(module, m_list);
	free(module->name);
	free(module);
}

void module_shutdown() {
	while (!LIST_EMPTY(&module_list))
		module_unload(LIST_FIRST(&module_list));
}

static __inline void *loadmod(u_char *name) {
	u_char fullpath[PATH_MAX];
	char *modpath, *tmp;
	char *path, *expath;
	void *hnd;

	modpath = ckstrdup(MODULE_PATH);
	tmp = modpath;

	hnd = NULL;
	while ((path = strsep(&tmp, ":")) != NULL && !hnd) {
		expath = tilde_expand(path);

		if (snprintf(fullpath, sizeof(fullpath), "%s/%s.so", expath,
				name) >= sizeof(fullpath)) {
			warnx("path to module exceeds PATH_MAX");
			goto next;
		}

		/*
		 * the only way to determine why dlopen failed is by
		 * calling dlerror; so try an access first so we don't
		 * print dlerrors for nonexistant files.
		 */
		if (access(fullpath, R_OK) != 0)
			goto next;

		hnd = dlopen(fullpath, RTLD_LAZY);
		if (!hnd)
			warnx("failed dlopen for %s: %s", fullpath, dlerror());
next:
		free(expath);
	}
	free(modpath);

	return hnd;
}

module_t *module_load(u_char *name) {
	u_char symname[512];
	module_t *module;
	int (*init)();
	void *hnd;

	hnd = loadmod(name);
	if (!hnd)
		return NULL;

	/* call the init routine, bail if it returns -1 */
	if (strlcpy(symname, SYM_PREFIX, sizeof(symname)) >= sizeof(symname))
		goto bail;
	if (strlcat(symname, name, sizeof(symname)) >= sizeof(symname))
		goto bail;
	if (strlcat(symname, "_init", sizeof(symname)) >= sizeof(symname))
		goto bail;
	init = dlsym(hnd, symname);
	if (init && init() == -1)
		goto bail;

	/* add the module to the list of loaded modules */
	module = ckmalloc(sizeof(module_t));
	module->name = ckstrdup(name);
	module->hnd = hnd;
	LIST_INSERT_HEAD(&module_list, module, m_list);

	return module;

bail:
	dlclose(hnd);
	return NULL;
}
