/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "perlapi.h"
#include "XSUB.h"

MODULE = Led		PACKAGE = Led

#
# register a command that will call into the perl
# sub of the supplied name; the command will be named
# after the sub with a 'perl_' prefix.
#
void
AddCommand(subname)
	char *subname

	PROTOTYPE: $
	PREINIT:
	char cmdname[strlen(subname) + strlen("perl_") + 1];

	CODE:
	strcpy(cmdname, "perl_");
	strcat(cmdname, subname);
	command_addf(cmdname, input_allocsyms(1), perlapi_cmdf);

#
# most perl scripts that do more than a single edit to a buffer
# should be calling these to make the operation undoable.  it
# should be noted that keys that are pushed are evaluated later, so
# they wont be included in an Undo unless they are used properly.
#
void
BeginUndo()

	CODE:
	undo_beginblock(view_findvdef(view_current)->hdr.buffer);

void
EndUndo()

	CODE:
	undo_endblock(view_findvdef(view_current)->hdr.buffer);

#
# send a command to the view.
#
void
Execute(command)
	char *command

	PROTOTYPE: $
	PREINIT:
	int sym;

	CODE:
	sym = command_getkeysym(command);
	if (sym != KEYSYM_NOP)
		input_dispatch(sym);

#
# equivalent of Execute for keys -- often more convienent than the other
# routines to modify text.
#
void
ExecuteStr(text)
	char *text

	PROTOTYPE: $
	PREINIT:
	STRLEN len;
	int i;

	CODE:
	SvPV(ST(0), len);
	for (i = 0; i < len; i++)
		input_dispatch(text[i]);

#
# pass some characters to the input handler of the view; if the script
# wants to reliably modify the buffer (i.e., regardless of what type of
# view we're looking at), it should use GetLine/SetLine.  The purpose
# of this is to allow a script to cause input to be readable to a command
# handler; this is acomplished by pushing the input string (or pushing a
# command with PushCommand) and then calling Execute().  The stuff works
# like a stack (hence the naming); so the thing last pushed is the first
# thing read.
#
# the string is pushed in reverse order for convienence.
#
void
PushKeys(keys)
	char *keys

	PROTOTYPE: $
	PREINIT:
	STRLEN len;
	int i;

	CODE:
	SvPV(ST(0), len);
	for (i = len - 1; i >= 0; i--)
		keybind_ungetkey(keys[i]);

void
PushCommand(name)
	char *name

	PROTOTYPE: $
	PREINIT:
	int sym;

	CODE:
	sym = command_getkeysym(name);
	if (sym != KEYSYM_NOP)
		keybind_ungetkey(sym);

#
# print something to the minibuffer output area
#
void
Echo(string)
	char *string

	PROTOTYPE: $

	CODE:
	minibuff_printf(string);

#
# return the position of the cursor in the active view's
# vdefault as a linenum,column list.
#
void
GetCursor()

	PREINIT:
	vdefault_t *vdef;

	PPCODE:
	vdef = view_findvdef(view_current);
	XPUSHs(sv_2mortal(newSViv(vdef->cursor.linenum)));
	XPUSHs(sv_2mortal(newSViv(vdef->cursor.col)));

#
# set the cursor position
#
void
SetCursor(linenum, col)
	int linenum
	int col

	PROTOTYPE: $;$
	PREINIT:
	vdefault_t *vdef;
	line_t *line;
	int num;

	CODE:
	vdef = view_findvdef(view_current);
	line = buffer_lineat(vdef->hdr.buffer, linenum, &num);
	if (num == linenum) {
		vdef->cursor.line = line;
		vdef->cursor.linenum = linenum;
	}
	if (col >= 0 && col <= vdef->cursor.line->length)
		vdef->cursor.col = col;
	vdefault_setscreencol(vdef);
	vdefault_checkscroll(vdef);
	vdef->redraw_header = 1;

#
# return the location of the mark; ($line, $col)
#
void
GetMark()

	PREINIT:
	vdefault_t *vdef;
	int l, c;

	PPCODE:
	vdef = view_findvdef(view_current);
	if (vdef->hdr.buffer->mark) {
		l = vdef->hdr.buffer->mark->linenum;
		c = vdef->hdr.buffer->mark->col;
	} else
		l = c = 0;
	XPUSHs(sv_2mortal(newSViv(l)));
	XPUSHs(sv_2mortal(newSViv(c)));

#
# push a mark at the specified position; more convienent than
# doing SetCursor and Execute("pushmark").  mark poppage should
# be done with Execute("popmark")
#
void
PushMark(linenum, col)
	int linenum
	int col

	PREINIT:
	vdefault_t *vdef;
	line_t *line;
	int num;

	CODE:
	vdef = view_findvdef(view_current);
	line = buffer_lineat(vdef->hdr.buffer, linenum, &num);
	if (col < 0)
		col = 0;
	else if (col > line->length)
		col = line->length;
	buffer_pushmark(vdef->hdr.buffer, line, linenum, col);

#
# return the number of lines in the current view's buffer
#
int
LineCount()

	CODE:
	RETVAL = view_findvdef(view_current)->hdr.buffer->linecnt;

	OUTPUT:
	RETVAL

#
# utility for inserting a string into a line at a column; usually used
# when wanting to put some text where the cursor is.  (more convienent
# than using GetLine/SetLine together).
#
void
InsertText(linenum, col, text)
	int linenum
	int col
	char *text

	PROTOTYPE: $;$;$
	PREINIT:
	vdefault_t *vdef;
	line_t *line;
	int realnum;
	STRLEN len;

	CODE:
	SvPV(ST(2), len);
	vdef = view_findvdef(view_current);
	line = buffer_lineat(vdef->hdr.buffer, linenum, &realnum);
	buffer_nputs(vdef->hdr.buffer, 1, line, realnum,
		col <= line->length ? col : 0, text, len);

#
# get text for a line
#
char *
GetLine(linenum)
	int linenum

	PROTOTYPE: $
	PREINIT:
	vdefault_t *vdef;
	line_t *line;
	SV *sv;

	PPCODE:
	vdef = view_findvdef(view_current);
	line = buffer_lineat(vdef->hdr.buffer, linenum, NULL);
	sv = line->length ? newSVpv(line->text, line->length) : newSVpv("", 0);
	XPUSHs(sv_2mortal(sv));

#
# set the text for a line
#
void
SetLine(linenum, text)
	int linenum
	char *text

	PROTOTYPE: $;$
	PREINIT:
	vdefault_t *vdef;
	line_t *line;
	STRLEN len;
	int num;

	CODE:
	SvPV(ST(1), len);
	vdef = view_findvdef(view_current);
	line = buffer_lineat(vdef->hdr.buffer, linenum, &num);
	undo_beginblock(vdef->hdr.buffer);
	buffer_rmstr(vdef->hdr.buffer, 1, line, num, 0, line->length);
	buffer_nputs(vdef->hdr.buffer, 1, line, num, 0, text, len);
	undo_endblock(vdef->hdr.buffer);

#
# interactively prompt the user to type a string of text; the string
# is placed into a scalar and returned to the perl script.  if the
# user cancels undef is returned.
#
char *
Prompt(prompt, def)
	char *prompt
	char *def

	PROTOTYPE: $;$
	PREINIT:
	u_char *str;
	SV *sv;

	PPCODE:
	input_nodraw--;
	str = minibuff_prompt(prompt, def);
	if (str) {
		sv = sv_2mortal(newSVpv(str, 0));
		free(str);
	} else
		sv = &PL_sv_undef;
	input_nodraw++;
	XPUSHs(sv);

#
# cause a redraw; normally redraws wont occur during the execution
# of the script.  most scripts probably wont redraw the screen.
#
void
Redraw()

	CODE:
	draw_screen();
	screen_refresh();

#
# allow the perl script to do some printing on the screen, or run
# other programs as children, and so forth.  the script _must_
# call resume after suspend.
#
void
Suspend()

	CODE:
	screen_clear();
	screen_refresh();
	screen_suspend();

void
Resume()

	CODE:
	screen_clear();
	screen_reinit();
	frame_forcedraw(frame_active);
	minibuff_forcedraw();
