/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "perlapi.h"

PerlInterpreter		*perlinterp;

extern void boot_DynaLoader(CV *);
extern void boot_Led(CV *);

/* extension module support */
static void xs_init() {
	char *file = __FILE__;

	newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
	newXS("Led::bootstrap", boot_Led, file);
}

/* command allowing the user to evaluate some perl */
static void perlapi_doeval(command_t *command) {
	u_char *text;

	text = minibuff_prompt("Perl", NULL);
	if (!text)
		return;
	input_nodraw++;
	perlapi_eval(text);
	input_nodraw--;
	free(text);
}

/* command to reload the perl interpreter and scripts */
static void perlapi_reload() {
	perlapi_shutdown();
	perlapi_init();
	minibuff_printf("Reloaded perl interpreter and scripts.");
}

int perlapi_init() {
	char str[2 * PATH_MAX];			/* arbitrary size */
	char *args[] = { "", "-e", "" };
	char *bootargs[] = { "Led", 0 };
	u_char *expath;

	/* add perl commands to the editor */
	command_addf("perl_eval", input_allocsyms(1), perlapi_doeval);
	command_addf("perl_reload", input_allocsyms(1), perlapi_reload);

	/* construct the perl interpreter */
	perlinterp = perl_alloc();
	perl_construct(perlinterp);
	if (perl_parse(perlinterp, xs_init, 3, args, NULL)) {
		perlapi_shutdown();
		return 1;
	}
	perl_call_argv("Led::bootstrap", G_DISCARD, bootargs);

	/* add the script search path to the @INC list */
	expath = tilde_expand(SCRIPT_PATH);
	if (snprintf(str, sizeof(str), "push @INC, split(/:/, \"%s\");",
			expath) < sizeof(str))
		perlapi_eval(str);
	free(expath);

	/*
	 * read in autorun.pl, if possible.  this allows the user to
	 * require any other perl scripts into the interpreter that
	 * they want.
	 */
	perlapi_eval("do \"autorun.pl\"");

	return 0;
}

void perlapi_shutdown() {
	command_t *c, *next;
	int idx;

	perl_destruct(perlinterp);
	perl_free(perlinterp);
	command_remove("perl_eval");

	/*
	 * easiest way to remove all our commands is to remove any
	 * function commands who's handlers are perlapi_cmdf.
	 */
	for (idx = 0; idx < COMMAND_BUCKETS; idx++) {
		c = LIST_FIRST(&command_hash[idx]);
		while (c) {
			next = LIST_NEXT(c, c_list);
			if (c->type == COMMAND_FUNC &&
					c->handler == perlapi_cmdf)
				command_remove(c->name);
			c = next;
		}
	}
}

/* evaluate some perl */
void perlapi_eval(u_char *string) {
	SV *sv;

	sv = newSVpv(string, 0);
	perl_eval_sv(sv, G_DISCARD | G_NOARGS);
	SvREFCNT_dec(sv);
}

/*
 * run the perl sub with name corresponding to the name of this
 * command.
 */
void perlapi_cmdf(command_t *command) {
	char *sub;

	sub = strchr(command->name, '_') + 1;
	input_nodraw++;
	perlapi_eval(sub);
	input_nodraw--;
}
