/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vfolding.h"

/* the color to use for drawing folded lines */
static int	color_folded;

/* first in the block of keysyms we're using */
int	vfolding_keybase;

/* markers to use for delim of folds */
u_char	*vfolding_begmark;
u_char	*vfolding_endmark;

static void vfolding_draw(vfolding_t *v) {
	fold_t *fold;
	int y;

	draw_standard(v->hdr.top, v->vdef);

	/* place the cursor, skipping folds */
	y = v->vdef->cursor.linenum - v->vdef->view.linenum;
	fold = fold_findafter(&v->folds, v->vdef->view.linenum);
	if (fold && fold->start < v->vdef->view.linenum)
		fold = TAILQ_NEXT(fold, f_list);		/* XXX */
	while (fold) {
		if (fold->start >= v->vdef->cursor.linenum)
			break;
		y -= min(v->vdef->cursor.linenum, fold->end) - fold->start;
		fold = TAILQ_NEXT(fold, f_list);
	}
	screen_gotoxy(v->hdr.geom.x + v->vdef->cursor.screencol -
		v->vdef->view.col, v->hdr.geom.y + y);
}

static void vfolding_drawpart(vfolding_t *v, line_t **line, int *linenum,
		int *col, int maxcols, int *y, int *viewcol) {
	fold_t *fold;

	/*
	 * it's impossible for us to properly fold text if a view above
	 * passing us chunks of lines.
	 * TODO: this can be alievated by some semantic changes to
	 * drawpart; will do eventually.
	 *
	 * XXX: when an explicit mark exists this case needs to be
	 * muted somehow....dunno how to deal with that yet.
	 */
	if (*col != 0 || maxcols != (*line)->length) {
		minibuff_printf("Incompatible configuration for "
			"vfolding; see docs");
		VIEWCALL(v->hdr.under, drawpart, line, linenum,
			col, maxcols, y, viewcol);
		return;
	}

	fold = fold_findafter(&v->folds, *linenum);
	if (fold && fold->start == *linenum) {
		screen_setcolor(color_folded);
		draw_string((*line)->text, v->hdr.geom.width,
			(*line)->length, view_findvdef((viewhdr_t *) v)->
			view.col, viewcol);
		draw_eol(v->vdef, *viewcol, *y);
		while (*linenum < fold->end) {
			*line = TAILQ_NEXT(*line, l_list);
			++*linenum;
		}
		*col = (*line)->length;
	} else
		VIEWCALL(v->hdr.under, drawpart, line, linenum,
			col, maxcols, y, viewcol);
}

static void vfolding_map(vfolding_t *v) {
	TAILQ_INIT(&v->folds);
	v->vdef = view_findvdef((viewhdr_t *) v);
}

/* {{{ vfolding_create */
static viewhdr_t *vfolding_create(viewhdr_t *under) {
	vfolding_t *vfolding;

	vfolding = VIEW_CREATE(vfolding_t, "vfolding", under);

	vfolding->hdr.op.draw		= VIEWOP(vfolding_draw);
	vfolding->hdr.op.drawpart	= VIEWOP(vfolding_drawpart);
	vfolding->hdr.op.viswid		= NULL;
	vfolding->hdr.op.drawhdr	= NULL;
	vfolding->hdr.op.input		= VIEWOP(vfolding_input);
	vfolding->hdr.op.map		= VIEWOP(vfolding_map);
	vfolding->hdr.op.geomchange	= NULL;
	vfolding->hdr.op.activate	= NULL;
	vfolding->hdr.op.rmline		= NULL;	/* TODO */
	vfolding->hdr.op.rmstr		= NULL;
	vfolding->hdr.op.addline	= NULL;
	vfolding->hdr.op.stradd		= NULL;
	vfolding->hdr.op.destroy	= VIEWOP(free);

	return (viewhdr_t *) vfolding;
}
/* }}} */

/* {{{ vfolding module initialization and shutdown */
static viewtypeops_t vfolding_typeops = {
	vfolding_create
};

static __inline void addkeys() {
	vfolding_keybase = input_allocsyms(VFNUMSYMS);

	command_add("vfolding_fold",	KEYSYM_FOLD);
}

int vfolding_init() {
	options_add("vfolding_color", OPT_COLOR, &color_folded, 1);
	options_add("vfolding_begmark", OPT_STRING, &vfolding_begmark, 1);
	options_add("vfolding_endmark", OPT_STRING, &vfolding_endmark, 1);
	vfolding_begmark = ckstrdup("{{{");
	vfolding_endmark = ckstrdup("}}}");
	color_folded = color_lookup("folded");
	if (color_folded == COLOR_NULL)
		color_folded = color_add("folded", "black", "white",
			SCREEN_A_NORM, SCREEN_A_STAND);
	addkeys();
	VIEW_ADDTYPE(vfolding_typeops, "vfolding");
	return 0;
}

void vfolding_shutdown() {
	options_rm("vfolding_color");
	view_rmtype("vfolding");
}
/* }}} */
