/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vhighlight.h"

/*
 * anchors not in use are placed on the free list for reuse to
 * prevent the need to constantly allocate them.  anchors are
 * never deallocated, just moved to this list.
 */
static anchorlist_t anchor_freelist = TAILQ_HEAD_INITIALIZER(anchor_freelist);

static __inline anchor_t *anchor_get() {
	anchor_t *anchor;

	anchor = TAILQ_FIRST(&anchor_freelist);
	if (!anchor)
		return ckmalloc(sizeof(anchor_t));
	TAILQ_REMOVE(&anchor_freelist, anchor, a_list);

	return anchor;
}

void anchor_shutdown() {
	anchor_t *anchor;

	while (!TAILQ_EMPTY(&anchor_freelist)) {
		anchor = TAILQ_FIRST(&anchor_freelist);
		TAILQ_REMOVE(&anchor_freelist, anchor, a_list);
		free(anchor);
	}
}

anchor_t *anchor_add(anchorlist_t *list, line_t *line,
		int linenum, synitem_t *state) {
	anchor_t *anchor;

	anchor = anchor_get();
	anchor->line = line;
	anchor->linenum = linenum;
	anchor->state = state;
	TAILQ_INSERT_TAIL(list, anchor, a_list);

	return anchor;
}

/* move an anchor to the freelist */
void anchor_remove(anchorlist_t *list, anchor_t *anchor) {
	TAILQ_REMOVE(list, anchor, a_list);
	TAILQ_INSERT_HEAD(&anchor_freelist, anchor, a_list);
}

/*
 * find the last anchor before linenum, and remove all the
 * ones after it (they aren't valid anymore).
 */
anchor_t *anchor_findlast(anchorlist_t *list, int linenum) {
	anchor_t *anchor;

	while (!TAILQ_EMPTY(list)) {
		anchor = TAILQ_LAST(list, anchorlist);
		if (anchor->linenum <= linenum)
			return anchor;
		anchor_remove(list, anchor);
	}

	return NULL;
}
