/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vhighlight.h"

/* syntax hash from language to syntax_t */
#define SYNTAX_BUCKETS	4
syntaxlist_t *syntax_list;

/* initialize the list of syntax highlighters */
void syntax_init() {
	syntax_list = hash_alloc("syntax_list", SYNTAX_BUCKETS,
		offsetof(syntax_t, s_list));
}

/* free our resources */
void syntax_shutdown() {
	syntax_t *syntax;
	synitem_t *item;
	int idx, i;

	for (idx = 0; idx < SYNTAX_BUCKETS; idx++)
		while (!LIST_EMPTY(&syntax_list[idx])) {
			syntax = LIST_FIRST(&syntax_list[idx]);
			LIST_REMOVE(syntax, s_list);

			while (!TAILQ_EMPTY(&syntax->items)) {
				item = TAILQ_FIRST(&syntax->items);
				TAILQ_REMOVE(&syntax->items, item, si_list);

				switch (item->type) {
				case SYNITEM_REGEXP:
					regfree(item->u.re);
					free(item->u.re);
					break;
				case SYNITEM_TOKENS:
				case SYNITEM_WORDS:
					for (i = 0; item->u.wordlist[i]; i++)
						free(item->u.wordlist[i]);
					free(item->u.wordlist);
					break;
				case SYNITEM_BLOCK:
					for (i = 0; item->u.blk.end[i]; i++)
						free(item->u.blk.end[i]);
					free(item->u.blk.end);
				case SYNITEM_TOEOL:
					for (i = 0; item->u.blk.begin[i]; i++)
						free(item->u.blk.begin[i]);
					free(item->u.blk.begin);
					break;
				}
				free(item);
			}
			free(syntax->lang);
			free(syntax);
		}

	hash_free(syntax_list);
}

/* find a syntax definition for a particular language */
syntax_t *syntax_findlang(u_char *lang) {
	syntax_t *syntax;
	int idx;

	/* try to find it in the hash */
	idx = hash_string(SYNTAX_BUCKETS, lang);
	LIST_FOREACH(syntax, &syntax_list[idx], s_list)
		if (!strcmp(lang, syntax->lang))
			return syntax;

	/* we couldn't find it, so try to load it now */
	syntax = syntax_load(lang);
	if (!syntax)
		return NULL;
	LIST_INSERT_HEAD(&syntax_list[idx], syntax, s_list);

	return syntax;
}

/* check for matching strings in an array of words at idx */
static int ckword(line_t *line, int *idx, u_char **words) {
	int find_i, find_w;
	int i, w;

	find_i = find_w = -1;
	i = w = 0;
	for (; words[w] != NULL; w++)
		for (i = 0;; i++) {
			if (words[w][i] == '\0') {
				find_i = i;
				find_w = w;
				break;
			}
			if (*idx + i == line->length) {
				if (find_i != -1)
					goto end;
				break;
			}
			if (words[w][i] > line->text[*idx + i])
				goto end;
			else if (words[w][i] != line->text[*idx + i])
				break;
		}

end:
	if (find_i != -1)
		*idx += find_i;
	return find_w;
}

/* look for a specific word anywhere after idx on line */
static int findword(line_t *line, int *idx, u_char *word) {
	u_char *s;

	s = strnstr(&line->text[*idx], word, line->length - *idx);
	if (s) {
		*idx += s - &line->text[*idx] + strlen(word);
		return 1;
	}

	return 0;
}

/* look for a set of words anywhere after idx on line */
static int findwords(line_t *line, int *idx, u_char **words) {
	int i;

	for (i = 0; words[i]; i++)
		if (findword(line, idx, words[i]))
			return 1;

	return 0;
}

/* eat identifier characters in text */
static int eatident(line_t *line, int *idx) {
	int firstidx;

	firstidx = *idx;
	while (*idx < line->length && (isalnum(line->text[*idx]) ||
			line->text[*idx] == '_'))
		(*idx)++;

	return firstidx != *idx;
}

/* eat numeric constants */
static int eatnum(line_t *line, int *idx) {
	int firstidx;
	int inxdigit;
	int decimal;

	decimal = inxdigit = 0;
	firstidx = *idx;

	/* check for hex constant */
	if (line->text[*idx] == '0') {
		if (++(*idx) >= line->length)
			return 1;
		if (line->text[*idx] == 'x') {
			inxdigit = 1;
			(*idx)++;
		}
	}

	/* eat as many digits as possible */
	while (*idx < line->length && (isdigit(line->text[*idx]) ||
			line->text[*idx] == '.' ||
			(inxdigit && tolower(line->text[*idx]) >= 'a' &&
			tolower(line->text[*idx]) <= 'f'))) {
		if (line->text[*idx] == '.') {
			if (firstidx == *idx || inxdigit || decimal)
				break;
			decimal = 1;
		}
		(*idx)++;
	}

	if (firstidx == *idx)
		return 0;

	/* if this isn't a hex constant, support suffixes */
	if (!inxdigit && (*idx < line->length &&
			(tolower(line->text[*idx]) == 'f' ||
			tolower(line->text[*idx]) == 'l')))
		(*idx)++;

	return 1;
}

/*
 * look through the text and match it against a syntax definition,
 * return the color to use for each chunk.
 */
int syntax_highlight(syntax_t *syntax, line_t *line, int *idx,
		synitem_t **state) {
	synitem_t *item;
	int startidx;

	/* handle starting up in a particular state */
	if (*state) {
		item = *state;
		switch (item->type) {
		case SYNITEM_BLOCK:
			if (findwords(line, idx, item->u.blk.end))
				*state = NULL;
			else
				*idx = line->length;
			return item->color;
		case SYNITEM_STRING:
			for (; *idx < line->length; (*idx)++)
				if (line->text[*idx] == item->u.str.quote) {
					if (*idx + 1 < line->length)
						(*idx)++;
				} else if (line->text[*idx] ==
						item->u.str.tok) {
					(*idx)++;
					*state = NULL;
					break;
				}
			return item->color;
		default:
			assert(0);
		}
	}

	/* try to match syntax items against the text */
	TAILQ_FOREACH(item, &syntax->items, si_list) {
		startidx = *idx;

		switch (item->type) {
		case SYNITEM_REGEXP: {
			regmatch_t match;
			int flags;

/*
 * REG_STARTEND is an extension to posix regexp found in some
 * implementations (such as henry spencer's) which allows using
 * non null-terminated strings.  If we're on a platform which
 * doesn't have it, we have to make sure our string has a '\0'
 * on then end.
 */
#ifdef REG_STARTEND
			match.rm_so = 0;
			match.rm_eo = line->length - *idx;
			flags = REG_STARTEND;
#else
			flags = 0;

			/* XXX */
			if (line->size == line->length) {
				line->size *= 2;
				line->text = ckrealloc(line->text,
					line->size);
			}
			if (line->text[line->length] != '\0')
				line->text[line->length] = '\0';
#endif

			if (*idx != 0)
				flags |= REG_NOTBOL;
			if (regexec(item->u.re, line->text + *idx, 1, &match,
					flags))
				continue;
			if (match.rm_so != 0)
				continue;
			*idx += match.rm_eo;
			return item->color;
		}
		case SYNITEM_WORDS:
		case SYNITEM_TOKENS:
			if (ckword(line, idx, item->u.wordlist) != -1) {
				if (item->type == SYNITEM_WORDS) {
					/*
					 * words differ from tokens for our
					 * syntax highlighting in that they
					 * aren't considered to be found if
					 * we can eat an identifier instead;
					 * try that first, and just continue
					 * in that case.
					 */
					if (eatident(line, idx)) {
						*idx = startidx;
						continue;
					}
				}
				return item->color;
			}
			break;
		case SYNITEM_BLOCK:
			if (ckword(line, idx, item->u.blk.begin) != -1) {
				*state = item;
				return item->color;
			}
			break;
		case SYNITEM_TOEOL:
			if (ckword(line, idx, item->u.wordlist) != -1) {
				*idx = line->length;
				return item->color;
			}
			break;
		case SYNITEM_STRING:
			if (line->text[*idx] == item->u.str.tok) {
				(*idx)++;
				*state = item;
				return item->color;
			}
			break;
		}
	}

	/* highlight constant numbers if this syntax wants this */
	if (syntax->hlnums && eatnum(line, idx))
		return vhighlight_numberclr;

	/* didn't match any words */
	if (eatident(line, idx))
		return COLOR_DEFAULT;

	/* eat any other single characters */
	(*idx)++;
	return COLOR_DEFAULT;
}
