%{
/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vhighlight.h"

extern int	syntax_lineno;
extern FILE	*syntax_in;
int		syntax_lex();

static syntax_t	*syntax_this;

/* yacc error reporting routine */
static int yyerror(char *s) {
	warnx("syntax definition '%s' line %d: %s", syntax_this->lang,
		syntax_lineno, s);
	usleep(500000);
	return 0;
}

/* read options for syntax_this */
static void readopt(u_char **opt) {
	int i;

	for (i = 0; opt[i]; i++) {
		if (!strcmp(opt[i], "hlnums"))
			syntax_this->hlnums = 1;
		free(opt[i]);
	}
	free(opt);
}

%}
%union {
	synitem_t	*item;
	u_char		**words;
	u_char		*str;
	int		n;
}

%token <words>	WORDLIST
%token <n>	WORDS TOKENS BEGINTOK END ENDATNL STRINGTOK
%token <n>	REGEXP QUOTE ARE OPTIONS
%token <str>	STRING

%type <item>	itemdef
%type <n>	color

%%

syntax		: /* empty */
		| syntax error
		| syntax itemdef ARE color ';' {
			TAILQ_INSERT_TAIL(&syntax_this->items, $2, si_list);
			$2->color = $4;
		}
		| syntax OPTIONS WORDLIST ';' {
			readopt($3);
		}
		;

itemdef		: REGEXP STRING {
			synitem_t *item;

			item = ckmalloc(sizeof(synitem_t));
			item->type = SYNITEM_REGEXP;
			item->u.re = ckmalloc(sizeof(regex_t));
			if (regcomp(item->u.re, $2, REG_EXTENDED))
				errx(1, "bad regexp \"%s\"", $2);
			free($2);
			$$ = item;
		}
		| WORDS WORDLIST {
			synitem_t *item;

			item = ckmalloc(sizeof(synitem_t));
			item->type = SYNITEM_WORDS;
			item->u.wordlist = sortwords($2);
			$$ = item;
		}
		| TOKENS WORDLIST {
			synitem_t *item;

			item = ckmalloc(sizeof(synitem_t));
			item->type = SYNITEM_TOKENS;
			item->u.wordlist = sortwords($2);
			$$ = item;
		}
		| BEGINTOK WORDLIST END WORDLIST {
			synitem_t *item;

			item = ckmalloc(sizeof(synitem_t));
			item->type = SYNITEM_BLOCK;
			item->u.blk.begin = sortwords($2);
			item->u.blk.end = sortwords($4);
			$$ = item;
		}
		| BEGINTOK WORDLIST ENDATNL {
			synitem_t *item;

			item = ckmalloc(sizeof(synitem_t));
			item->type = SYNITEM_TOEOL;
			item->u.wordlist = sortwords($2);
			$$ = item;
		}
		| STRINGTOK STRING QUOTE STRING {
			synitem_t *item;

			item = ckmalloc(sizeof(synitem_t));
			item->type = SYNITEM_STRING;
			item->u.str.tok = $2[0];
			item->u.str.quote = $4[0];
			$$ = item;
			free($2);
			free($4);
		}
		;

color		: STRING {
			int clr;

			clr = color_lookup($1);
			if (clr == COLOR_NULL)
				clr = COLOR_DEFAULT;

			free($1);
			$$ = clr;
		}
		;

%%

/*
 * load syntax definition for a language
 */
syntax_t *syntax_load(u_char *name) {
	char fullpath[PATH_MAX];
	char *synpath, *tmp, *path;
	char *expath;
	FILE *fp;

	/*
	 * prevent attempts to open a directory which causes breakage
	 * on some systems by failing the read() but not the open().
	 */
	if (*name == '\0')
		return NULL;

	/* make a local copy of the path for strsep */
	synpath = ckstrdup(SYNTAX_PATH);
	tmp = synpath;

	/* loop through the : deliminated path */
	fp = NULL;
	while ((path = strsep(&tmp, ":")) != NULL && !fp) {
		expath = tilde_expand(path);

		/* build the full path */
		if (snprintf(fullpath, sizeof(fullpath), "%s/%s",
				expath, name) >= sizeof(fullpath)) {
			warnx("path to syntax definition exceeds PATH_MAX");
			goto next;
		}

		/* try to open it */
		fp = fopen(fullpath, "r");

next:
		free(expath);
	}
	free(synpath);

	if (!fp)
		return NULL;

	/* parse the definition file */
	syntax_this = ckcalloc(1, sizeof(syntax_t));
	syntax_this->lang = ckstrdup(name);
	TAILQ_INIT(&syntax_this->items);
	syntax_in = fp;
	syntax_parse();
	fclose(fp);

	return syntax_this;
}
