%{
/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vhighlight.h"
#include "syntax_gram.tab.h"

/* build a word list */
static u_char **mkwordlist(char *yytext, int yyleng) {
	u_char **words;
	u_char *wd;
	int cnt, i;
	int wi, len;
	int quote;

	/* chop the ':' off the end */
	yytext[--yyleng] = '\0';

	/*
	 * count the number of words in the list so we know how
	 * big to make the memory allocation
	 */
	wd = NULL;
	cnt = 0;
	for (quote = 0, i = 1; i < yyleng; i++) {
		if (isspace(yytext[i]) && !quote) {
			if (wd) {
				cnt++;
				wd = NULL;
			}
		} else {
			if (!wd)
				wd = &yytext[i];
			if (yytext[i] == '\\') {
				quote = 1;
				continue;
			}
		}

		quote = 0;
	}

	/* add one for a null entry */
	cnt++;

	/* allocate the words array and fill it out */
	words = ckmalloc(cnt * sizeof(u_char *));
	for (quote = 0, wi = 0, i = 1; i < yyleng; i++) {
		if (isspace(yytext[i]) && !quote) {
			if (wd) {
				yytext[i] = '\0';
				words[wi++] = ckstrdup(wd);
				wd = NULL;
			}
		} else {
			if (!wd)
				wd = &yytext[i];
			if (yytext[i] == '\\') {
				quote = 1;
				continue;
			}
		}

		quote = 0;
	}
	assert(wi == cnt - 1);
	words[wi] = NULL;

	/* translate \ escapes */
	for (wi = 0; words[wi]; wi++) {
		len = strlen(words[wi]);
		for (i = 0; i < len; i++) {
			if (words[wi][i] == '\\') {
				if (++i >= len)
					break;
				switch (words[wi][i]) {
				case ':': case '\\':
				case ' ': case '\t':
					words[wi][i - 1] = words[wi][i];
					memmove(&words[wi][i],
						&words[wi][i + 1], len - i);
					len--;
					break;
				}
			}
		}
	}

	return words;
}

%}

%option yylineno
%option noyywrap
%option nounput

WORDLIST	":"([^\\:]|\\.)*":"
STRING		\"([^\\\"]|\\.|\\\n)*"\""

%%

	/*
	 * keywords
	 */

"regexp"	return REGEXP;
"words"		return WORDS;
"tokens"	return TOKENS;
"begin"		return BEGINTOK;
"end"		return END;
"endatnl"	return ENDATNL;
"string"	return STRINGTOK;
"quote"		return QUOTE;
"are"		return ARE;
"options"	return OPTIONS;


	/*
	 * datatypes
	 */

{WORDLIST} {
	syntax_lval.words = mkwordlist(yytext, yyleng);
	return WORDLIST;
}

{STRING} {
	syntax_lval.str = rclex_makestr(yytext, yyleng);
	return STRING;
}

	/*
	 * comments
	 */

";"		return ';';
"#".*\n		/* eat */
.|\n		/* eat */

%%
