/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vhighlight.h"

/* color to use to highlight numbers */
int vhighlight_numberclr;

/* keysym for setting vhighlight language */
static int vhighlight_setlang		= KEYSYM_NOP;

/* set the line considered valid for highlighting state */
static __inline void vhighlight_vstate(vhighlight_t *v, line_t *line,
		int linenum) {
	if (v->valid.linenum > linenum) {
		v->valid.line = line;
		v->valid.linenum = linenum;
	}
}

static void vhighlight_drawpart(vhighlight_t *v, line_t **line, int *linenum,
		int *col, int maxcols, int *y, int *viewcol) {
	anchor_t *anchor;
	synitem_t *state;
	int idx, start;
	int color, topcol;
	int low, len;

	/* vhighlights without syntax information can't draw */
	if (!v->syntax) {
		VIEWCALL(v->hdr.under, drawpart, line, linenum, col,
			maxcols, y, viewcol);
		return;
	}

	topcol = *col + maxcols;
	draw_setxy(v->vdef, *viewcol, *y);

	/*
	 * find the last valid anchor, if the valid state line is in front
	 * of the current one, we have to scan up to this line in order
	 * to get the state to use.
	 */
	anchor = anchor_findlast(&v->anchors, min(v->valid.linenum, *linenum));
	state = anchor ? anchor->state : NULL;
	while (v->valid.linenum < *linenum) {
		for (idx = 0; idx < v->valid.line->length;)
			syntax_highlight(v->syntax, v->valid.line,
				&idx, &state);

		/*
		 * now this line has a valid state, add an anchor if the
		 * state is different from the previous anchor.
		 */
		v->valid.line = TAILQ_NEXT(v->valid.line, l_list);
		v->valid.linenum++;
		if (!anchor || state != anchor->state)
			anchor = anchor_add(&v->anchors, v->valid.line,
				v->valid.linenum, state);
	}

	/*
	 * draw each portions of the line as the highlighter returns
	 * colors for them.
	 */
	idx = 0;
	while (idx < (*line)->length) {
		start = idx;
		color = syntax_highlight(v->syntax, *line, &idx, &state);
		assert(idx <= (*line)->length + 1);
		if (idx > (*line)->length)
			idx = (*line)->length;
		screen_setcolor(color);

		/* draw the section if it's in the part we're drawing */
		low = max(start, *col);
		if (low < idx) {
			len = idx - low;
			if (*col + len > topcol)
				len = topcol - *col;
			draw_string((*line)->text + low, v->hdr.geom.width,
				len, v->vdef->view.col, viewcol);
			*col += len;
			if (*col >= topcol)
				break;
		}
	}

	/* set up new valid state information */
	if (*col >= (*line)->length) {
		if (TAILQ_NEXT(*line, l_list)) {
			v->valid.line = TAILQ_NEXT(*line, l_list);
			v->valid.linenum = *linenum + 1;
			if (!anchor || state != anchor->state)
				anchor_add(&v->anchors, v->valid.line,
					v->valid.linenum, state);
		}
	} else
		vhighlight_vstate(v, *line, *linenum);
}

static void vhighlight_input(vhighlight_t *v, int keysym) {
	u_char *lang;

	/*
	 * XXX: this should be cleaned up a bit.
	 */
	if (keysym == vhighlight_setlang) {
		lang = minibuff_prompt("Language", NULL);
		if (!lang)
			goto callunder;
		anchor_emptylist(&v->anchors);
		v->valid.line = TAILQ_FIRST(&v->hdr.buffer->lines);
		v->valid.linenum = 0;
		v->syntax = syntax_findlang(lang);
		v->vdef->redraw_text = 1;
		free(lang);
		return;
	}

callunder:
	VIEWCALL(v->hdr.under, input, keysym);
}

/*
 * detect what language should be used to highlight for;
 * use a few simple rules to detect most cases:
 *
 * - check file extensions for matches with the languages
 * - check the first line for a #! /interpreter syntax
 * - check for a Emacs-style -*- language -*- on the first line
 */
static syntax_t *detectlang(buffer_t *buffer) {
	syntax_t *syntax;
	u_char *lang, *s;
	line_t *first;
	int len, i;

	/* extensions */
	lang = strrchr(buffer->name, '.');
	if (lang && (syntax = syntax_findlang(lang + 1)) != NULL)
		return syntax;

	/* #! interpreter */
	syntax = NULL;
	first = TAILQ_FIRST(&buffer->lines);
	if (first->text[0] == '#' && first->text[1] == '!') {
		lang = first->text;
		len = first->length;
		while ((s = memchr(lang, '/', len - 1)) != NULL) {
			s++;
			len -= s - lang;
			lang = s;
		}
		if (lang == first->text)
			goto out;

		/* make a copy so we can play with it */
		s = ckmalloc(len + 1);
		memcpy(s, lang, len);
		lang = s;
		lang[len] = '\0';

		/* special case for 'env', use what we're enving */
		if (!memcmp("env", lang, min(3, len))) {
			lang += 3;
			lang += strspn(lang, " \t");
			if (!*lang)
				goto free;
		}

		/* put a null at the first whitespace */
		for (i = 0; lang[i]; i++)
			if (lang[i] == ' ' || lang[i] == '\t') {
				lang[i] = '\0';
				break;
			}

		syntax = syntax_findlang(lang);
free:
		free(s);
		if (syntax)
			return syntax;
	}
out:

	/* emacs-style -*- language -*- on the first line */
	syntax = NULL;
	if ((lang = strnstr(first->text, "-*-", first->length)) != NULL) {
		lang += 3;
		len = first->length - (lang - first->text);
		s = ckmalloc(len + 1);
		memcpy(s, lang, len);
		lang = strnstr(s, "-*-", len);
		if (!lang)
			goto free2;
		*lang = '\0';
		lang = s + strspn(s, " \t");

		/* put a null at the first whitespace */
		for (i = 0; lang[i]; i++)
			if (lang[i] == ' ' || lang[i] == '\t') {
				lang[i] = '\0';
				break;
			}

		syntax = syntax_findlang(lang);
free2:
		free(s);
		if (syntax)
			return syntax;
	}

	return NULL;
}

static void vhighlight_map(vhighlight_t *v) {
	v->syntax = detectlang(v->hdr.buffer);
	anchor_emptylist(&v->anchors);
	v->valid.line = TAILQ_FIRST(&v->hdr.buffer->lines);
	v->valid.linenum = 0;

	v->vdef->redraw_header = 1;
	v->vdef->redraw_text = 1;
}

static void vhighlight_rmline(vhighlight_t *v, line_t *line, int linenum) {
	anchor_findlast(&v->anchors, linenum);
	if (TAILQ_PREV(line, linelist, l_list))
		vhighlight_vstate(v, TAILQ_PREV(line, linelist,
			l_list), linenum - 1);
	else {
		/*
		 * line being removed is the first line in the buffer,
		 * so set state to valid on the next line (which will
		 * become the first line) with number at 0.
		 */
		v->valid.line = TAILQ_NEXT(line, l_list);
		v->valid.linenum = 0;
	}
	VIEWCALL(v->hdr.under, rmline, line, linenum);
}

static void vhighlight_rmstr(vhighlight_t *v, line_t *line, int linenum,
		int loc, int slen) {
	anchor_findlast(&v->anchors, linenum);
	vhighlight_vstate(v, line, linenum);
	VIEWCALL(v->hdr.under, rmstr, line, linenum, loc, slen);
}

static void vhighlight_addline(vhighlight_t *v, line_t *line, int linenum) {
	anchor_findlast(&v->anchors, linenum);
	vhighlight_vstate(v, line, linenum);
	VIEWCALL(v->hdr.under, addline, line, linenum);
}

static void vhighlight_stradd(vhighlight_t *v, line_t *line, int linenum,
		int loc, int slen) {
	anchor_findlast(&v->anchors, linenum);
	vhighlight_vstate(v, line, linenum);
	VIEWCALL(v->hdr.under, stradd, line, linenum, loc, slen);
}

static void vhighlight_destroy(vhighlight_t *v) {
	anchor_emptylist(&v->anchors);
	free(v);
}

static viewhdr_t *vhighlight_create(viewhdr_t *under) {
	vhighlight_t *vhighlight;

	/* create the vhighlight structure */
	vhighlight = VIEW_CREATE(vhighlight_t, "vhighlight", under);
	vhighlight->vdef = view_findvdef((viewhdr_t *) vhighlight);
	TAILQ_INIT(&vhighlight->anchors);

	/* set view operations */
	vhighlight->hdr.op.draw		= NULL;
	vhighlight->hdr.op.drawpart	= VIEWOP(vhighlight_drawpart);
	vhighlight->hdr.op.viswid	= NULL;
	vhighlight->hdr.op.drawhdr	= NULL;
	vhighlight->hdr.op.input	= VIEWOP(vhighlight_input);
	vhighlight->hdr.op.map		= VIEWOP(vhighlight_map);
	vhighlight->hdr.op.geomchange	= NULL;
	vhighlight->hdr.op.activate	= NULL;
	vhighlight->hdr.op.rmline	= VIEWOP(vhighlight_rmline);
	vhighlight->hdr.op.rmstr	= VIEWOP(vhighlight_rmstr);
	vhighlight->hdr.op.addline	= VIEWOP(vhighlight_addline);
	vhighlight->hdr.op.stradd	= VIEWOP(vhighlight_stradd);
	vhighlight->hdr.op.destroy	= VIEWOP(vhighlight_destroy);

	return (viewhdr_t *) vhighlight;
}

/* type operations for the highlighter view */
static viewtypeops_t vhighlight_typeops = {
	vhighlight_create
};

/* register the view type for the highlighter view */
int vhighlight_init() {
	vhighlight_setlang = input_allocsyms(1);
	options_add("vhighlight_numberclr", OPT_COLOR,
		&vhighlight_numberclr, 1);
	vhighlight_numberclr = color_lookup("number");
	if (vhighlight_numberclr == COLOR_NULL)
		vhighlight_numberclr = COLOR_DEFAULT;
	command_add("vhighlight_setlang", vhighlight_setlang);
	VIEW_ADDTYPE(vhighlight_typeops, "vhighlight");
	syntax_init();

	return 0;
}

/* free up memory */
void vhighlight_shutdown() {
	syntax_shutdown();
	anchor_shutdown();
	command_remove("vhighlight_setlang");
	view_rmtype("vhighlight");
}
