/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "vtrailspace.h"

/* color to use to draw trailing whitespace */
static int color_trailspace;

/*
 * check the line the cursor is on, and the line the cursor was on
 * to see if they have trailing space; if they do, we need to redraw.
 * by checking this we manage to avoid drawing when the user moves the
 * cursor across lines that don't have trailing space.
 */
static int linetrails(vtrailspace_t *v) {
	u_char c;

	if (v->oldcursor.line) {
		c = v->oldcursor.line->text[v->oldcursor.line->length - 1];
		if (c == ' ' || c == '\t')
			return 1;
	}
	c = v->vdef->cursor.line->text[v->vdef->cursor.line->length - 1];
	if (c == ' ' || c == '\t')
		return 1;
	return 0;
}

static void vtrailspace_draw(vtrailspace_t *v) {
	if (v->vdef->redraw_text || v->hdr.buffer->dirty)
		goto draw;
	if (v->vdef->redraw_header) {
		if (linetrails(v))
			goto draw;
		if (!v->hdr.buffer->mark_implicit && v->hdr.buffer->mark)
			goto draw;
	}
	goto exit;

draw:
	draw_text(v->hdr.top, v->vdef);
	v->vdef->redraw_text = 0;
	v->vdef->redraw_header = 0;
exit:
	draw_mvcursor(v->vdef);
}

static void vtrailspace_drawpart(vtrailspace_t *v, line_t **line,
		int *linenum, int *col, int maxcols, int *y, int *viewcol) {
	int trail, len;

	/* save the cursor location to maybe avoid some redraws */
	memcpy(&v->oldcursor, &v->vdef->cursor, sizeof(v->oldcursor));

	/* don't draw it if their cursor is on the end of this line */
	if (v->vdef->cursor.line == *line &&
			v->vdef->cursor.col == (*line)->length)
		goto under;

	/* find out how much space is on the end */
	trail = (*line)->length;
	for (;;) {
		if ((*line)->text[trail - 1] != '\t' &&
				(*line)->text[trail - 1] != ' ')
			break;
		if (--trail < *col) {
			trail = *col;
			break;
		}
	}
	if (trail >= (*line)->length)
		goto under;
	if (trail - *col >= maxcols)
		goto under;
	if (trail - *col) {
		maxcols -= trail - *col;
		while (*col < trail)
			VIEWCALL(v->hdr.under, drawpart, line, linenum, col,
				trail - *col, y, viewcol);
	}
	screen_setcolor(color_trailspace);
	draw_setxy(v->vdef, *viewcol, *y);
	len = min((*line)->length - *col, maxcols);
	assert(len > 0);
	draw_string(&(*line)->text[*col], v->hdr.geom.width, len,
		v->vdef->view.col, viewcol);
	*col += len;

	return;

under:
	VIEWCALL(v->hdr.under, drawpart, line, linenum, col,
		maxcols, y, viewcol);
}

static void vtrailspace_map(vtrailspace_t *v) {
	v->oldcursor.line = NULL;
	v->vdef->redraw_text = 1;
	v->vdef->redraw_header = 1;
}

static void vtrailspace_rmline(vtrailspace_t *v, line_t *line, int linenum) {
	if (v->oldcursor.line == line)
		v->oldcursor.line = NULL;
	else if (v->oldcursor.line && linenum < v->oldcursor.linenum)
		v->oldcursor.linenum--;
	VIEWCALL(v->hdr.under, rmline, line, linenum);
}

static void vtrailspace_addline(vtrailspace_t *v, line_t *line, int linenum) {
	if (v->oldcursor.line && linenum < v->oldcursor.linenum)
		v->oldcursor.linenum++;
	VIEWCALL(v->hdr.under, addline, line, linenum);
}

static viewhdr_t *vtrailspace_create(viewhdr_t *under) {
	vtrailspace_t *vtrailspace;

	vtrailspace = VIEW_CREATE(vtrailspace_t, "vtrailspace", under);
	vtrailspace->vdef = view_findvdef((viewhdr_t *) vtrailspace);

	/* setup view operations */
	vtrailspace->hdr.op.draw	= VIEWOP(vtrailspace_draw);
	vtrailspace->hdr.op.drawpart	= VIEWOP(vtrailspace_drawpart);
	vtrailspace->hdr.op.viswid	= NULL;
	vtrailspace->hdr.op.drawhdr	= NULL;
	vtrailspace->hdr.op.input	= NULL;
	vtrailspace->hdr.op.map		= VIEWOP(vtrailspace_map);
	vtrailspace->hdr.op.geomchange	= NULL;
	vtrailspace->hdr.op.activate	= NULL;
	vtrailspace->hdr.op.rmline	= VIEWOP(vtrailspace_rmline);
	vtrailspace->hdr.op.rmstr	= NULL;
	vtrailspace->hdr.op.addline	= VIEWOP(vtrailspace_addline);
	vtrailspace->hdr.op.stradd	= NULL;
	vtrailspace->hdr.op.destroy	= VIEWOP(free);

	return (viewhdr_t *) vtrailspace;
}

static viewtypeops_t vtrailspace_typeops = {
	vtrailspace_create
};

int vtrailspace_init() {
	options_add("vtrailspace_color", OPT_COLOR, &color_trailspace, 1);
	color_trailspace = color_lookup("trailspace");
	if (color_trailspace == COLOR_NULL)
		color_trailspace = color_add("trailspace", "default", "red",
			SCREEN_A_NORM, SCREEN_A_STAND);
	VIEW_ADDTYPE(vtrailspace_typeops, "vtrailspace");
	return 0;
}

void vtrailspace_shutdown() {
	options_rm("vtrailspace_color");
	view_rmtype("vtrailspace");
}
