%{
/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

extern FILE	*yyin;
extern int	yylineno;

int	yylex();

/*
 * yacc error reporting function, display the error message and
 * pause for a half a second so the user can see the error before
 * we go into the fullscreen editing mode.
 */
static int yyerror(char *s) {
	warnx("line %d: %s", yylineno, s);
	usleep(500000);
	return 0;
}

%}
%union {
	u_char		*str;
	int		num;
	u_char		ch;
}

%token <num>	INTEGER
%token <str>	STRING
%token <ch>	CHARACTER
%token <num>	OPT KEY KEYCLEAR CMDCLEAR CLEARALLKEYS LOADMOD

%%

rcfile		: /* empty */
		| rcfile error
		| rcfile options_stmt ';'
		| rcfile keys_stmt ';'
		| rcfile LOADMOD STRING ';' {
			module_load($3);
			free($3);
		}
		;

options_stmt	: error
		| OPT STRING STRING {
			opt_t *opt;

			/*
			 * unknown option names are silently ignored; the
			 * option may be valid for a module that wasn't loaded
			 * or built in.
			 */
			opt = options_lookup($2);
			if (opt)
				options_set(opt, $3);
			free($2);
			free($3);
		}
		;

keys_stmt	: error
		| KEY STRING STRING {
			int keysym;

			/*
			 * unknown key names are silently ignored; the reason
			 * for this is that the user may have commands bound
			 * in the rc file that are for a module that wasn't
			 * built in.
			 */
			keysym = command_getkeysym($2);
			if (keysym != KEYSYM_NOP) {
				if (keybind_define(keysym, $3) == -1)
					errx(1, "invalid key string '%s'", $3);
			}
			free($2);
			free($3);
		}
		| KEYCLEAR STRING {
			if (keybind_define(0, $2) == -1)
				errx(1, "invalid key string '%s'", $2);
			free($2);
		}
		| CMDCLEAR STRING {
			int keysym;

			keysym = command_getkeysym($2);
			if (keysym == KEYSYM_NOP)
				errx(1, "unknown key name %s", $2);
			keybind_define(command_getkeysym($2), NULL);
			free($2);
		}
		| CLEARALLKEYS {
			keybind_clear();
		}
		;

%%

/* exported rcfile parsing function */
void rcfile_parse() {
	u_char rcfn[PATH_MAX];
	FILE *fp;

	/* find the rcfile in the user's home directory */
	snprintf(rcfn, sizeof(rcfn), "%s/.led/ledrc", getenv("HOME"));

	/* attempt to open the file, if we can't just silently exit */
	fp = fopen(rcfn, "r");
	if (!fp)
		return;

	/* parse the file and then close and exit */
	yyin = fp;
	yyparse();
	fclose(fp);
}
