/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

void region_order(region_t *r) {
	line_t *tline;
	int tmp;

	if (r->first.linenum < r->last.linenum)
		return;
	if (r->first.linenum == r->last.linenum)
		if (r->first.col <= r->last.col)
			return;

	/* we need to switch */
	swap(tmp, r->first.linenum, r->last.linenum);
	swap(tmp, r->first.col, r->last.col);
	swap(tline, r->first.line, r->last.line);
}

/*
 * delete region from buffer; the contents of the passed in region_t
 * are modified.
 */
void region_delete(buffer_t *buffer, int undoable, region_t *r) {
	line_t *line;

	if (r->first.linenum == r->last.linenum) {
		if (r->first.col < r->last.col)
			buffer_rmstr(buffer, undoable, r->first.line,
				r->first.linenum, r->first.col,
				r->last.col - r->first.col);
		return;
	}

	/* remove the first section */
	undo_beginblock(buffer);
	buffer_rmstr(buffer, undoable, r->first.line, r->first.linenum,
		r->first.col, r->first.line->length - r->first.col);

	/* remove every middle line */
	line = TAILQ_NEXT(r->first.line, l_list);
	for (; line != r->last.line;) {
		line = TAILQ_NEXT(line, l_list);
		buffer_rmline(buffer, undoable, TAILQ_PREV(line, linelist,
			l_list), r->first.linenum + 1);
		r->last.linenum--;
	}

	/* delete the last part; copy the remaining portion to the first */
	buffer_rmstr(buffer, undoable, r->last.line, r->last.linenum,
		0, r->last.col);
	buffer_nputs(buffer, undoable, r->first.line, r->first.linenum,
		r->first.col, r->last.line->text, r->last.line->length);
	buffer_rmline(buffer, undoable, r->last.line, r->last.linenum);

	undo_endblock(buffer);
}

/*
 * used to enumerate the contents of a region; the string for each
 * line is returned in *string, with length *len.  *len may be zero,
 * and should be checked on the caller end.
 */
int region_nextstr(buffer_t *buffer, region_t *r, line_t **line,
		int *linenum, int *col, u_char **string, int *len) {
	if (*line == NULL) {
		*line = r->first.line;
		*linenum = r->first.linenum;
		*col = r->first.col;
	}

	*string = &(*line)->text[*col];

	/* check if this is the last line in the region */
	if (*line == r->last.line) {
		*len = r->last.col - *col;
		return 1;
	}

	*len = (*line)->length - *col;
	*line = TAILQ_NEXT(*line, l_list);
	(*linenum)++;
	*col = 0;

	return 0;
}
