/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* should a beep happen on the next screen_refresh() */
int		screen_needbell = 0;

/* are keysyms enabled */
int		screen_keysyms	= 1;

/*
 * because we combine the attributes of a character into the
 * concept of a color from outside this module, colors are managed
 * inside here using this.
 */
static int	numclrs	= 0;
static chtype	*clrs	= NULL;

/* does the tty support color */
static int	screen_color = 0;

/* signal handler for window size changes */
static void sigwinch_handler(int sig) {
	screen_reinit();
}

/* initialize the screen */
void screen_init() {
	initscr();
	signal(SIGWINCH, sigwinch_handler);
	raw();
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	scrollok(stdscr, FALSE);

	/* check for colors */
	if (has_colors()) {
		screen_color = 1;
		start_color();
		use_default_colors();
	}
}

/* perform neccessary cleanup for the end */
void screen_shutdown() {
	if (clrs)
		free(clrs);
	endwin();
}

/*
 * define a color; attrs is attributes to use, bwattrs are used if the tty
 * can't do color.
 */
void screen_defcolor(int ident, u_char *fg, u_char *bg,
		int attrs, int bwattrs) {
	char *chars[2];
	short num[2];
	int i;

	/* get room for the color on the internal color list */
	if (ident > numclrs) {
		clrs = ckrealloc(clrs, sizeof(chtype) * ident);
		numclrs = ident;
	}

	/* just use bwattrs if there's no color available */
	if (!screen_color) {
		clrs[ident - 1] = bwattrs;
		return;
	}

	/* copy the color strings */
	chars[0] = fg;
	chars[1] = bg;

	/* find color to use based on name */
	for (i = 0; i < 2; i++) {
		if (!strcmp(chars[i], "default")) {
			num[i] = -1;
		} else if (strstr(chars[i], "black") != NULL) {
			num[i] = COLOR_BLACK;
		} else if (!strcmp(chars[i], "gray")) {
			if (i == 0)
				attrs |= A_BOLD;
			num[i] = COLOR_BLACK;
		} else if (strstr(chars[i], "red") != NULL) {
			num[i] = COLOR_RED;
		} else if (strstr(chars[i], "green") != NULL) {
			num[i] = COLOR_GREEN;
		} else if (strstr(chars[i], "brown") != NULL) {
			num[i] = COLOR_YELLOW;
		} else if (!strcmp(chars[i], "yellow")) {
			if (i == 0)
				attrs |= A_BOLD;
			num[i] = COLOR_YELLOW;
		} else if (strstr(chars[i], "blue") != NULL) {
			num[i] = COLOR_BLUE;
		} else if (strstr(chars[i], "magenta") != NULL) {
			num[i] = COLOR_MAGENTA;
		} else if (strstr(chars[i], "cyan") != NULL) {
			num[i] = COLOR_CYAN;
		} else {
			/*
			 * should be for a string named "white", but we
			 * treat unknowns like this also.
			 */
			num[i] = COLOR_WHITE;
		}

		/*
		 * additionally we allow "bright" in front of any color name
		 * to add the bold attribute.
		 */
		if (strstr(chars[i], "bright") == chars[i] && i == 0)
			attrs |= A_BOLD;
	}

	/* initialize the color pair */
	init_pair(ident, num[0], num[1]);

	/* set the clr in our colors information array */
	clrs[ident - 1] = COLOR_PAIR(ident) | attrs;
}

/* set the current color */
void screen_setcolor(int ident) {
	if (ident && ident - 1 < numclrs)
		attrset(clrs[ident - 1]);
}

/* turn on or off keysyms */
void screen_nokeysyms(int b) {
	screen_keysyms = !b;
	keypad(stdscr, !b);
}

/*
 * return the ident of a pressed key, the identifiers for simple
 * characters are the value of the character ('a', 'b', and so forth)
 */
int screen_getkey() {
	/*
	 * XXX: for some reason this is neccessary in order
	 * for ncurses not to screw up SIGWINCH stuff.  I'm not
	 * sure why, but just calling getch (without calling
	 * timeout(), which is what screen_key_pending does)
	 * results in getting stopped, and thus messes up
	 * winches and stuff until another key comes across
	 * the tty.
	 */
	for (;;)
		if (screen_key_pending(100000))
			break;
	return getch();
}

/* put a key back, return -1 on error */
int screen_ungetkey(int ch) {
	if (ungetch(ch) == ERR)
		return -1;
	else
		return 0;
}

/* check for pending input w/ timeout of m in milliseconds */
int screen_key_pending(int m) {
	int c;

	timeout(m);
	c = getch();
	if (c != ERR)
		ungetch(c);
	timeout(-1);

	return (c == ERR) ? 0 : 1;
}

/* print formatted output to the screen */
int screen_nprintf(int n, u_char *fmt, ...) {
	u_char buff[n + 1];	/* + 1 for a \0 */
	va_list args;
	int cnt;

	va_start(args, fmt);
	cnt = vsnprintf(buff, sizeof(buff), fmt, args);
	va_end(args);

	screen_nputs(buff, n);

	return min(n, cnt);
}

/* flush all changes to the user's tty */
void screen_refresh() {
	if (screen_needbell) {
		beep();
		screen_needbell = 0;
	}
	refresh();
}
