/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* return the next character in a buffer, or -1 if we ran out */
static __inline int nextch(line_t **line, int *linenum, int *col) {
	if (*col < (*line)->length)
		return (*line)->text[(*col)++];
	if (!TAILQ_NEXT(*line, l_list))
		return -1;
	*line = TAILQ_NEXT(*line, l_list);
	(*linenum)++;
	*col = 0;

	return '\n';
}

/*
 * return the character at the current location, and move backward
 * one, return a -1 if we ran out.
 */
static __inline int prevch(line_t **line, int *linenum, int *col) {
	if (*col)
		return (*line)->text[--(*col)];
	if (!TAILQ_PREV(*line, linelist, l_list))
		return -1;
	*line = TAILQ_PREV(*line, linelist, l_list);
	(*col) = (*line)->length;
	(*linenum)--;

	return '\n';
}

/* find the next occurrence of a string in a linelist */
int search_next(u_char *s, int slen, line_t **rline,
		int *rlinenum, int *rcol) {
	line_t *line, *save_line;
	int col, linenum;
	int save_col, save_linenum;
	int c, i;

	line = *rline;
	linenum = *rlinenum;
	col = *rcol;
	for (;;) {
		c = nextch(&line, &linenum, &col);
		if (c == -1)
			break;
		save_line = line;
		save_linenum = linenum;
		save_col = col;

		/* see if we match the string */
		for (i = 0;; i++) {
			if (options.icasesearch ? tolower(c) != tolower(s[i])
					: c != s[i])
				break;
			if (i + 1 == slen) {
				*rline = save_line;
				*rlinenum = save_linenum;
				*rcol = save_col;
				prevch(rline, rlinenum, rcol);
				return 0;
			}
			c = nextch(&line, &linenum, &col);
			if (c == -1)
				break;
		}

		/* go back to the position following the save */
		line = save_line;
		linenum = save_linenum;
		col = save_col;
	}

	return -1;
}

/* find the previous occurrence of a string in a linelist */
int search_prev(u_char *s, int slen, line_t **rline,
		int *rlinenum, int *rcol) {
	line_t *line, *save_line;
	int col, linenum;
	int save_col, save_linenum;
	int c, i;

	line = *rline;
	linenum = *rlinenum;
	col = *rcol;
	for (;;) {
		c = prevch(&line, &linenum, &col);
		if (c == -1)
			break;
		save_line = line;
		save_linenum = linenum;
		save_col = col;

		/* see if we match the string */
		for (i = 0;; i++) {
			c = nextch(&line, &linenum, &col);
			if (c == -1)
				break;
			if (options.icasesearch ? tolower(c) != tolower(s[i])
					: c != s[i])
				break;
			if (i + 1 == slen) {
				*rline = save_line;
				*rlinenum = save_linenum;
				*rcol = save_col;
				return 0;
			}
		}

		/* go back to the position preceding the save */
		line = save_line;
		linenum = save_linenum;
		col = save_col;
	}

	return -1;
}
