/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

void undo_beginblock(buffer_t *buffer) {
	undo_t *undo;

	undo = ckmalloc(sizeof(undo_t));
	undo->type = UNDO_BEGINBLOCK;
	undo_additem(buffer, undo);
}

void undo_endblock(buffer_t *buffer) {
	undo_t *undo, *prev;

	/*
	 * remove empty blocks; if the undoitem on the end of
	 * the list is a block begin, just trash it and return.
	 */
	assert(!buffer->undoitem);
	prev = TAILQ_LAST(&buffer->undolist, undolist);
	if (prev->type == UNDO_BEGINBLOCK) {
		undo_remove(&buffer->undolist, prev);
		return;
	}

	undo = ckmalloc(sizeof(undo_t));
	undo->type = UNDO_ENDBLOCK;
	undo_additem(buffer, undo);
}

void undo_remove(undolist_t *undolist, undo_t *undo) {
	if (undo->type == UNDO_TEXTDELETE || undo->type == UNDO_TEXTINSERT)
		free(undo->op.text.s);
	TAILQ_REMOVE(undolist, undo, u_list);
	free(undo);
}

void undo_emptylist(buffer_t *buffer) {
	while (!TAILQ_EMPTY(&buffer->undolist))
		undo_remove(&buffer->undolist, TAILQ_FIRST(&buffer->undolist));
	buffer->undoitem = NULL;
}

/*
 * add a new undo item; if the current undoitem pointer in the buffer
 * is not at the end of the undolist, we need to wipe the rest of the
 * list.
 */
void undo_additem(buffer_t *buffer, undo_t *undo) {
	if (buffer->undoitem) {
		while (TAILQ_NEXT(buffer->undoitem, u_list))
			undo_remove(&buffer->undolist,
				TAILQ_NEXT(buffer->undoitem, u_list));
		undo_remove(&buffer->undolist, buffer->undoitem);
		buffer->undoitem = NULL;
	}
	TAILQ_INSERT_TAIL(&buffer->undolist, undo, u_list);
}

/*
 * rollback a number of buffer operations; treat anything
 * within a UNDO_BEGINBLOCK and UNDO_ENDBLOCK pair as a
 * single operation.  allows nesting of undo blocks.
 */
int undo_rollback(buffer_t *buffer, int *linenum) {
	undo_t *undo;
	line_t *line;
	int blklvl, rolled;

	undo = buffer->undoitem;
	rolled = blklvl = 0;
	do {
		if (undo) {
			if (!TAILQ_PREV(undo, undolist, u_list))
				break;
			undo = TAILQ_PREV(undo, undolist, u_list);
		} else {
			undo = TAILQ_LAST(&buffer->undolist, undolist);
			if (!undo)
				break;
		}
		if (undo->type != UNDO_BEGINBLOCK &&
				undo->type != UNDO_ENDBLOCK)
			*linenum = undo->linenum;

		switch (undo->type) {
		case UNDO_TEXTINSERT:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_rmstr(buffer, 0, line, undo->linenum,
				undo->op.text.col, undo->op.text.slen);
			break;
		case UNDO_TEXTDELETE:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_nputs(buffer, 0, line, undo->linenum,
				undo->op.text.col, undo->op.text.s,
				undo->op.text.slen);
			break;
		case UNDO_PUTCHAR:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_rmch(buffer, 0, line, undo->linenum,
				undo->op.text.col);
			break;
		case UNDO_RMCHAR:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_putc(buffer, 0, line, undo->linenum,
				undo->op.ch.col, undo->op.ch.c);
			break;
		case UNDO_LINEINSERT:
			line = buffer_lineat(buffer, undo->linenum + 1, NULL);
			buffer_rmline(buffer, 0, line, undo->linenum + 1);
			break;
		case UNDO_LINEDELETE:
			line = !undo->linenum ? NULL
				: buffer_lineat(buffer, undo->linenum - 1,
				NULL);
			buffer_addline(buffer, 0, line, undo->linenum - 1);
			break;
		case UNDO_SETNLSTYLE:
			buffer->nlstyle = buffer->nlstyle == NLSTYLE_LF ?
				NLSTYLE_CRLF : NLSTYLE_LF;
			break;
		case UNDO_BEGINBLOCK:
			blklvl--;
			break;
		case UNDO_ENDBLOCK:
			blklvl++;
			break;
		}

		rolled++;
		assert(blklvl >= 0);
	} while (blklvl);
	assert(blklvl == 0);

	if (buffer->undodisk)
		buffer->modified = undo != TAILQ_NEXT(buffer->undodisk,
			u_list);
	else
		buffer->modified = undo != TAILQ_FIRST(&buffer->undolist);
	buffer->undoitem = undo;
	buffer->dirty = 1;

	return rolled;
}

/*
 * roll forward a number of buffer changes; very similar to
 * undo_rollback, but goes in the other direction.
 */
int undo_rollforward(buffer_t *buffer, int *linenum) {
	undo_t *undo;
	line_t *line;
	int blklvl, rolled;

	undo = buffer->undoitem;
	rolled = blklvl = 0;
	do {
		if (!undo)
			break;

		switch (undo->type) {
		case UNDO_TEXTINSERT:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_nputs(buffer, 0, line, undo->linenum,
				undo->op.text.col, undo->op.text.s,
				undo->op.text.slen);
			break;
		case UNDO_TEXTDELETE:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_rmstr(buffer, 0, line, undo->linenum,
				undo->op.text.col, undo->op.text.slen);
			break;
		case UNDO_PUTCHAR:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_putc(buffer, 0, line, undo->linenum,
				undo->op.ch.col, undo->op.ch.c);
			break;
		case UNDO_RMCHAR:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_rmch(buffer, 0, line, undo->linenum,
				undo->op.ch.col);
			break;
		case UNDO_LINEINSERT:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_addline(buffer, 0, line, undo->linenum);
			break;
		case UNDO_LINEDELETE:
			line = buffer_lineat(buffer, undo->linenum, NULL);
			buffer_rmline(buffer, 0, line, undo->linenum);
			break;
		case UNDO_SETNLSTYLE:
			buffer->nlstyle = buffer->nlstyle == NLSTYLE_LF ?
				NLSTYLE_CRLF : NLSTYLE_LF;
			break;
		case UNDO_BEGINBLOCK:
			blklvl++;
			break;
		case UNDO_ENDBLOCK:
			blklvl--;
			break;
		}

		rolled++;
		assert(blklvl >= 0);
		if (undo->type != UNDO_BEGINBLOCK &&
				undo->type != UNDO_ENDBLOCK)
			*linenum = undo->linenum;
		undo = TAILQ_NEXT(undo, u_list);
	} while (blklvl);
	assert(blklvl == 0);

	if (buffer->undodisk)
		buffer->modified = undo != TAILQ_NEXT(buffer->undodisk,
			u_list);
	else
		buffer->modified = undo != TAILQ_FIRST(&buffer->undolist);
	buffer->undoitem = undo;
	buffer->dirty = 1;

	return rolled;
}
