/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef UNDO_H
#define UNDO_H

/* types of undoable operations */
enum {
	UNDO_TEXTINSERT,
	UNDO_TEXTDELETE,
	UNDO_PUTCHAR,
	UNDO_RMCHAR,
	UNDO_LINEINSERT,	/* only should be used with empty lines */
	UNDO_LINEDELETE,
	UNDO_SETNLSTYLE,

	/*
	 * some operations that should seem to the user to be a single
	 * operation (such as killing a region consisting of multiple
	 * lines) are made to operate as such with the following.
	 */
	UNDO_BEGINBLOCK,
	UNDO_ENDBLOCK
};

/*
 * an undoable buffer operation; undo operations can be rolled
 * in either direction (for 'redo' support), so all of the undo
 * deltas for insertion store the data they inserted just like
 * the ones for deletion.
 */
struct undo {
	TAILQ_ENTRY(undo) u_list;

	int	type;		/* undoable change type */
	int	linenum;	/* line number where the operation occured */

	/* undo operation data */
	union {
		/* text inserted or deleted */
		struct {
			int	col;	/* column it occured at */
			u_char	*s;	/* string deleted or inserted */
			int	slen;	/* length of the string */
		} text;

		/* character inserted or deleted */
		struct {
			int	col;
			u_char	c;	/* the character */
		} ch;

		/*
		 * line additions and removals are done only on
		 * empty lines, so it is not neccesary to store
		 * any additional data for those operations.
		 */
	} op;
};

TAILQ_HEAD(undolist, undo);

void	undo_beginblock(buffer_t *);
void	undo_endblock(buffer_t *);
void	undo_remove(undolist_t *, undo_t *undo);
void	undo_emptylist(buffer_t *);
void	undo_additem(buffer_t *, undo_t *);
int	undo_rollback(buffer_t *, int *linenum);
int	undo_rollforward(buffer_t *, int *linenum);

#endif
