/*
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

static void vdefault_draw(vdefault_t *v) {
	draw_standard(v->hdr.top, v);
	draw_mvcursor(v);
}

static void vdefault_drawpart(vdefault_t *v, line_t **line, int *linenum,
		int *col, int maxcols, int *y, int *viewcol) {
	int len;

	screen_setcolor(COLOR_DEFAULT);
	draw_setxy(v, *viewcol, *y);

	len = min(maxcols, (*line)->length - *col);
	if (len) {
		draw_string(&(*line)->text[*col], v->hdr.geom.width, len,
			v->view.col, viewcol);
		*col += len;
	}
}

static int vdefault_viswid(viewhdr_t *v, int viewcol, u_char c) {
	if (c == '\t')
		return draw_next_tabstop(viewcol);
	else if (input_isctrl(c)) {
		if (c < 32 || c == '\177')
			return 2;
		else
			return 4;
	} else
		return 1;
}

/* view header drawing for vdefault */
static void vdefault_drawhdr(vdefault_t *v, int x, int y, int width) {
	/*
	 * the redraw_header flag isn't reset here, because we will
	 * be interested in its contents when vdefault_draw is called
	 * to determine if we need to draw text when there's an explicit
	 * mark set.
	 */
	if (v->hdr.buffer->dirty || v->redraw_header)
		draw_header(v, x, y, width);
}

/* input handler for vdefault */
static void vdefault_input(vdefault_t *v, int keysym) {
	v->in_command = 1;

	/* always treat quoted keys like normal characters */
	if (input_lastkeysym == KEYSYM_QUOTENEXT) {
		assert(keysym <= UCHAR_MAX);
		goto type;
	}

	switch (keysym) {
	case '\r': case '\n':
	case KEYSYM_ENTER:	command_enter(v);	break;
	case KEYSYM_UP:		command_up(v);		break;
	case KEYSYM_DOWN:	command_down(v);	break;
	case KEYSYM_LEFT:	command_left(v);	break;
	case KEYSYM_RIGHT:	command_right(v);	break;
	case KEYSYM_HOME:	command_home(v);	break;
	case KEYSYM_END:	command_end(v);		break;
	case KEYSYM_NEXTWORD:	command_nextword(v);	break;
	case KEYSYM_PREVWORD:	command_prevword(v);	break;
	case KEYSYM_NEXTPARA:	command_nextpara(v);	break;
	case KEYSYM_PREVPARA:	command_prevpara(v);	break;
	case KEYSYM_VIEWTOP:	command_viewtop(v);	break;
	case KEYSYM_VIEWBOTTOM:	command_viewbottom(v);	break;
	case KEYSYM_PGUP:	command_pgup(v);	break;
	case KEYSYM_PGDN:	command_pgdn(v);	break;
	case KEYSYM_LINEUP:	command_lineup(v);	break;
	case KEYSYM_LINEDN:	command_linedn(v);	break;
	case KEYSYM_FIRSTLINE:	command_firstline(v);	break;
	case KEYSYM_LASTLINE:	command_lastline(v);	break;
	case KEYSYM_HALFPGUP:	command_halfpgup(v);	break;
	case KEYSYM_HALFPGDN:	command_halfpgdn(v);	break;
	case KEYSYM_CENTER:	command_center(v);	break;
	case KEYSYM_BACKSPACE:	command_backspace(v);	break;
	case KEYSYM_DELETE:	command_delete(v);	break;
	case KEYSYM_OPENLINE:	command_openline(v);	break;
	case KEYSYM_TOGNLSTYLE:	command_tognlstyle(v);	break;
	case KEYSYM_UNDO:	command_undo(v);	break;
	case KEYSYM_REDO:	command_redo(v);	break;
	case KEYSYM_ISEARCH:	command_isearch(v);	break;
	case KEYSYM_RISEARCH:	command_risearch(v);	break;
	case KEYSYM_GOTOLN:	command_gotoln(v);	break;
	case KEYSYM_REPLACE:	command_replace(v);	break;
	case KEYSYM_TRANCHARS:	command_tranchars(v);	break;
	case KEYSYM_TRANWORDS:	command_tranwords(v);	break;
	case KEYSYM_TRANLINES:	command_tranlines(v);	break;
	case KEYSYM_KILLRGN:	command_killrgn(v);	break;
	case KEYSYM_KILLLINE:	command_killline(v);	break;
	case KEYSYM_KILLEOL:	command_killeol(v);	break;
	case KEYSYM_KILLWORD:	command_killword(v);	break;
	case KEYSYM_FWDKILLWORD:command_fwdkillword(v);	break;
	case KEYSYM_APPENDKILL:	command_appendkill(v);	break;
	case KEYSYM_YANK:	command_yank(v);	break;
	case KEYSYM_YANKPOP:	command_yankpop(v);	break;
	case KEYSYM_FWDYANKPOP:	command_fwdyankpop(v);	break;
	case KEYSYM_SETMARK:	command_setmark(v);	break;
	case KEYSYM_POPMARK:	command_popmark(v);	break;
	case KEYSYM_PROMOTEMARK:command_promotemark(v);	break;
	case KEYSYM_DEMOTEMARK:	command_demotemark(v);	break;
	case KEYSYM_XCHGMARK:	command_xchgmark(v);	break;
	case KEYSYM_FILTER:	command_filter(v);	break;
	case KEYSYM_INSERTFILE:	command_insertfile(v);	break;
	default:
		if (keysym <= UCHAR_MAX && !input_isctrl(keysym)) {
type:
			command_type(v, keysym);
		} else
			input_default_handler((viewhdr_t *) v, keysym);
		break;
	}

	/*
	 * set protection flags back to nothing so commands
	 * dont have to.  clear the in_command flag.
	 */
	v->rm_protect = 0;
	v->in_command = 0;
}

/* handle initial view mapping */
static void vdefault_map(vdefault_t *v) {
	v->cursor.line = v->view.line = TAILQ_FIRST(&v->hdr.buffer->lines);
	v->cursor.linenum = v->view.linenum = 0;
	v->cursor.col = v->view.col = 0;
	v->cursor.lastcol = 0;
	v->cursor.screencol = 0;
	v->inhibit_parenflash = v->in_command = v->rm_protect = 0;

	/* we'll need to do a first draw */
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/*
 * handle changes in the view geometry; set flags for redrawing
 * and make sure the cursor didn't get scrolled off the visible
 * portion of the view.
 */
static void vdefault_geomchange(vdefault_t *v) {
	vdefault_setscreencol(v);
	vdefault_checkscroll(v);
	v->redraw_header = 1;
	v->redraw_text = 1;
}

static void vdefault_activate(vdefault_t *v) {
	v->redraw_header = 1;
	v->redraw_text = 1;
}

/*
 * someone (or us) removed a line from our buffer; pay attention to
 * bits in rm_protect so that commands can prevent us from messing up
 * their changes.
 */
static void vdefault_rmline(vdefault_t *v, line_t *line, int linenum) {
	if (!(v->rm_protect & PROTECT_VIEW)) {
		if (v->view.line == line) {
			v->view.line = TAILQ_NEXT(v->view.line, l_list);
			if (!v->view.line) {
				v->view.line = TAILQ_PREV(line, linelist,
					l_list);
				v->view.linenum--;
			}
		} else if (linenum < v->view.linenum)
			v->view.linenum--;
		v->redraw_text = 1;
	}
	if (!(v->rm_protect & PROTECT_CURSOR)) {
		if (v->cursor.line == line) {
			v->cursor.line = TAILQ_NEXT(v->cursor.line, l_list);
			if (!v->cursor.line) {
				v->cursor.line = TAILQ_PREV(line, linelist,
					l_list);
				v->cursor.linenum--;
			}
			vdefault_trylastcol(v);
			if (!(v->rm_protect & PROTECT_VIEW))
				vdefault_checkscroll(v);
			v->redraw_header = 1;
		} else if (linenum < v->cursor.linenum)
			v->cursor.linenum--;
	}
}

/* handle string removal from our buffer */
static void vdefault_rmstr(vdefault_t *v, line_t *line, int linenum,
		int loc, int slen) {
	if (!(v->rm_protect & PROTECT_CURSOR) && v->cursor.line == line &&
			v->cursor.col >= loc) {
		if (v->cursor.col < loc + slen)
			v->cursor.col = loc <= line->length - slen ? loc :
				(loc ? loc - 1 : 0);
		else
			v->cursor.col -= slen;
		vdefault_setscreencol(v);
		if (!(v->rm_protect & PROTECT_VIEW))
			vdefault_checkscroll(v);
		v->redraw_header = 1;
	}
}

/*
 * if lines are added above the view/cursor line it is neccesary to
 * increment linenumbers to maintain consistency.
 */
static void vdefault_addline(vdefault_t *v, line_t *line, int linenum) {
	if (linenum < v->view.linenum)
		v->view.linenum++;
	if (linenum < v->cursor.linenum) {
		v->cursor.linenum++;
		v->redraw_header = 1;
	}
}

/*
 * update the cursor screen column if characters are added before
 * the cursor col
 */
static void vdefault_stradd(vdefault_t *v, line_t *line, int linenum,
		int loc, int slen) {
	if (v->in_command)
		return;
	if (line == v->cursor.line && loc < v->cursor.col) {
		v->cursor.col += slen;
		vdefault_setscreencol(v);
		vdefault_checkscroll(v);
		v->redraw_header = 1;
	}
}

/* create a new vdefault view */
static viewhdr_t *vdefault_create(viewhdr_t *under) {
	vdefault_t *vdefault;

	/* create the view structure */
	vdefault = VIEW_CREATE(vdefault_t, "vdefault", under);

	/* set view operations */
	vdefault->hdr.op.draw		= VIEWOP(vdefault_draw);
	vdefault->hdr.op.drawpart	= VIEWOP(vdefault_drawpart);
	vdefault->hdr.op.viswid		= VIEWOP(vdefault_viswid);
	vdefault->hdr.op.drawhdr	= VIEWOP(vdefault_drawhdr);
	vdefault->hdr.op.input		= VIEWOP(vdefault_input);
	vdefault->hdr.op.map		= VIEWOP(vdefault_map);
	vdefault->hdr.op.geomchange	= VIEWOP(vdefault_geomchange);
	vdefault->hdr.op.activate	= VIEWOP(vdefault_activate);
	vdefault->hdr.op.rmline		= VIEWOP(vdefault_rmline);
	vdefault->hdr.op.rmstr		= VIEWOP(vdefault_rmstr);
	vdefault->hdr.op.addline	= VIEWOP(vdefault_addline);
	vdefault->hdr.op.stradd		= VIEWOP(vdefault_stradd);
	vdefault->hdr.op.destroy	= VIEWOP(free);

	return (viewhdr_t *) vdefault;
}

/* type operations for vdefault */
static viewtypeops_t vdefault_typeops = {
	vdefault_create
};

/* register the vdefault view type */
void vdefault_init() {
	VIEW_ADDTYPE(vdefault_typeops, "vdefault");
}

/*
 * fill out a region_t with the region defined by the location of
 * the cursor and the current mark.  the caller is responsible
 * for making sure a mark exists before calling this method.
 */
void vdefault_getmarkrgn(vdefault_t *v, region_t *r) {
	memcpy(&r->last, v->hdr.buffer->mark, sizeof(r->last));
	memcpy(&r->first, &v->cursor, sizeof(r->first));
	region_order(r);
}

/*
 * recompute the screen column of the cursor for a vdefault view, and
 * save it as the new lastcol for the vdefault.
 */
void vdefault_setscreencol(vdefault_t *v) {
	int i;

	v->cursor.screencol = 0;
	for (i = 0; i < v->cursor.col; i++)
		v->cursor.screencol += VIEWCALLI(v->hdr.top, viswid,
			v->cursor.screencol, v->cursor.line->text[i]);
	v->cursor.lastcol = v->cursor.screencol;
}

/*
 * try to move the cursor column to the last cursor column, if this is not
 * possible just put it on the end of the line.  this is done so that
 * when the user moves the cursor across lines it stays as close to
 * the column it was at as is possible.
 */
void vdefault_trylastcol(vdefault_t *v) {
	int i;

	for (i = v->cursor.screencol = 0; i < v->cursor.line->length
			&& v->cursor.screencol < v->cursor.lastcol; i++)
		v->cursor.screencol += VIEWCALLI(v->hdr.top, viswid,
			v->cursor.screencol, v->cursor.line->text[i]);
	v->cursor.col = i;
}

/*
 * called from commands that could cause the visible portion of the
 * buffer to change, and update the scrolling appropriately.
 */
void vdefault_checkscroll(vdefault_t *v) {
	int i;

	/* vertical scrolling */
	if (v->cursor.linenum < v->view.linenum) {
		if (options.chunkyscroll) {
			i = 1 + v->cursor.linenum - v->hdr.geom.height / 2;
			while (TAILQ_PREV(v->view.line, linelist, l_list)) {
				if (v->view.linenum < i)
					break;
				v->view.line = TAILQ_PREV(v->view.line,
					linelist, l_list);
				v->view.linenum--;
			}
		} else {
			v->view.line = v->cursor.line;
			v->view.linenum = v->cursor.linenum;
		}
		v->redraw_text = 1;
	} else if (v->cursor.linenum >= v->view.linenum + v->hdr.geom.height) {
		if (options.chunkyscroll)
			i = 1 + v->cursor.linenum - v->hdr.geom.height / 2;
		else
			i = 1 + v->cursor.linenum - v->hdr.geom.height;
		while (TAILQ_NEXT(v->view.line, l_list)) {
			if (v->view.linenum >= i)
				break;
			v->view.line = TAILQ_NEXT(v->view.line, l_list);
			v->view.linenum++;
		}
		v->redraw_text = 1;
	}

	/* horizontal scrolling */
	if (v->cursor.screencol < v->view.col) {
		v->view.col = v->cursor.screencol - v->hdr.geom.width / 2;
		if (v->view.col < 0)
			v->view.col = 0;
		v->redraw_text = 1;
	} else if (v->cursor.screencol >= v->view.col + v->hdr.geom.width) {
		v->view.col = v->cursor.screencol - v->hdr.geom.width +
			v->hdr.geom.width / 2;
		v->redraw_text = 1;
	}
}
