/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

/* the hash of all view types */
viewtypelist_t	*view_types;

/*
 * initialization, calls inits for any view types that are
 * built into the editor binary.
 */
void view_init() {
	view_types = hash_alloc("view_types", VTYPE_BUCKETS,
		offsetof(viewtype_t, vt_list));
	vdefault_init();
	vminibuff_init();
}

void view_shutdown() {
	viewtype_t *type;
	int i;

	/* free up the viewtype hash */
	for (i = 0; i < VTYPE_BUCKETS; i++)
		while (!LIST_EMPTY(&view_types[i])) {
			type = LIST_FIRST(&view_types[i]);
			LIST_REMOVE(type, vt_list);
			free(type);
		}
	hash_free(view_types);
}

/* locate a viewtype_t by name */
viewtype_t *view_findtype(u_char *name) {
	viewtype_t *type;
	int idx;

	idx = hash_string(VTYPE_BUCKETS, name);
	LIST_FOREACH(type, &view_types[idx], vt_list)
		if (!strcmp(type->name, name))
			return type;
	return NULL;
}

void view_rmtype(u_char *name) {
	viewtype_t *type;

	type = view_findtype(name);
	assert(type);
	LIST_REMOVE(type, vt_list);
	free(type);
}

/*
 * create a new view, if under is nonzero, the new view
 * will be stacked on top of it.
 */
viewhdr_t *view_create(viewtype_t *type, viewhdr_t *under) {
	viewhdr_t *view;

	view = type->op.create(under);

	return view;
}

/*
 * destroy a view; views stacked below the view are left
 * untouched.  a pointer to the view immediatly below
 * the one destroyed is returned.
 */
viewhdr_t *view_destroy(viewhdr_t *view) {
	viewhdr_t *under;

	under = view->under;
	buffer_rele(view->buffer);
	VIEWCALL(view, destroy);

	return under;
}

/*
 * duplicate a stack of views for a new viewport
 */
viewhdr_t *view_duplicate(viewhdr_t *v) {
	return view_create(view_findtype(v->typename),
		v->under ? view_duplicate(v->under) : NULL);
}

/* map a view stack onto a specified screen position and buffer */
void view_mapstack(viewhdr_t *v, buffer_t *buffer, int x, int y,
		int width, int height) {
	while (v) {
		v->geom.x = x;
		v->geom.y = y;
		v->geom.width = width;
		v->geom.height = height;
		if (v->buffer != buffer) {
			if (v->buffer)
				buffer_rele(v->buffer);
			v->buffer = buffer_ref(buffer);
		}
		VIEWCALL(v, map);
		v = v->under;
	}
}

/* resize a view without changing buffers */
void view_resize(viewhdr_t *v, int x, int y, int width, int height) {
	viewhdr_t *vtop;

	vtop = v;
	while (v) {
		v->geom.x = x;
		v->geom.y = y;
		v->geom.width = width;
		v->geom.height = height;
		v = v->under;
	}

	VIEWCALL(vtop, geomchange);
}

/* view operation used for an empty view operation */
void view_nop(viewhdr_t *v) {}

/*
 * stack a new view over an old view, and update the entries
 * in the view list so the new one replaces the old one.
 */
viewhdr_t *view_push(viewlist_t *list, viewtype_t *type, viewhdr_t *v) {
	viewhdr_t *after, *new;

	after = TAILQ_PREV(v, viewlist, v_list);
	TAILQ_REMOVE(list, v, v_list);
	new = view_create(type, v);
	new->node = v->node;
	v->node = NULL;
	new->node->d.v = new;
	if (after)
		TAILQ_INSERT_AFTER(list, after, new, v_list);
	else
		TAILQ_INSERT_HEAD(list, new, v_list);
	new->buffer = buffer_ref(v->buffer);
	new->geom.x = v->geom.x;
	new->geom.y = v->geom.y;
	new->geom.width = v->geom.width;
	new->geom.height = v->geom.height;
	VIEWCALL(new, map);

	return new;
}

/* pop a view off the top of a view stack */
void view_pop(viewlist_t *list, viewhdr_t *v) {
	viewhdr_t *after, *under;
	framenode_t *node;

	assert(v->under);

	after = TAILQ_PREV(v, viewlist, v_list);
	TAILQ_REMOVE(list, v, v_list);
	node = v->node;
	v = view_destroy(v);
	if (after)
		TAILQ_INSERT_AFTER(list, after, v, v_list);
	else
		TAILQ_INSERT_HEAD(list, v, v_list);
	v->node = node;
	node->d.v = v;

	/*
	 * set all the top pointers to reflect the removal
	 * of the top node.
	 */
	under = v;
	do {
		under->top = (viewhdr_t *) v;
		under = under->under;
	} while (under);
}
