/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifndef VIEW_H
#define VIEW_H

/*
 * a view is a filter over the input and output of data from a buffer;
 * views control how input is processed, and how the data in the buffer
 * is represented on the screen.  views may be stacked on top of other
 * views, and inherit parts of their functionality.
 *
 * the vdefault view is the main view type, and is always built into
 * the program binary.  any stack of views should have a vdefault at
 * the bottom.
 */

/* number of buckets in the viewtype hashtable */
#define VTYPE_BUCKETS	6

/* operations for view types */
struct viewtypeops {
	/*
	 * create a new view of this type; passes in the view (if any)
	 * that this new view should be stacked on top of.
	 */
	viewhdr_t	*(*create)(viewhdr_t *);
};

/* description of a type of a view */
struct viewtype {
	LIST_ENTRY(viewtype) vt_list;

	viewtypeops_t	op;		/* viewtype operations */
	u_char		*name;		/* name of the viewtype */
};

/* operations for a view */
struct viewops {
	/* view redraw entry point */
	void	(*draw)(viewhdr_t *);

	/*
	 * redraw part of the contents of a view; the position in the
	 * buffer and screen location to draw at are passed in, along
	 * with a limit on how much to draw (in terms of the buffer
	 * text, not in terms of screen).  this view operation is
	 * provided to make it simpler to extend a small portion of
	 * the drawing functionality of a view that uses the vdefault
	 * drawing operation.  other types of views who completly change
	 * the normal drawing behavior need not provide this operation,
	 * and should override draw instead.
	 *
	 * maxcols is the maximum characters to draw from the buffer (this
	 * has no bearing on the number of screen characters taken up); less
	 * than maxcols may be drawn.
	 */
	void	(*drawpart)(viewhdr_t *, line_t **line, int *linenum,
			int *col, int maxcols, int *y, int *viewcol);

	/*
	 * operation to ask the width of a character; views that use
	 * drawpart to override pieces of simple functioniality who
	 * change the width of characters as they are displayed on
	 * the screen should override this routine.  this allows those
	 * changes in character screen width to be available to views
	 * residing underneath by calling from their viewhdr's top pointer.
	 */
	int	(*viswid)(viewhdr_t *, int viewcol, u_char);

	/*
	 * draw the header for the view; all views have a header
	 * of a single line; the view can place any information
	 * in the header that it wants, and the underlying buffer
	 * name should probably be displayed.  because the header
	 * isn't included in the view geometry, the location and
	 * size of the header is passed into this function.
	 */
	void	(*drawhdr)(viewhdr_t *, int x, int y, int width);

	/*
	 * called when the view is focused, and user input should
	 * be handled by the view.  the view may perform any operations
	 * it believes appropriate on the underlying buffer, or pass
	 * some keys (with or without modification) to the lower layer,
	 * or ignore certain keys altogether.
	 */
	void	(*input)(viewhdr_t *, int keysym);

	/*
	 * called when a view is mapped to a screen position over
	 * a buffer; allows the view to set up internal state
	 * as appropriate.
	 */
	void	(*map)(viewhdr_t *);

	/*
	 * sizing changes will occur if the user resizes the tty and
	 * the editor recieves a SIGWINCH, or if views in a frame
	 * are split.  the geomchange call is placed after the view's
	 * geom rect is modified.  this differs from map in that the
	 * buffer is unchanged.
	 */
	void	(*geomchange)(viewhdr_t *);

	/*
	 * notification that the frame this view resides on has become
	 * active; this means the view should be prepared to fully redraw
	 * on the next draw call.
	 */
	void	(*activate)(viewhdr_t *);

	/*
	 * because many views will want to hold state about positions
	 * in a buffer, such as lines and columns in it, it is neccesary
	 * to notify views when another view of the same buffer has changed
	 * a line or range of characters.  the removal op for lines (rmline)
	 * is called immediatly before the remove operation happens (to allow
	 * the view to continue to indirect through any stored line pointers
	 * while it reacts appropriately), the other operations are called
	 * after the operation in question.
	 */
	void	(*rmline)(viewhdr_t *, line_t *, int linenum);
	void	(*rmstr)(viewhdr_t *, line_t *, int linenum,
			int loc, int slen);
	void	(*addline)(viewhdr_t *, line_t *, int linenum);
	void	(*stradd)(viewhdr_t *, line_t *, int linenum,
			int loc, int slen);

	/*
	 * destroy the view; view destruction works from a top to
	 * bottom order, so the destroy routine is allowed to do
	 * things to the view underneath this one (if any).
	 *
	 * it isn't neccesarily true that when this view is
	 * destroying the view(s) beneath it will also be destroyed.
	 *
	 * all views must provide this operation.
	 */
	void	(*destroy)(viewhdr_t *);
};

/* header for all views */
struct viewhdr {
	TAILQ_ENTRY(viewhdr) v_list;	/* view list for this frame */
	framenode_t	*node;		/* node this view is on */

	viewops_t	op;		/* view operations */
	rect_t		geom;		/* screen rect this view uses */
	u_char		*typename;	/* the name of this view's type */
	viewhdr_t	*under;		/* the view stacked below this one */
	viewhdr_t	*top;		/* the top view in the stack */
	buffer_t	*buffer;	/* buffer this view displays */
};

TAILQ_HEAD(viewlist, viewhdr);
LIST_HEAD(viewtypelist, viewtype);

/*
 * macro for the basic part of creation of views; allocate memory and
 * set up top pointers for any underlying views.
 */
#define VIEW_CREATE(vtype, vtname, vunder) ({				\
	viewhdr_t *uv;							\
	vtype *vcreate;							\
									\
	vcreate = ckcalloc(1, sizeof(vtype));				\
	((viewhdr_t *) vcreate)->typename = (vtname);			\
	((viewhdr_t *) vcreate)->under = vunder;			\
									\
	uv = (viewhdr_t *) vcreate;					\
	do {								\
		uv->top = (viewhdr_t *) vcreate;			\
		uv = uv->under;						\
	} while (uv);							\
									\
	vcreate;							\
})

/*
 * macro for registering new view types
 */
#define VIEW_ADDTYPE(viewtypeops, typename) do {			\
	viewtype_t *viewtype;						\
	int idx;							\
									\
	assert(view_findtype(typename) == NULL);			\
	viewtype = ckmalloc(sizeof(viewtype_t));			\
	if (!viewtype)							\
		break;							\
	memcpy(&viewtype->op, &viewtypeops, sizeof(viewtypeops_t));	\
	viewtype->name = typename;					\
	idx = hash_string(VTYPE_BUCKETS, typename);			\
	LIST_INSERT_HEAD(&view_types[idx], viewtype, vt_list);		\
} while (0)

/*
 * used to allow the view operation handlers to take a pointer
 * of a type appropriate for that particular view, rather than
 * a viewhdr_t *.
 */
#define VIEWOP(name)	((void *) (name))

/*
 * view operation calling macro
 */
#define VIEWCALL(v, opname, args...) do {			\
	viewhdr_t *vcallunder = (v);				\
	do {							\
		if (vcallunder->op.opname) {			\
			vcallunder->op.opname(vcallunder	\
				, ## args );			\
			break;					\
		}						\
		vcallunder = vcallunder->under;			\
	} while (vcallunder);					\
} while (0)

/*
 * view operation calling for view operations that have
 * return integer values.
 */
#define VIEWCALLI(v, opname, args...) ({				\
	__label__ found;						\
	viewhdr_t *vcallunder;						\
	int rval;							\
									\
	rval = 0;							\
	vcallunder = (v);						\
	do {								\
		if (vcallunder->op.opname) {				\
			rval = vcallunder->op.opname(vcallunder		\
				, ## args);				\
			goto found;					\
		}							\
		vcallunder = vcallunder->under;				\
	} while (vcallunder);						\
	rval = -1;							\
found:									\
	rval;								\
})

/*
 * view operation calling for all views on a particular
 * buffer.
 */
#define VIEWCALLB(buffer, opname, args...) do {				\
	frame_t *vcallfr;						\
	viewhdr_t *vcallb;						\
									\
	TAILQ_FOREACH(vcallfr, &frame_list, f_list)			\
		TAILQ_FOREACH(vcallb, &vcallfr->views, v_list)		\
			if (vcallb->buffer == (buffer))			\
				VIEWCALL(vcallb, opname , ## args );	\
} while (0)

/*
 * macro to obtain the view with the input focus; it is always on
 * the end of the views list in the active frame.
 */
#define view_current	TAILQ_LAST(&frame_active->views, viewlist)

extern viewtypelist_t	*view_types;

void		view_init();
void		view_shutdown();
viewtype_t	*view_findtype(u_char *name);
void		view_rmtype(u_char *name);
viewhdr_t	*view_create(viewtype_t *, viewhdr_t *under);
viewhdr_t	*view_destroy(viewhdr_t *);
viewhdr_t	*view_duplicate(viewhdr_t *);
void		view_mapstack(viewhdr_t *, buffer_t *buffer, int x, int y,
			int width, int height);
void		view_resize(viewhdr_t *, int x, int y, int width,
			int height);
void		view_nop(viewhdr_t *);
viewhdr_t	*view_push(viewlist_t *, viewtype_t *, viewhdr_t *v);
void		view_pop(viewlist_t *, viewhdr_t *v);

/*
 * set the buffer a view is looking at; implement by remaping the
 * view to it's current position with the new buffer, so a new
 * view callback isn't required.
 */
static __inline void view_setbuffer(viewhdr_t *v, buffer_t *buffer) {
	view_mapstack(v, buffer, v->geom.x, v->geom.y,
		v->geom.width, v->geom.height);
}

/*
 * returns the vdefault in a view stack; this allows views to be
 * built that use/modify some of the vdefault internals but can have
 * any number of views stacked between them and the vdefault.
 */
static __inline vdefault_t *view_findvdef(viewhdr_t *v) {
	while (v) {
		if (!strcmp(v->typename, "vdefault"))
			return (vdefault_t *) v;
		v = v->under;
	}

	assert(0);
	return NULL;
}

#endif
