/*-
 * Copyright (c) 2002 Jordan DeLong
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of contributors may be
 *    used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "editor.h"

static __inline void vminibuff_redraw(vminibuff_t *v) {
	int i;

	screen_setcolor(COLOR_HEADING);
	screen_gotoxy(v->hdr.geom.x, v->hdr.geom.y);
	screen_puts(" -");
	for (i = 0; i < v->hdr.geom.width - 4; i++)
		screen_putc(' ');
	screen_puts("- ");
}

static __inline void vminibuff_drawmsg(vminibuff_t *v) {
	line_t *line;
	int len, i;

	screen_setcolor(COLOR_HEADING);
	screen_gotoxy(v->hdr.geom.x, v->hdr.geom.y);
	screen_puts(" | ");
	line = TAILQ_FIRST(&v->hdr.buffer->lines);
	len = min(line->length, v->hdr.geom.width - 6);
	screen_nputs(line->text, len);
	for (i = len + 3; i < v->hdr.geom.width - 3; i++)
		screen_putc(' ');
	screen_gotoxy(v->hdr.geom.x + v->hdr.geom.width - 3, v->hdr.geom.y);
	screen_puts(" | ");
}

static __inline void vminibuff_drawprompt(vminibuff_t *v) {
	vdefault_t *vdefault;

	/* we need to use vdefault members for scrolling */
	vdefault = (vdefault_t *) v->hdr.under;

	/* do the left side trimming */
	screen_setcolor(COLOR_HEADING);
	screen_gotoxy(v->hdr.geom.x, v->hdr.geom.y);
	screen_puts(v->inputcback ? " * " : " | ");

	/* draw the prompt, if one is specified */
	if (v->prompt) {
		screen_nputs(v->prompt, strlen(v->prompt));
		screen_putc(':');
		screen_setcolor(COLOR_DEFAULT);
		screen_putc(' ');
	}

	/* the right side trimming */
	screen_setcolor(COLOR_HEADING);
	screen_gotoxy(v->hdr.geom.x + v->hdr.geom.width - 3, v->hdr.geom.y);
	screen_puts(v->inputcback ? " * " : " | ");

	/*
	 * to draw the user's text we just hand off to the
	 * vdefault draw operation; this is possible because
	 * we are hiding the real view size from vdefault by
	 * manipulating its geom structure.
	 */
	VIEWCALL((viewhdr_t *) vdefault, draw);
}

/* draw a minibuffer view */
static void vminibuff_draw(vminibuff_t *v) {
	switch (v->state) {
	case MINIBUFF_NORMAL:
		break;
	case MINIBUFF_NEEDREDRAW:
		vminibuff_redraw(v);
		v->state = MINIBUFF_NORMAL;
		break;
	case MINIBUFF_DRAWMSG:
		vminibuff_drawmsg(v);
		v->state = MINIBUFF_NEEDREDRAW;
		break;
	case MINIBUFF_PROMPT:
		vminibuff_drawprompt(v);
		break;
	}
}

/*
 * find as much as can be unambiguously completed and
 * append that to the text the user is typing.
 */
static void vminibuff_comp_unambig(vminibuff_t *v, int startidx,
		u_char **comps) {
	line_t *line;
	int unambig, i, n;
	int len;

	/* need at least one completion */
	if (!comps[0])
		return;

	/* find out how much is unambiguous, if any */
	unambig = INT_MAX;
	line = TAILQ_FIRST(&v->hdr.buffer->lines);
	for (i = 1; comps[i]; i++) {
		for (n = 0; comps[i][n] && comps[i - 1][n]; n++)
			if (comps[i][n] != comps[i - 1][n])
				break;

		if (n < unambig)
			unambig = n;
		if (!unambig)
			return;
	}
	if (unambig == INT_MAX)
		unambig = strlen(comps[0]);

	/* complete as much unambiguous text as possible */
	assert(unambig > 0);
	len = line->length - startidx;
	if (unambig - len == 0)
		return;
	buffer_nputs(v->hdr.buffer, 0, line, 0, line->length,
		&comps[0][len], unambig - len);

	/* put the cursor at the end of the line */
	VIEWCALL(v->hdr.under, input, KEYSYM_END);
}

/* print the possible completions in a columned format */
static void vminibuff_showcomps(vminibuff_t *v, u_char **comps,
		line_t *line) {
	int longest, count, linenum, len;
	int numcols, numrows, endcol, base, chcnt, cnt;
	int r, c, i;

	count = longest = linenum = 0;

	/* find out what the longest item is */
	for (i = 0; comps[i]; i++, count++) {
		len = strlen(comps[i]);
		if (len > longest)
			longest = len;
	}

	/*
	 * figure out how much space between columns; if the amount
	 * we'd need is longer than the width of this view, just do a
	 * simple listing of the results
	 */
	longest = (longest + options.tabwidth) & ~(options.tabwidth - 1);
	if (longest >= v->popup_view->geom.width) {
		for (i = 0; comps[i]; i++) {
			buffer_nputs(v->popup_buffer, 0, line, linenum, 0,
				comps[i], strlen(comps[i]));
			line = buffer_addline(v->popup_buffer, 0,
				line, linenum++);
		}

		return;
	}

	/*
	 * send the output to the popup view in columns; this is
	 * adapted from FreeBSD's /usr/src/usr.bin/column/column.c
	 */

	numcols = v->popup_view->geom.width / longest;
	numrows = count / numcols;
	if (count % numcols)
		numrows++;

	for (r = 0; r < numrows; r++) {
		endcol = longest;
		for (base = r, chcnt = c = 0; c < numcols; c++) {
			len = strlen(comps[base]);
			chcnt += len;
			buffer_nputs(v->popup_buffer, 0, line, linenum,
				line->length, comps[base], len);
			if ((base += numrows) >= count)
				break;
			while ((cnt = ((chcnt + options.tabwidth) &
					~(options.tabwidth - 1))) <= endcol) {
				buffer_putc(v->popup_buffer, 0, line, linenum,
					line->length, '\t');
				chcnt = cnt;
			}
			endcol += longest;
		}

		line = buffer_addline(v->popup_buffer, 0, line, linenum++);
	}
}

/* complete the user's text for minibuffer */
static void vminibuff_complete(vminibuff_t *v) {
	u_char **comps;
	line_t *line;
	int startidx;
	int i;

	/*
	 * get the possible completions for what has been typed
	 * so far from the datatype-specific completion rountine.
	 */
	comps = v->compfunc(v, &startidx);
	if (!comps)
		return;
	sortwords(comps);
	vminibuff_comp_unambig(v, startidx, comps);

	/*
	 * don't bother with showing completions if we there
	 * wasn't any ambiguity.
	 */
	if (!comps[0] || !comps[1])
		goto free;

	/*
	 * create the popup if it isn't visible yet, or empty
	 * out an existing one.  then call the routine to display
	 * the possible completions.
	 */
	if (!v->popup_view) {
		v->popup_buffer = buffer_create("*completions", 1);
		line = TAILQ_FIRST(&v->popup_buffer->lines);
		vminibuff_popup(v, v->popup_buffer);
	} else {
		for (;;) {
			line = TAILQ_FIRST(&v->popup_buffer->lines);
			if (line == TAILQ_LAST(&v->popup_buffer->lines,
					linelist))
				break;
			buffer_rmline(v->popup_buffer, 0, line, 1);
		}
		VIEWCALL(v->popup_view, map);
	}
	vminibuff_showcomps(v, comps, line);

free:
	for (i = 0; comps[i]; i++)
		free(comps[i]);
	free(comps);
}

/*
 * handle input for a minibuffer when it is editing something; most
 * editing operations are passed through to the underlying view
 * type (likely a vdefault), but certain things that would cause
 * opening of a new line aren't allowed.  hitting enter will accept
 * the minibuffer input.
 */
static void vminibuff_input(vminibuff_t *v, int keysym) {
	killbuffer_t *killbuff;

	assert(v->state == MINIBUFF_PROMPT);

	/* call the input callback if there is one */
	if (v->inputcback && v->inputcback(v, keysym))
		goto post;

	if (input_lastkeysym == KEYSYM_QUOTENEXT) {
		VIEWCALL(v->hdr.under, input, keysym);
		goto post;
	}

	switch (keysym) {
	case KEYSYM_CANCEL:
		v->canceled = 1;
	case '\r': case '\n':
	case KEYSYM_ENTER:
		vminibuff_exit(v);
		break;
	case KEYSYM_SUSPEND:
	case KEYSYM_LEFT:
	case KEYSYM_RIGHT:
	case KEYSYM_HOME:
	case KEYSYM_END:
	case KEYSYM_NEXTWORD:
	case KEYSYM_PREVWORD:
	case KEYSYM_DELETE:
	case KEYSYM_BACKSPACE:
	case KEYSYM_UNDO:
	case KEYSYM_REDO:
	case KEYSYM_TRANCHARS:
	case KEYSYM_KILLWORD:
	case KEYSYM_FWDKILLWORD:
	case KEYSYM_SETMARK:
	case KEYSYM_PROMOTEMARK:
	case KEYSYM_DEMOTEMARK:
	case KEYSYM_XCHGMARK:
	case KEYSYM_RECORD:
	case KEYSYM_PLAYBACK:
		VIEWCALL(v->hdr.under, input, keysym);
		break;
	case KEYSYM_KILLEOL: {
		vdefault_t *vdef;

		vdef = view_findvdef((viewhdr_t *) v);
		if (vdef->cursor.col == vdef->cursor.line->length)
			break;
	}
	case KEYSYM_KILLLINE:
		VIEWCALL(v->hdr.under, input, keysym);
		TAILQ_FIRST(&killring.buffers)->trailnl = 0;
		break;

	/* allow yanks, but prevent yanking multi-line kills */
	case KEYSYM_YANK:
	case KEYSYM_YANKPOP:
	case KEYSYM_FWDYANKPOP:
		killbuff = killring.current;
		if (!killbuff)
			break;
		if (keysym == KEYSYM_YANKPOP) {
			killbuff = TAILQ_NEXT(killbuff, kb_list);
			if (!killbuff) {
				killbuff = TAILQ_FIRST(&killring.buffers);
				if (!killbuff)
					break;
			}
		} else if (keysym == KEYSYM_FWDYANKPOP) {
			killbuff = TAILQ_PREV(killbuff, killbufferlist, kb_list);
			if (!killbuff) {
				killbuff = TAILQ_LAST(&killring.buffers, killbufferlist);
				if (!killbuff)
					break;
			}
		}
		if (!TAILQ_FIRST(&killbuff->lines))
			break;
		if (TAILQ_NEXT(TAILQ_FIRST(&killbuff->lines), l_list))
			break;
		VIEWCALL(v->hdr.under, input, keysym);
		break;

	/* allow scrolling of a popup window */
	case KEYSYM_PGUP:
	case KEYSYM_PGDN:
		if (v->popup_view)
			VIEWCALL(v->popup_view, input, keysym);
		break;

	default:
		if (keysym == '\t' && v->compfunc)
			vminibuff_complete(v);
		else if (keysym <= UCHAR_MAX)
			VIEWCALL(v->hdr.under, input, keysym);
		else
			screen_bell();
	}

post:
	/* call the input callback if there is one */
	if (v->postinputcback)
		v->postinputcback(v, keysym);
}

/* handle view mappage */
static void vminibuff_map(vminibuff_t *v) {
	v->canceled = 0;
	v->state = MINIBUFF_NEEDREDRAW;
	v->popup_view = NULL;
	v->inputcback = NULL;
	v->compfunc = NULL;
}

/* handle changes in the size of the minibuff (when winch's happen) */
static void vminibuff_geomchange(vminibuff_t *v) {
	if (v->state == MINIBUFF_PROMPT)
		vminibuff_hidesz(v);
	else
		v->state = MINIBUFF_NEEDREDRAW;
}

/* create a new minibuffer view */
static viewhdr_t *vminibuff_create(viewhdr_t *under) {
	vminibuff_t *vminibuff;

	/*
	 * this only supports stacking directly on top of
	 * a vdefault view.
	 */
	assert(under && !strcmp(under->typename, "vdefault"));

	/* allocate the view structure */
	vminibuff = VIEW_CREATE(vminibuff_t, "vminibuff", under);

	/* set up view operations */
	vminibuff->hdr.op.draw		= VIEWOP(vminibuff_draw);
	vminibuff->hdr.op.drawpart	= NULL;
	vminibuff->hdr.op.viswid	= NULL;
	vminibuff->hdr.op.drawhdr	= VIEWOP(view_nop);
	vminibuff->hdr.op.input		= VIEWOP(vminibuff_input);
	vminibuff->hdr.op.map		= VIEWOP(vminibuff_map);
	vminibuff->hdr.op.geomchange	= VIEWOP(vminibuff_geomchange);
	vminibuff->hdr.op.activate	= NULL;
	vminibuff->hdr.op.rmline	= VIEWOP(abort);
	vminibuff->hdr.op.rmstr		= NULL;
	vminibuff->hdr.op.stradd	= NULL;
	vminibuff->hdr.op.destroy	= VIEWOP(free);

	return (viewhdr_t *) vminibuff;
}

/* type operations for minibuffer views */
static viewtypeops_t vminibuff_typeops = {
	vminibuff_create
};

/* register the minibuff view type */
void vminibuff_init() {
	VIEW_ADDTYPE(vminibuff_typeops, "vminibuff");
}

/* manipulate the size of the underlying vdefault */
void vminibuff_hidesz(vminibuff_t *v) {
	int promptlen;
	int tried;

	/*
	 * hide the real size from the underlying vdefault, so
	 * it properly scrolls itself.  take into account the prompt,
	 * the trim padding of " | " (3) on each side of the minibuff,
	 * and the ": " (2) placed after the prompt.
	 *
	 * if we aren't big enough for at least 1 column of vdefault,
	 * cancel the minibuffer edit immediately.
	 */
	tried = 0;
retry:
	v->hdr.under->geom.x = v->hdr.geom.x + 3;
	v->hdr.under->geom.width = v->hdr.geom.width - 6;
	if (v->prompt) {
		promptlen = strlen(v->prompt);
		v->hdr.under->geom.x += promptlen + 2;
		v->hdr.under->geom.width -= promptlen + 2;
	}

	/* XXX */
	if (v->hdr.under->geom.width <= 0) {
		if (tried)
			/*
			 * unreasonable window size.
			 */
			abort();
		v->prompt = "...";
		tried = 1;
		goto retry;
	}

	/* let the vdefault know about it */
	VIEWCALL(v->hdr.under, geomchange);
}

/*
 * popup a viewport displaying data in buffer; we shadow whatever
 * type of view is here with a vdefault (which wont pass anything
 * through to the view underneath it) to make sure we have the
 * behavior we want for drawing.
 */
void vminibuff_popup(vminibuff_t *v, buffer_t *buffer) {
	v->popup_view = view_push(&frame_active->views,
		view_findtype("vdefault"), view_current);

	/*
	 * don't set the buffer through normal means so the
	 * view underneath wont hear about it.
	 */
	buffer_rele(v->popup_view->buffer);
	v->popup_view->buffer = buffer_ref(buffer);
	VIEWCALL(v->popup_view, map);
}

/* prepare for minibuffer entry */
void vminibuff_prepare(vminibuff_t *v, u_char *prompt, u_char *def) {
	vdefault_t *vdefault;
	line_t *line;
	int deflen;

	/*
	 * make sure there are no attempts to enter the minibuffer
	 * when it is already editing a prompt
	 */
	assert(v->state != MINIBUFF_PROMPT);

	/* prepare for entry */
	v->state = MINIBUFF_PROMPT;
	v->prompt = prompt;
	vminibuff_hidesz(v);
	v->canceled = 0;
	if (v->canceled)
		return;
	line = TAILQ_FIRST(&v->hdr.buffer->lines);
	line->length = 0;
	vdefault = (vdefault_t *) v->hdr.under;
	vdefault->view.col = 0;
	vdefault->redraw_text = 1;
	if (def) {
		deflen = strlen(def);
		line_nputs(line, 0, def, deflen);
		vdefault->cursor.col = deflen;
	} else
		vdefault->cursor.col = 0;
	vdefault_setscreencol(vdefault);
}

void vminibuff_enter(vminibuff_t *v) {
	v->canceled = 0;
	vminibuff_hidesz(v);

	/* enter an input loop for minibuffer prompting */
	input_recurse();

	/* free up any popup that may exist */
	if (v->popup_view) {
		view_pop(&frame_active->views, v->popup_view);
		VIEWCALL(view_current, activate);
		v->popup_view = NULL;
	}
}

/*
 * temporarily exit the minibuffer (to be reentered); this only is
 * workable if you are going to reenter; the real vminibuff_exit
 * must be called on the final exit.
 */
void vminibuff_tmpexit(vminibuff_t *v) {
	input_exitloop = 1;
}

/* exit input loop doing editing on a minibuff prompt */
void vminibuff_exit(vminibuff_t *v) {
	input_exitloop = 1;
	undo_emptylist(v->hdr.buffer);
	v->inputcback = NULL;
	v->postinputcback = NULL;
	v->compfunc = NULL;
	v->prompt = NULL;
	v->state = MINIBUFF_NEEDREDRAW;
	VIEWCALL(v->hdr.under, map);
}
